(* ----------------------------------------------------------------------- *)
UNIT DATA_;
(* ----------------------------------------------------------------------- *)

INTERFACE
(* ----------------------------------------------------------------------- *)

USES

  SysUtils,
  Graphics,

  CONSTANTS_,
  DRAW_,
  ELEMENTS_;
(* ----------------------------------------------------------------------- *)

TYPE

  KeyData = Class(Elements)
   (* --- Basic Description ---------------------------------------------- *)
    Key    : Integer;  // A unique identifier - here an integer for example
    Color  : TColor;   // Its color for visualization purposes
    Number : Integer;  // The index of this elements in the DS it belongs
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( S: Integer; C: Codes);
    Constructor SetAs( S: String; C: Codes);
   (* --- Handling Operations -------------------------------------------- *)
    Function    GetNumber: Integer;                                Override;
    Procedure   SetNumber( N: Integer);                            Override;
    Function    GetColor: TColor;                                  Override;
    Procedure   SetColor( C: TColor);                              Override;
   (* --- Visualization -------------------------------------------------- *)
    Function    AsText: String;                                    Override;
    Procedure   Draw( W: TCanvas; k,l: Integer);                   Override;
   (* -------------------------------------------------------------------- *)
  end;

Function NumericOrder( E1,E2: Elements): Codes;
Function HashLocation( Key, Range, ReTry: Integer): Integer;
(* ======================================================================= *)

IMPLEMENTATION

Constructor KeyData.SetUp( S: Integer; C: Codes);
begin
  Key   := S;
  Color := C;
end;

constructor KeyData.SetAs(S: String; C: Codes);
begin
  Key   := StrToInt(S);
  Color := C;
end;

Function KeyData.AsText: String;
begin
  AsText := Format('%d', [Key]);
end;

Procedure KeyData.Draw( W: TCanvas; k,l: Integer);
Var
  Adjust,
  Dx, Dy : Integer;
  S      : String;
begin
  With W do begin
    Dx := CellWidth;
    Dy := CellHeight;
    S := AsText;
    Font.Size := FontSize;
    Font.Color := _KeyColor_;
    DrawFrame( W, k-2,l-1, k+Dx+2, l+Dy+1, Color);
   (* --- Record the 'clickable' areas ------ *)
    If NmrOfHitAreas<_MaxHitAreas_ then begin
      Inc(NmrOfHitAreas);
      With HitAreas[NmrOfHitAreas] do begin
        x1 := k;     y1 := l;
        x2 := x1+Dx; y2 := y1+Dy;
        Hit := S; // = AsText
      end;
    end;
   (* --- Do not show content if small ------ *)
    If Not (FontSize<6) then begin
      Adjust := (CellWidth-TextWidth(S)) Div 2;
      TextOut( k+Adjust,l, S );
      If ShowNumbers and (Number>0) then begin
        S  := Format(' %d:', [Number]);
        Font.Size := FontSize-1;
        Font.Color := _NumberColor_;
        Brush.Style := bsSolid;
        Brush.Color := _FillColor_;
        TextOut( k-TextWidth(S)-4, l, S );
      end;
    end;
   (* --------------------------------------- *)
  end;
end;

Function KeyData.GetColor: TColor;
begin
  GetColor := Color;
end;

Procedure KeyData.SetColor( C: TColor);
begin
  If MarkUsed then
    Case C=_UsedColor_ of
      Ok: If (Color=_FillColor_) then Color := C;
      No: Color := C;
    end;
end;

Function KeyData.GetNumber: Integer;
begin
  GetNumber := Number;
end;

Procedure KeyData.SetNumber(N: Integer);
begin
  Number := N;
end;

Function NumericOrder(E1,E2: Elements): Codes;
Var
  K1, K2: KeyData;
begin
  Try
    K1 := (E1 As KeyData);
    K2 := (E2 As KeyData);
    If K1.Key < K2.Key then NumericOrder := _Less_ else
    If K1.Key = K2.Key then NumericOrder := _Equal_ else
    If K1.Key > K2.Key then NumericOrder := _Greater_;
    //
   (* --- Color relevant element as 'USED' ------------------- *)
    E1.Color := _UsedColor_;
    E2.Color := _UsedColor_;
   (* -------------------------------------------------------- *)
    //
  except
    NumericOrder := _Null_;
  end;
end;

Function HashLocation( Key, Range, ReTry: Integer): Integer;
begin
  // Just for example...
  HashLocation := ((Key * 509) + Retry*37 ) Mod Range + 1;
end;

(* ======================================================================= *)

begin
end.

