(* ----------------------------------------------------------------------- *)
UNIT DSTRUCTURES_;
(* ----------------------------------------------------------------------- *)

INTERFACE
(* ----------------------------------------------------------------------- *)

USES

  SysUtils,
  Graphics,

  CONSTANTS_,
  ELEMENTS_,
  REPORTING_,
  DRAW_,
  DATA_,
  LINKS_,
  SSCHEMES_;
(* ----------------------------------------------------------------------- *)

TYPE

  Structures = Class( Elements )
   (* --- Basic Description ---------------------------------------------- *)
    Size  : Integer;          // Number of elements inserted so far
    Order : OrderFunctions;   // Primary order defined on its elements
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp;
    Constructor Build( S: Structures);
    Destructor  Clear;                                             Override;
    Function    AsText: String;                                    Override;
   (* --- Store Methods -------------------------------------------------- *)
    Function    Scheme : StorageSchemes;                            Virtual;
    Procedure   Insert( E: Elements);                               Virtual;
    Function    Delete{Current}: Elements;                          Virtual;
    Procedure   Promote( {Current to:} P: Elements);                Virtual;
   (* --- Union/Split Methods -------------------------------------------- *)
    Procedure   Split( Var E,S: Elements; Var S1,S2: Structures);   Virtual;
    Procedure   Union ( E: Elements; S1,S2: Structures );           Virtual;
   (* --- Search Methods ------------------------------------------------- *)
    Function    Find( C: Codes): Elements;                          Virtual;
    Function    FindKey( K: Elements): Elements;                    Virtual;
    Function    FindNmr( N: Integer): Elements;                     Virtual;
   (* --- Report Methods ------------------------------------------------- *)
    Function    IndexOf( E: Elements): Integer;                     Virtual;
    Procedure   ReportLess( Up: Elements; R: Reporter);             Virtual;
    Procedure   ReportInRange( Lf, Rt : Elements; R: Reporter);     Virtual;
    Procedure   ReportInArea( Up,Lf,Rt: Elements; R: Reporter);     Virtual;
   (* -------------------------------------------------------------------- *)
  end;

(* --- CATALOGUES -------------------------------------------------------- *)

  _Catalogs = Class( Structures )
   (* --- Basic Description ---------------------------------------------- *)
                                                      //Abstract Structure
   (* -------------------------------------------------------------------- *)
  end;

  CAT_Tables = Class( _Catalogs )
   (* --- Basic Description ---------------------------------------------- *)
    Table   : Tables;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( T: Tables);
    Function    Scheme : StorageSchemes;                           Override;
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: Elements);                              Override;
    Function    Delete: Elements;                                  Override;
   (* -------------------------------------------------------------------- *)
  end;

  CAT_DynamicTables = Class( CAT_Tables)
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: Elements);                              Override;
    Function    Delete: Elements;                                  Override;
   (* -------------------------------------------------------------------- *)
  end;

  CAT_Lists = Class( _Catalogs )
   (* --- Basic Description ---------------------------------------------- *)
    List : Lists;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( L: Lists);
    Function    Scheme : StorageSchemes;                           Override;
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: Elements);                              Override;
    Function    Delete: Elements;                                  Override;
   (* -------------------------------------------------------------------- *)
  end;

(* --- STACKS ------------------------------------------------------------ *)

  _Stacks = Class( Structures )
   (* --- Basic Description ---------------------------------------------- *)
                                                      //Abstract Structure
   (* -------------------------------------------------------------------- *)
    Procedure   Push( E: Elements);                                 Virtual;
    Function    Pop: Elements;                                      Virtual;
    Function    Top: Elements;                                      Virtual;
   (* -------------------------------------------------------------------- *)
  end;

  STACK_Tables = Class( _Stacks )
   (* --- Basic Description ---------------------------------------------- *)
    Table   : Tables;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( T: Tables);
    Function    Scheme : StorageSchemes;                           Override;
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: Elements);                              Override;
    Function    Delete: Elements;                                  Override;
   (* --- Search Methods ------------------------------------------------- *)
    Function    Find( C: Codes): Elements;                         Override;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   DrawLinks( W: TCanvas; k,l: Integer);              Override;
   (* -------------------------------------------------------------------- *)
  end;

  STACK_Lists = Class( _Stacks )
   (* --- Basic Description ---------------------------------------------- *)
    List : Lists;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( L: Lists);
    Function    Scheme : StorageSchemes;                           Override;
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: ELements);                              Override;
    Function    Delete: Elements;                                  Override;
   (* --- Search Methods ------------------------------------------------- *)
    Function    Find( C: Codes): Elements;                         Override;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   DrawLinks( W: TCanvas; k,l: Integer);              Override;
   (* -------------------------------------------------------------------- *)
  end;

(* ----------------------------------------------------------------------- *)

  _Queues = Class( Structures )
   (* --- Basic Description ---------------------------------------------- *)
                                                      //Abstract Structure
   (* -------------------------------------------------------------------- *)
    Procedure   EnQueue(E: Elements);                               Virtual;
    Function    DeQueue: Elements;                                  Virtual;
    Function    Top: Elements;                                      Virtual;
   (* -------------------------------------------------------------------- *)
  end;

  QUEUE_Lists = Class( _Queues)
   (* --- Basic Description ---------------------------------------------- *)
    List : Lists;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( L: Lists);
    Function    Scheme : StorageSchemes;                           Override;
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: ELements);                              Override;
    Function    Delete: Elements;                                  Override;
   (* --- Search Methods ------------------------------------------------- *)
    Function    Find( C: Codes): Elements;                         Override;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   DrawLinks( W: TCanvas; k,l: Integer);              Override;
   (* -------------------------------------------------------------------- *)
  end;

  QUEUE_CircularTables = Class( _Queues)
   (* --- Basic Description ---------------------------------------------- *)
    CircularTable : CircularTables;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( T: CircularTables);
    Function    Scheme : StorageSchemes;                           Override;
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: ELements);                              Override;
    Function    Delete: Elements;                                  Override;
   (* --- Search Methods ------------------------------------------------- *)
    Function    Find( C: Codes): Elements;                         Override;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   DrawLinks( W: TCanvas; k,l: Integer);              Override;
   (* -------------------------------------------------------------------- *)
  end;

(* ----------------------------------------------------------------------- *)

  _PriorityQueues = Class( Structures )
   (* --- Basic Description ---------------------------------------------- *)
                                                      //Abstract Structure
   (* -------------------------------------------------------------------- *)
    Procedure   EnQueue( E: Elements);                              Virtual;
    Function    DeQueue: Elements;                                  Virtual;
    Function    Top: Elements;                                      Virtual;
    Procedure   Promote( P: Elements);                             Override;
   (* -------------------------------------------------------------------- *)
  end;

  PQ_HeapTables = Class( _PriorityQueues )
   (* --- Basic Description ---------------------------------------------- *)
    Table : Tables;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( T: Tables);
    Function    Scheme : StorageSchemes;                           Override;
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: Elements);                              Override;
    Function    Delete: Elements;                                  Override;
    Procedure   Promote( P: Elements);                             Override;
   (* --- Search Methods ------------------------------------------------- *)
    Function    Find( C: Codes): Elements;                         Override;
   (* --- Report Methods ------------------------------------------------- *)
    Procedure   ReportLess( Up: Elements; R: Reporter);            Override;
   (* --- Auxiliary Methods ---------------------------------------------- *)
    Procedure   BubbleUp( P: Integer);
    Procedure   BubbleDown( Var P: Integer);
   (* --- Visualization -------------------------------------------------- *)
    Procedure   DrawLinks( W: TCanvas; k,l: Integer);               Override;
   (* -------------------------------------------------------------------- *)
  end;

  PQ_BinaryHeaps = Class( _PriorityQueues )
   (* --- Basic Description ---------------------------------------------- *)
    Tree : BinaryTrees;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( T: BinaryTrees);
    Function    Scheme : StorageSchemes;                           Override;
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: Elements);                              Override;
    Function    Delete: Elements;                                  Override;
    Procedure   Promote( P: Elements);                             Override;
   (* --- Search Methods ------------------------------------------------- *)
    Function    Find( C: Codes): Elements;                         Override;
   (* --- Report Methods ------------------------------------------------- *)
    Procedure   ReportLess( Up: Elements; R: Reporter);            Override;
   (* --- Auxiliary Methods ---------------------------------------------- *)
    Procedure   BubbleUp( Var L: Elements);
    Procedure   BubbleDown( Var L: Elements);
   (* --- Visualization -------------------------------------------------- *)
    Procedure   DrawLinks( W: TCanvas; k,l: Integer);              Override;
   (* -------------------------------------------------------------------- *)
  end;

  PQ_BinomialHeaps = Class( _PriorityQueues )
   (* --- Basic Description ---------------------------------------------- *)
    Forest     : Forests;
    TopLink    : Elements;
    AlwaysPack : Boolean;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp;
    Function    Scheme : StorageSchemes;                           Override;
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert(E: Elements);                               Override;
    Function    Delete: Elements;                                  Override;
    Procedure   Promote( P: Elements);                             Override;
   (* --- Search Methods ------------------------------------------------- *)
    Function    Find( C: Codes ): Elements;                        Override;
   (* --- Report Methods ------------------------------------------------- *)
    Procedure   ReportLess( Up: Elements; R: Reporter);            Override;
   (* --- Auxiliary Methods ---------------------------------------------- *)
    Procedure   BubbleUp( Var L: Elements);
    Procedure   MoveToRoot( Var L: Elements);
    Procedure   HangFrom( Var TreeA, TreeB: AnyDegTrees);           Virtual;
    Procedure   PackRootLevel;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   DrawLinks( W: TCanvas; k,l: Integer);              Override;
   (* -------------------------------------------------------------------- *)
  end;

  PQ_AmortizedBinomialHeaps = Class( PQ_BinomialHeaps )
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp;
   (* -------------------------------------------------------------------- *)
  end;

  PQ_FibonacciHeaps = Class( PQ_AmortizedBinomialHeaps )
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Promote( P: Elements);                             Override;
   (* --- Auxiliary Methods ---------------------------------------------- *)
    Procedure   HangFrom( Var TreeA, TreeB: AnyDegTrees);          Override;
   (* -------------------------------------------------------------------- *)
  end;

  PQ_LeftistHeaps = Class( PQ_BinaryHeaps )
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: Elements);                              Override;
    Function    Delete: Elements;                                  Override;
   (* --- Auxiliary Methods ---------------------------------------------- *)
    Function    Meld( L,R: Elements): Elements;
    Procedure   HandleRanks( E: Elements);                          Virtual;
   (* -------------------------------------------------------------------- *)
  end;

  PQ_SkewHeaps = Class( PQ_LeftistHeaps )
   (* --- Auxiliary Methods ---------------------------------------------- *)
    Procedure   HandleRanks( E: Elements);                         Override;
   (* -------------------------------------------------------------------- *)
  end;

  PQ_PairingHeaps = Class( PQ_BinaryHeaps )
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: Elements);                              Override;
    Function    Delete: Elements;                                  Override;
    Procedure   Promote( P: Elements);                             Override;
   (* --- Report Methods ------------------------------------------------- *)
    Procedure   ReportLess( Up: Elements; R: Reporter);            Override;
   (* --- Auxiliary Methods ---------------------------------------------- *)
    Function    Meld( L,R: Elements): Elements;
    Procedure   HeapSplayUp( E: Elements);
   (* -------------------------------------------------------------------- *)
  end;

(* ----------------------------------------------------------------------- *)

  _Directories = Class( Structures )
   (* --- Basic Description ---------------------------------------------- *)
                                                      //Abstract Structure
   (* -------------------------------------------------------------------- *)
  end;

  DIR_SortedTables = Class( _Directories)
   (* --- Basic Description ---------------------------------------------- *)
    Table : Tables;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( T: Tables);
    Function    Scheme: StorageSchemes;                            Override;
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: Elements);                              Override;
    Function    Delete: Elements;                                  Override;
   (* --- Search Methods ------------------------------------------------- *)
    Function    Find( C: Codes ): Elements;                        Override;
    Function    FindKey( K: Elements): Elements;                   Override;
   (* --- Report Methods ------------------------------------------------- *)
    Procedure   ReportLess( Up: Elements; R: Reporter);            Override;
    Procedure   ReportInRange( Lf,Rt: Elements; R: Reporter);      Override;
   (* --- Auxiliary Methods ---------------------------------------------- *)
    Function    BinarySearch( K: Elements; Var Found: Boolean): Integer;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   DrawLinks( W: TCanvas; k,l: Integer);              Override;
   (* -------------------------------------------------------------------- *)
  end;

  DIR_BinaryTrees = Class( _Directories )
   (* --- Basic Description ---------------------------------------------- *)
    Tree    : BinaryTrees;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( T: BinaryTrees);
    Function    Scheme : StorageSchemes;                           Override;
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: Elements);                              Override;
    Function    Delete: Elements;                                  Override;
   (* --- Search Methods ------------------------------------------------- *)
    Function    Find( C: Codes): Elements;                         Override;
    Function    FindKey( K: Elements): Elements;                   Override;
    Function    IndexOf( E: Elements): Integer;                    Override;
   (* --- Report Methods ------------------------------------------------- *)
    Procedure   ReportLess( Up: Elements; R: Reporter);            Override;
    Procedure   ReportInRange( Lf,Rt: ELements; R: Reporter);      Override;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   DrawLinks( W: TCanvas; k,l: Integer);              Override;
   (* -------------------------------------------------------------------- *)
  end;

  DIR_SkipLists = Class( _Directories )
   (* --- Basic Description ---------------------------------------------- *)
    SkipList : SkipLists;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( SL: SkipLists);
    Function    Scheme : StorageSchemes;                           Override;
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: Elements);                              Override;
    Function    Delete: Elements;                                  Override;
   (* --- Search Methods ------------------------------------------------- *)
    Function    Find( C: Codes): Elements;                         Override;
    Function    FindKey( K: Elements): Elements;                   Override;
   (* --- Report Methods ------------------------------------------------- *)
    Procedure   ReportInRange( Lf, Rt : Elements; R: Reporter);    Override;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   DrawLinks( W: TCanvas; k,l: Integer);              Override;
   (* --- Auxiliary Methods ---------------------------------------------- *)
    Function    IsAfter( E: Elements; L: Lists; var Here: Elements): Boolean;
    Function    WhereAfter( E: Elements; L: Lists; Here: Elements): Elements;
   (* -------------------------------------------------------------------- *)
  end;

  DIR_Btrees = Class( _Directories)
   (* --- Basic Description ---------------------------------------------- *)
    BTree : Btrees;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( MinS, MaxS: Integer);
    Function    Scheme : StorageSchemes;                           Override;
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: Elements);                              Override;
    Function    Delete: Elements;                                  Override;
   (* --- Search Methods ------------------------------------------------- *)
    Function    Find( C: Codes): Elements;                         Override;
    Function    FindKey( K: Elements): Elements;                   Override;
   (* --- Report Methods ------------------------------------------------- *)
    Procedure   ReportInRange( Lf, Rt : Elements; R: Reporter);    Override;
   (* -------------------------------------------------------------------- *)
  end;

  DIR_SelfAdjustedLists= Class( _Directories)
   (* --- Basic Description ---------------------------------------------- *)
    List: Lists;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( L: Lists);
    Function    Scheme : StorageSchemes;                           Override;
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: Elements);                              Override;
    Function    Delete: Elements;                                  Override;
   (* --- Search Methods ------------------------------------------------- *)
    Function    FindKey( K: Elements): Elements;                   Override;
   (* -------------------------------------------------------------------- *)
  end;

  DIR_OpenHashTables = Class( _Directories )
   (* --- Basic Description ---------------------------------------------- *)
    HashTable : Tables;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( T: Tables);
    Function    Scheme : StorageSchemes;                           Override;
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: Elements);                              Override;
    Function    Delete: Elements;                                  Override;
   (* --- Search Methods ------------------------------------------------- *)
    Function    FindKey( K: Elements): Elements;                   Override;
   (* -------------------------------------------------------------------- *)
  end;

  DIR_OpenDynamicHashTables = Class( DIR_OpenHashTables )
   (* --- Basic Description ---------------------------------------------- *)
    LoLimit,
    HiLimit : Real;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( T: Tables);
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: Elements);                              Override;
    Function    Delete: Elements;                                  Override;
   (* --- Auxiliary Methods ---------------------------------------------- *)
    Procedure   CheckLoadFactor( C: Codes);
    Procedure   ReHashTo( N: Integer);
   (* -------------------------------------------------------------------- *)
  end;

  DIR_ClosedHashTables = Class( _Directories )
   (* --- Basic Description ---------------------------------------------- *)
    HashTable : Buckets;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( N: Integer);
    Function    Scheme : StorageSchemes;                           Override;
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: Elements);                              Override;
    Function    Delete: Elements;                                  Override;
   (* --- Search Methods ------------------------------------------------- *)
    Function    FindKey( K: Elements): Elements;                   Override;
   (* --- Auxiliary Methods ---------------------------------------------- *)
    Function    LookFor( E: Elements; L: Lists): Elements;
   (* -------------------------------------------------------------------- *)
  end;

  DIR_DoubleEntryHashTables = Class( DIR_ClosedHashTables )
   (* --- Basic Description ---------------------------------------------- *)
    BucketSize : Array of Integer;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( N: Integer);
    Destructor  Clear;                                             Override;
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: Elements);                              Override;
    Function    Delete: Elements;                                  Override;
   (* --- Search Methods ------------------------------------------------- *)
    Function    FindKey( K: Elements): Elements;                   Override;
   (* --- Auxiliary Methods ---------------------------------------------- *)
    Procedure   LookFor( E: Elements; L1,L2: Lists; Var F1, F2: Boolean);
   (* -------------------------------------------------------------------- *)
  end;

(* ----------------------------------------------------------------------- *)

  _PrioritySearchSDirectories = Class( Structures )
   (* -------------------------------------------------------------------- *)
                                                      //Abstract Structure
    HeapOrder : OrderFunctions; // Secondary Order defined on its elements
   (* -------------------------------------------------------------------- *)
  end;

  PSDIR_BinaryTrees = Class( _PrioritySearchSDirectories )
   (* -------------------------------------------------------------------- *)
    Tree      : BinaryTrees;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( T: BinaryTrees);
    Function    Scheme: StorageSchemes;                            Override;
   (* --- Store Methods -------------------------------------------------- *)
    Procedure   Insert( E: Elements);                              Override;
    Function    Delete: Elements;                                  Override;
   (* --- Search Methods ------------------------------------------------- *)
    Function    FindKey( K: Elements): Elements;                   Override;
    Function    Find( C: Codes): Elements;                         Override;
   (* --- Report Methods ------------------------------------------------- *)
    Procedure   ReportInArea( Up,Lf,Rt: Elements; R: Reporter);    Override;
   (* -------------------------------------------------------------------- *)
  end;
(* ----------------------------------------------------------------------- *)

IMPLEMENTATION
(* ----------------------------------------------------------------------- *)

(* --- STRUCTURES -------------------------------------------------------- *)
Constructor Structures.SetUp;
begin
  Size  := 0;
  Order := NumericOrder;
end;

Constructor Structures.Build( S: Structures);
begin
  Size := 0;
  With S, Scheme do begin
    Go(_First_);
    While Current<>NIL do begin
      Self.Insert( Current);
      Go(_Step_);
    end;
  end;
end;

Destructor Structures.Clear;
begin
  Size := 0;
  Scheme.Clear;
end;

Function Structures.AsText: String;
begin
  AsText := 'DATA STRUCTURE ' + ClassName + ', ' +
            'IMPLEMENTED with  ' + Scheme.ClassName + ' ';
  Case Size=0 of
    Ok: AsText := Result + '/ Void';
    No: AsText := Result + Format( '/ Size = %d elements', [Size]);
  end;
  If Scheme.MaxCapacity<_MaximumSize_ then
    AsText := Result + Format( ' (Maximum: %d)', [Scheme.MaxCapacity]);
end;

Function Structures.Scheme : StorageSchemes;
begin
  Scheme := NIL;
end;

Procedure Structures.Insert( E: Elements);
begin
end;

Function Structures.Delete: Elements;
begin
  Delete := NIL;
end;

Procedure Structures.Promote;
begin
  // Not generally available - here implemented only for "Priority Queues"
end;

Function Structures.Find( C: Codes): Elements; // Must return valid "Current"
begin
  Case C of
    _Any_  : Case Size>0 of
               Ok : Find := FindNmr(1+Random(Size));
               No : Find := NIL;
             end;
  Else
    Find := NIL;
  end;
end;

Function Structures.FindNmr( N: Integer): Elements;
begin
  FindNmr := Scheme.GoNmr(N);
end;

Function Structures.FindKey( K: Elements): Elements;
begin
  With Scheme do begin
    Go(_First_);
    While (Current<>NIL) and Not Equal( K, Current, Order ) do
      Go(_Step_);
    FindKey := Current;
  end;
end;

Procedure Structures.Split( Var E,S: Elements; Var S1,S2: Structures);
begin
  With Scheme do begin
    Go(_First_);
    While Current<>NIL do begin
      Case Order( E, Current ) of
        _Less_    : S1.Insert(Current);
        _Equal_   : S := Current;
        _Greater_ : S2.Insert(Current);
      end;
      Go(_Step_);
    end;
  end;
  Clear;
end;

Procedure Structures.Union( E: Elements; S1,S2: Structures );
begin
  SetUp;
  Insert(E);
  With S1, Scheme do begin
    Go(_First_);
    While Current<>NIL do begin Insert(Current); Go(_Step_); end;
  end;
  S1.Clear;
  With S2, Scheme do begin
    Go(_First_);
    While Current<>NIL do begin Insert(Current); Go(_Step_); end;
  end;
  S2.Clear;
end;

Procedure Structures.ReportLess( Up: Elements; R: Reporter);
begin
  With Scheme do begin
    Go(_First_);
    While Current<>NIL do begin
      If LessOrEqual( Current, Up, Order) then R.Announce(Current);
      Go(_Step_);
    end;
  end;
end;

Procedure Structures.ReportInRange( Lf,Rt: Elements; R: Reporter);
begin
  With Scheme do begin
    Go(_First_);
    While Current<>NIL do begin
      If InRange( Current, Lf,Rt, Order) then R.Announce(Current);
      Go(_Step_);
    end;
  end;
end;

Procedure Structures.ReportInArea( Up,Lf,Rt: Elements; R: Reporter);
begin
  With Scheme do begin
    Go(_First_);
    While Current<>NIL do begin
      With Current do
        If LessOrEqual( Current, Up, Order)
           and InRange( Current, Lf,Rt, Order) then R.Announce(Current);
      Go(_Step_);
    end;
  end;
end;

Function Structures.IndexOf( E: Elements ): Integer;
begin
  IndexOf := 1;
  With Scheme do begin
    Go(_First_);
    While Current<>NIL do begin
      If Less( Current, E, Order ) then IndexOf := Result+1;
      Go(_Step_);
    end;
  end;
end;
(* ----------------------------------------------------------------------- *)

(* --- CATALOGUES with TABLES -------------------------------------------- *)
Constructor CAT_Tables.SetUp( T: Tables);
begin
  Inherited SetUp;
  Case T<>NIL of
    Ok: Table := T;
    No: Table := Tables.SetUp(100);
  end;
end;

Function CAT_Tables.Scheme : StorageSchemes;
begin
  Scheme := Table;
end;

Procedure CAT_Tables.Insert( E: Elements);
begin
  With Table do
    If (E<>NIL) and (Size<MaximumNmr) then begin
      Inc(Size);
      Item[Size] := E;
    end;
end;

Function CAT_Tables.Delete: Elements;
begin
  With Table do
    Case (1<=CurrentNmr) and (CurrentNmr<=Size) of
      Ok: begin
            Delete := Current;
            Item[CurrentNmr] := Item[Size];
            Item[Size] := NIL;
            CurrentNmr := 0;
            Dec(Size);
          end;
      No: Delete := NIL;
    end;
end;
(* ----------------------------------------------------------------------- *)

(* --- CATALOGUES with DYNAMIC TABLES ------------------------------------ *)
Procedure CAT_DynamicTables.Insert( E: Elements);
Var
  i : Integer;
begin
  With Table do
    If (Size=MaximumNmr) then begin
     (* --- Color relevant element as 'USED' --------------------- *)
      For i := 1 to Size do
        If Item[i]<>NIL then Item[i].Color := _UsedColor_;
     (* ---------------------------------------------------------- *)
      Resize(2*MaximumNmr);
    end;
  Inherited Insert(E);
end;

Function CAT_DynamicTables.Delete: Elements;
Var
  i : Integer;
begin
  Delete := Inherited Delete;
  With Table do
    If (Size<=(MaximumNmr div 4)) then begin
     (* --- Color relevant element as 'USED' --------------------- *)
      For i := 1 to Size do
        If Item[i]<>NIL then Item[i].Color := _UsedColor_;
     (* ---------------------------------------------------------- *)
      Resize(MaximumNmr div 2);
    end;
end;
(* ----------------------------------------------------------------------- *)

(* --- CATALOGUES with LISTS --------------------------------------------- *)
Constructor CAT_Lists.SetUp( L: Lists);
begin
  Inherited SetUp;
  Case L<>NIL of
    Ok: List := L;
    No: List := Lists.SetUp;
  end;
end;

Function CAT_Lists.Scheme : StorageSchemes;
begin
  Scheme := List;
end;

Procedure CAT_Lists.Insert( E: Elements);
begin
  If E<>NIL then begin
    Inc(Size);
    With List do
      Link( LinkFor(E), _After_, Tail);
  end;
end;

Function CAT_Lists.Delete;
begin
  With List do begin
    Delete := Current;
    If Result<>NIL then begin
      UnLink( CurrentLink );
      CurrentLink.Clear; // garbage colection
      Dec(Size);
    end;
  end;
end;
(* ----------------------------------------------------------------------- *)

(* --- STACKS ------------------------------------------------------------ *)
Procedure _Stacks.Push( E: Elements);
begin
  Insert(E);
end;

Function _Stacks.Pop: Elements;
begin
  Pop := Find(_Top_);
  Delete;
end;

Function _Stacks.Top: Elements;
begin
  Top := Find(_Top_);
end;
(* ----------------------------------------------------------------------- *)

(* --- STACKS with TABLES ------------------------------------------------ *)
Constructor STACK_Tables.SetUp( T: Tables);
begin
  Inherited SetUp;
  Case T<>NIL of
    Ok: Table := T;
    No: Table := Tables.SetUp(100);
  end;
end;

Function STACK_Tables.Scheme : StorageSchemes;
begin
  Scheme := Table;
end;

Procedure STACK_Tables.Insert( E: Elements);
begin
  With Table do
    If (E<>NIL) and (Size<MaximumNmr) then begin
      Inc(Size);
      Item[Size] := E;
    end;
end;

Function STACK_Tables.Delete: Elements;
Var
  k    : Integer;
begin
  With Table do
    Case (1<=CurrentNmr) and (CurrentNmr<=Size) of
      Ok: begin
            Delete := Current;
            For k := CurrentNmr+1 to Size do Item[k-1] := Item[k];
            Item[Size] := NIL;
            CurrentNmr := 0;
            Dec(Size);
          end;
      No: Delete := NIL;
    end;
end;

Function STACK_Tables.Find( C: Codes): Elements;
begin
  Case C of
    _Top_ : Table.CurrentNmr := Size;
  Else
    Inherited Find(C);
  end;
  Find := Table.Current;
end;

Procedure STACK_Tables.DrawLinks;
begin
  If Size>0 then
    With Table do begin
    //DrawAnnotation( W, ' ', Location(1), 0,-4);
      DrawAnnotation( W, 'Oldest element', Location(1), 0,-4);
    //DrawAnnotation( W, '  (Top)', Location(Size), +1,-2);
      DrawAnnotation( W, 'Newest element (Top)', Location(Size), +1,-2);
    end;
end;
(* ----------------------------------------------------------------------- *)

(* --- STACKS with LISTS ------------------------------------------------- *)
Constructor STACK_Lists.SetUp( L: Lists);
begin
  inherited SetUp;
  Case L<>NIL of
    Ok: List := L;
    No: List := Lists.SetUp;
  end;
end;

Function STACK_Lists.Scheme : StorageSchemes;
begin
  Scheme := List;
end;

Procedure STACK_Lists.Insert( E: Elements);
begin
  If E<>NIL then begin
    Inc(Size);
    List.Link( List.LinkFor(E), _After_, List.Tail);
  end;
end;

Function STACK_Lists.Delete: Elements;
begin
  With List do begin
    Delete := Current;
    If Result<>NIL then begin
      UnLink( CurrentLink );
      CurrentLink.Clear; // garbage colection
      Dec(Size);
    end;
  end;
end;

Function STACK_Lists.Find(C: Codes): Elements;
begin
  Case C of
    _Top_ : List.Go(_Last_);
  Else
    Inherited Find(C);
  end;
  Find := List.Current;
end;

Procedure STACK_Lists.DrawLinks;
begin
  With List do begin
  //DrawAnnotation( W, ' ', Head, +1,-5);
    DrawAnnotation( W, 'Oldest element', Head, +1,-5);
  //DrawAnnotation( W, '  (Top)', Tail, +2,-3);
    DrawAnnotation( W, 'Newest element (Top)', Tail, +2,-3);
  end;
end;
(* ----------------------------------------------------------------------- *)

(* --- QUEUES ------------------------------------------------------------ *)
Procedure _QUEUES.EnQueue(E: Elements);
begin
  Insert(E);
end;

Function _QUEUES.DeQueue: Elements;
begin
  DeQueue := Find(_Top_);
  Delete;
end;

Function _QUEUES.Top: Elements;
begin
  Top := Find(_Top_);
end;
(* ----------------------------------------------------------------------- *)

(* --- QUEUES with LISTS ------------------------------------------------- *)
Constructor QUEUE_Lists.SetUp( L: Lists);
begin
  inherited SetUp;
  Case L<>NIL of
    Ok: List := L;
    No: List := Lists.SetUp;
  end;
end;

Function QUEUE_Lists.Scheme : StorageSchemes;
begin
  Scheme := List;
end;

Procedure QUEUE_Lists.Insert( E: Elements);
begin
  If E<>NIL then begin
    Inc(Size);
    List.Link( List.LinkFor(E), _Before_, NIL);
  end;
end;

Function QUEUE_Lists.Delete: Elements;
begin
  With List do begin
    Delete := Current;
    If Result<>NIL then begin
      UnLink( CurrentLink );
      CurrentLink.Clear; // garbage colection
      Dec(Size);
    end;
  end;
end;

Function QUEUE_Lists.Find( C: Codes): Elements;
begin
  Case C of
    _Top_ : List.Go(_First_);
  Else
    Inherited Find(C);
  end;
  Find := List.Current;
end;

Procedure QUEUE_Lists.DrawLinks;
begin
  With List do begin
  //DrawAnnotation( W, '  (Top)', Head, +1,-5);
    DrawAnnotation( W, 'Oldest element (Top)', Head, +1,-5);
  //DrawAnnotation( W, ' ', Tail, +2,-3);
    DrawAnnotation( W, 'Newest element', Tail, +2,-3);
  end;
end;
(* ----------------------------------------------------------------------- *)

(* --- QUEUES with TABLES ------------------------------------------------ *)
Constructor QUEUE_CircularTables.SetUp( T: CircularTables);
begin
  Inherited SetUp;
  Case T<>NIL of
    Ok: CircularTable := T;
    No: CircularTable := CircularTables.SetUp(50);
  end;
end;

Function QUEUE_CircularTables.Scheme : StorageSchemes;
begin
  Scheme := CircularTable;
end;

Procedure QUEUE_CircularTables.Insert( E: Elements);
begin
  With CircularTable do
    If (E<>NIL) and (Size<MaximumNmr) then begin
      Inc(Size);
      Item[Size] := E;
    end;
end;

Function QUEUE_CircularTables.Delete: Elements;
Var
  k : Integer;
begin
  With CircularTable do begin
    Case (1<=CurrentNmr) and (CurrentNmr<=Size) of
      Ok: begin
            Delete := Current;
            For k := CurrentNmr-1 DownTo 1 do Item[k+1] := Item[k];
            Item[1] := NIL;
            Dec(Size);
            CurrentNmr := 0;
            First := (First Mod MaximumNmr) + 1;
          end;
      No: Delete := NIL;
    end;
  end;
end;

Function QUEUE_CircularTables.Find( C: Codes): Elements;
begin
  Case C of
    _Top_ : CircularTable.CurrentNmr := 1;
  Else
    Inherited Find(C);
  end;
  Find := CircularTable.Current;
end;

Procedure QUEUE_CircularTables.DrawLinks;
begin
  If Size>0 then
    With CircularTable do begin
    //DrawAnnotation( W, '  (Top)', Location(1), -2,-2);
      DrawAnnotation( W, 'Oldest element (Top)', Location(1), -2,-2);
    //DrawAnnotation( W, ' ', Location(Size), -2,-2);
      DrawAnnotation( W, 'Newest element', Location(Size), -2,-2);
    end;
end;
(* ----------------------------------------------------------------------- *)

(* ----------------------------------------------------------------------- *)
Procedure _PriorityQueues.EnQueue(E: Elements);
begin
  Insert(E);
end;

Function _PriorityQueues.DeQueue: Elements;
begin
  DeQueue := Top;
  Delete;
end;

Procedure _PriorityQueues.Promote;
begin
  If Current<>NIL then Delete.Clear;
  Insert(P);
end;

Function _PriorityQueues.Top: Elements;
begin
  Top := Find(_Top_);
end;
(* ----------------------------------------------------------------------- *)

(* --- PQ with HEAPS in TABLES ------------------------------------------- *)
Constructor PQ_HeapTables.SetUp( T: Tables);
begin
  Inherited SetUp;
  Case T<>NIL of
    Ok: Table := T;
    No: Table := Tables.SetUp(250); // Just for example
  end;
end;

Function PQ_HeapTables.Scheme : StorageSchemes;
begin
  Scheme := Table;
end;

Procedure PQ_HeapTables.Insert( E: Elements);
begin
  If (E<>NIL) then
    If (Size<Table.MaximumNmr) then begin
      Inc(Size);
      Table.Item[Size] := E;
      BubbleUp(Size);
    end;
end;

Function PQ_HeapTables.Delete: Elements;
Var
  This : Integer;
begin
  Delete := Table.Current;
  If Result<>NIL then begin
    This := Table.CurrentNmr;
    If (1<=This) and (This<=Size) then begin
      BubbleDown(This);
      If (This<>Size) then begin    // Useless If Not Different
        Table.SwapItems( This, Size);
        BubbleUp(This);
      end;
      Table.Item[Size] := NIL;
      Dec(Size);
    end;
  end;
end;

Procedure PQ_HeapTables.Promote;
begin
  With Table do
    If (1<=CurrentNmr) and (CurrentNmr<=Size) then begin
      Location(CurrentNmr)[_NewInfo_] := P;
      BubbleUp( CurrentNmr);
    end;
end;

function PQ_HeapTables.Find(C: Codes): Elements;
begin
  Case C of
    _Top_ : With Table do CurrentNmr := 1;
  Else
    Inherited Find(C);
  end;
  Find := Table.Current;
end;

Procedure PQ_HeapTables.ReportLess( Up: Elements; R: Reporter);
//
  Procedure Report( This: Integer);
  begin
    If (This<=Size) and LessOrEqual( Table.Item[This], Up, Order) then begin
      R.Announce( Table.Item[This]);
      Report(2*This);
      Report(2*This+1);
    end;
  end;
//
begin
  Report(1);
end;

Procedure PQ_HeapTables.BubbleUp( P: Integer);
Var
  Here : Integer;
  More : Boolean;
begin
  With Table do
    Repeat
      Here := P Div 2;    // = the _Parent_
      More := (Here>0) and Less( Item[P], Item[Here], Order);
      If More then begin
        SwapItems( P, Here);
        P := Here;
      end;
    Until Not More;
end;

Procedure PQ_HeapTables.BubbleDown( Var P : Integer);
Var
  L,R  : Elements;
  C    : Codes;
begin
  If (1<=P) and (P<=Size) then
    With Table do
      Repeat
        If 2*P   <= Size then L := Item[2*P]   else L := NIL;    // _Lf_
        If 2*P+1 <= Size then R := Item[2*P+1] else R := NIL;    // _Rt_
        C := Minimum( L, R, Order);
        Case C Of
          _1st_: begin SwapItems( P, 2*P);   P := 2*P;   end;
          _2nd_: begin SwapItems( P, 2*P+1); P := 2*P+1; end;
        end;
      Until (C=_None_);
end;

Procedure PQ_HeapTables.DrawLinks( W: TCanvas; k,l: Integer);
begin
  If Size>0 then
  //DrawAnnotation( W, '  (Top)', Table.Location(1), +1,-2);
    DrawAnnotation( W, 'Minimum element (Top)', Table.Location(1), +1,-2);
end;
(* ----------------------------------------------------------------------- *)

(* --- PQ with BINARY TREES ---------------------------------------------- *)
Constructor PQ_BinaryHeaps.SetUp( T: BinaryTrees);
begin
  Inherited SetUp;
  Case T<>NIL of
    Ok: Tree := T;
    No: Tree := BinaryTrees.SetUp;
  end;
end;

Function PQ_BinaryHeaps.Scheme : StorageSchemes;
begin
  Scheme := Tree;
end;

Function PQ_BinaryHeaps.Find( C: Codes): Elements;
begin
  Case C of
    _Top_ : Tree.Go(_Root_);
  Else
    Inherited Find(C);
  end;
  Find := Tree.Current;
end;

Procedure PQ_BinaryHeaps.BubbleUp( Var L: Elements);
Var
  Here   : Elements;
  More : Boolean;
begin
  Repeat
    Here := L[_Parent_];
    More := (Here<>NIL) and Less( L[_Info_], Here[_Info_], Order);
    If More then begin
      SwapParts( L, _Info_, Here, _Info_);
      L := Here;
    end;
  Until Not More;
end;

Procedure PQ_BinaryHeaps.BubbleDown( Var L: Elements);
Var
  Here : Elements;
  A,B  : Elements;
begin
  If L<>NIL then
    Repeat
      A := L[_Lf_]; If A<>NIL then A := A[_Info_];
      B := L[_Rt_]; If B<>NIL then B := B[_Info_];
      Case Minimum(A,B, Order) Of
        _1st_: Here := L[_Lf_];
        _2nd_: Here := L[_Rt_];
      Else
        Here := NIL;
      end;
      If (Here<>NIL) then begin
        SwapParts( L, _Info_, Here, _Info_);
        L := Here;
      end;
    Until (Here=NIL);
end;

Procedure PQ_BinaryHeaps.Insert( E: Elements);
Var
  This,
  Here : Elements;
begin
  If E<>NIL then begin
    This := Tree.LinkFor(E);
    Inc(Size);  // Now! Correct "Sizing" Is Important Here
    With Tree do
      Case Root=NIL of
        Ok: Root := This;
        No: begin
              Here := GoBinary( Size Div 2);
              Link( This, Child[Size Mod 2], Here );
              BubbleUp(This);
            end;
      end;
  end;
end;

Function PQ_BinaryHeaps.Delete: Elements;
Var
  This,
  Here : Elements;
begin
  Delete := Tree.Current;
  This   := Tree.CurrentLink;
  If Result<>NIL then begin
    BubbleDown(This);
    Here := Tree.GoBinary(Size); // Now and not before!
    If (Here<>This) then begin
      SwapParts( Here, _Info_, This, _Info_);
      BubbleUp(This);
    end;
    Tree.UnLink(Here);
    Here.Clear; // garbage collection
    Dec(Size);
  end;
end;

Procedure PQ_BinaryHeaps.Promote;
begin
  With Tree do
    If CurrentLink<>NIL then begin
      CurrentLink[_NewInfo_] := P;
      BubbleUp( Tree.CurrentLink);
    end;
end;

Procedure PQ_BinaryHeaps.ReportLess( Up: Elements; R: Reporter);
//
  Procedure Report( This: Elements);
  begin
    If (This<>NIL) and LessOrEqual( This[_Info_], Up, Order) then begin
      R.Announce( This[_Info_]);
      Report( This[_Lf_]);
      Report( This[_Rt_]);
    end;
  end;
//
begin
  Report(Tree.Root);
end;

Procedure PQ_BinaryHeaps.DrawLinks;
begin
//DrawAnnotation( W, '  (Top)', Tree.Root, +3,-4);
  DrawAnnotation( W, 'Minimum element (Top)', Tree.Root, +3,-4);
end;

(* ----------------------------------------------------------------------- *)

Constructor PQ_BinomialHeaps.SetUp;
begin
  Inherited SetUp;
  Forest     := Forests.SetUp;
  TopLink    := NIL;
  AlwaysPack := True;
end;

Function PQ_BinomialHeaps.Scheme : StorageSchemes;
begin
  Scheme := Forest;
end;

Function PQ_BinomialHeaps.Find( C: Codes): Elements;  
Var
  Top,
  Here  : Elements;
begin
  Case C of
    _Top_ : Case TopLink<>NIL of
              Ok: With Forest do begin
                    ListOfTrees.CurrentLink := TopLink;
                    With CurrentTree do CurrentLink := Root;
                  end;
              No: With Forest do begin
                    With ListOfTrees do begin
                      Go(_First_);
                      TopLink := CurrentLink;
                      Go(_Step_);
                    end;
                    While (CurrentTree<>NIL) do begin
                      Here := CurrentTree.Root;
                      Top  := RootOf(TopLink);
                      If Less( Here[_Info_], Top[_Info_], Order )then
                        TopLink := ListOfTrees.CurrentLink;
                      ListOfTrees.Go(_Step_);
                    end;
                  end;
            end;
  Else
    Inherited Find(C);
  end;
  Find := Forest.Current;
end;

Procedure PQ_BinomialHeaps.Insert( E: Elements);
Var
  This : Elements;
begin
  If E<>NIL then begin
    With Forest do begin
      This := LinkFor(E);
      ListOfTrees.Link( This, _After_, ListOfTrees.Tail);
      If (TopLink=NIL) or (Order( E, RootOf(TopLink)[_Info_]) = _Less_) then
        TopLink := This;
    end;
    Inc(Size);
    If AlwaysPack then PackRootLevel;
  end;
end;

Function PQ_BinomialHeaps.Delete: Elements;
Var
  This,
  Here,
  Next : Elements;
  Tree : RootedTrees;
begin
  With Forest do begin
    Tree := CurrentTree;
    Case Tree=NIL of
      Ok: Delete := NIL;
      No: begin
            Delete := Tree.Current;
            This   := Tree.CurrentLink;
            If Result<>NIL then begin
              MoveToRoot(This);
              With Forest.ListOfTrees do begin
                UnLink(CurrentLink);
                CurrentLink.Clear; // garbage collection
                Here := Tree.Root[_FirstChild_];
                While Here<>NIL do begin
                  Next := Here[_NextSibling_];
                  Here[_Parent_] := NIL;
                  Here[_NextSibling_] := NIL;
                  Here[_PrevSibling_] := NIL;
                  // Wrap children as roots of a new trees
                  Tree := AnyDegTrees.SetUp;
                  Tree.Root := Here;
                  Link( LinkFor(Tree), _After_, Tail);
                  //
                  Here := Next;
                end;
              end;
              PackRootLevel;
              Dec(Size);
            end;
          end;
    end;
  end;
end;

procedure PQ_BinomialHeaps.Promote;
Var
  Top,
  Here : Elements;
begin
  With Forest do
    If CurrentTree<>NIL then begin
      Here := CurrentTree.CurrentLink;
      If Here<>NIL then Here[_NewInfo_] := P;
      // Take Care of Min Element...
      Top := Forest.RootOf(TopLink);
      If Less( Here[_Info_], Top[_Info_], Order) then
        SwapParts( Here, _Info_, Top, _Info_);
      //
      BubbleUp(Here);
    end;
end;

procedure PQ_BinomialHeaps.ReportLess(Up: Elements; R: Reporter);
//
  Procedure Report( This: Elements);
  Var
    Child: Elements;
  begin
    If (This<>NIL) and LessOrEqual( This[_Info_], Up, Order) then begin
      R.Announce( This[_Info_]);
      Child := This[_FirstChild_];
      While Child<>NIL do begin
        Report(Child);
        Child := Child[_NextSibling_];
      end;
    end;
  end;
//
begin
  PackRootLevel;
  With Forest do begin
    ListOfTrees.Go(_First_);
    While (CurrentTree<>NIL) do begin
      Report(CurrentTree.Root);
      ListOfTrees.Go(_Step_);
    end;
  end;
end;

Procedure PQ_BinomialHeaps.MoveToRoot( Var L: Elements);
Var
  Here : Elements;
begin
  Repeat
    Here := L[_Parent_];
    If Here<>NIL then begin
      SwapParts( L, _Info_, Here, _Info_);
      L := Here;
    end;
  Until (Here=NIL);
end;

Procedure PQ_BinomialHeaps.BubbleUp( Var L: Elements);
Var
  Here : Elements;
  More : Boolean;
begin
  Repeat
    Here := L[_Parent_];
    More := (Here<>NIL) and (Order( L[_Info_], Here[_Info_]) = _Less_);
    If More then begin
      SwapParts( L, _Info_, Here, _Info_);
      L := Here;
    end;
  Until Not More;
end;

Procedure PQ_BinomialHeaps.HangFrom( Var TreeA, TreeB: AnyDegTrees);
begin
  Case Minimum( TreeA.Root[_Info_], TreeB.Root[_Info_], Order ) of
    _1st_ : begin
              TreeA.Link( TreeB.Root, _Child_, TreeA.Root);
              With TreeA.Root do Rank := Rank+1;
            end;
    _2nd_ : begin
              TreeB.Link( TreeA.Root, _Child_, TreeB.Root);
              With TreeB.Root do Rank := Rank+1;
              TreeA := TreeB;
            end;
  end;
end;

Procedure PQ_BinomialHeaps.PackRootLevel;
Var
  R       : Integer;
  HasRank : Array[0..64] of AnyDegTrees; // Enough for 2^64 elements
  Tree    : AnyDegTrees;
//
  Procedure CLearListOfTrees;
  Var
    Here, This : Elements;
  begin
    Here := Forest.ListOfTrees.Head;
    While Here<>NIL do begin
      This := Here;
      Here := Here[_Next_];
      This.Clear;
    end;
  end;
//
begin
  For R := 0 to 64 do HasRank[R] := NIL;
  With Forest do begin
    ListOfTrees.Go(_First_);
    While (CurrentTree<>NIL) do begin
      Tree := CurrentTree;
      R    := Tree.Root.Rank;
      While HasRank[R]<>NIL do begin
        HangFrom( Tree, HasRank[R] );
        HasRank[R] := NIL;
        R := Tree.Root.Rank;
      end;
      HasRank[R] := Tree;
      ListOfTrees.Go(_Step_);
    end;
   (* --- Clear and rebuild ListOfTrees ---------------------------------- *)
    ClearListOfTrees;
    With ListOfTrees do begin
      Head := NIL;
      Tail := NIL;
    end;
    For R := 0 to 64 do
      If HasRank[R]<>NIL then
        With ListOfTrees do
          Link( LinkFor(HasRank[R]), _After_, Tail);
  end;
  //
  TopLink := NIL;
  Find(_Top_);
end;

Procedure PQ_BinomialHeaps.DrawLinks;
begin
  With Forest do
  //DrawAnnotation( W, '  (Top)', RootOf(TopLink), +1,-2);
    DrawAnnotation( W, 'Minimum element (Top)', RootOf(TopLink), +1,-2);
end;

(* ----------------------------------------------------------------------- *)

Constructor PQ_AmortizedBinomialHeaps.SetUp;
begin
  Inherited SetUp;
  AlwaysPack := False;
end;

(* ----------------------------------------------------------------------- *)

Procedure PQ_FibonacciHeaps.Promote;
Var
  Up,
  Top,
  Cut  : Elements;
  More : Boolean;
//
  Procedure LinkToRootLevel(E: Elements);
  var
    This : Elements;
    Tree : RootedTrees;
  begin
    With Forest do begin
      // Wrap it as root of a new tree
      CurrentTree.UnLink(E);
      Tree := AnyDegTrees.SetUp;
      Tree.Root := E;
      This := ListOfTrees.LinkFor(Tree);
      ListOfTrees.Link( This, _Before_, NIL);
      If (TopLink=NIL)or (Order( E[_Info_], RootOf(TopLink)[_Info_]) = _Less_)
      then
        TopLink := This;
    end;
  end;
//
begin
  Cut := Forest.CurrentTree.CurrentLink;
  If (Cut<>NIL) then begin
    Cut[_NewInfo_] := P;
    // Take care of min element...
    Top := Forest.RootOf(TopLink);
    If Less( Cut[_Info_], Top[_Info_], Order) then
      SwapParts( Cut, _Info_, Top, _Info_);
    //
    Up := Cut[_Parent_];
    If (Up<>NIL) and Less( Cut[_Info_], Up[_Info_], Order) then
      Repeat
        LinkToRootLevel(Cut);
        Up.Rank := Up.Rank-1;
        Case Up.Marked of
          Ok: begin
                Cut     := Up;
                Up      := Cut[_Parent_];
                More    := (Up<>NIL);
              end;
          No: begin
                Up.Marked := Ok;
                More    := No;
              end;
        end;
      Until Not More;
  end;
end;

Procedure PQ_FibonacciHeaps.HangFrom( Var TreeA, TreeB: AnyDegTrees);
begin
  Inherited HangFrom( TreeA, TreeB);
  (* --- Now and not earlier for visualization purposes ------------------ *)
  TreeA.Root.Marked := No;
end;

(* ----------------------------------------------------------------------- *)

Procedure PQ_LeftistHeaps.Insert( E: Elements);
begin
  If E<>NIL then begin
    E := Tree.LinkFor(E);
    E.Rank := 1;
    Tree.Root := Meld( Tree.Root, E);
    Inc(Size);
  end;
end;

Function PQ_LeftistHeaps.Delete: Elements;
Var
  Here,
  This,
  Child  : Elements;
  AsWhat : Codes;
begin
  Delete := Tree.Current;
  This   := Tree.CurrentLink;
  If Result<>NIL then begin
    AsWhat := This.IsWhat;
    Here   := This[_Parent_];
    Tree.UnLink(This);
    Child  := Meld( This[_Lf_], This[_Rt_]);
    Tree.Link( Child, AsWhat, Here);
    While Here<>NIL do begin
      HandleRanks( Here);
      Here := Here[_Parent_];
    end;
    This.Clear; // garbage collection
    Dec(Size);
  end;
end;

Procedure PQ_LeftistHeaps.HandleRanks( E: Elements);
begin
  // Check ranks and swap accordingly...
  If E<>NIL then begin
    If (E.RankOf(_Lf_) < E.RankOf(_Rt_)) then SwapParts( E, _Lf_, E, _Rt_);
    E.Rank := E.RankOf(_Rt_) + 1;
  end;
end;

Function PQ_LeftistHeaps.Meld( L, R: Elements): Elements;
Var
 Min,
 Max : Elements;
begin
  Case NotNIL( L, R) of
    _Both_: begin
              Case Less( L[_Info_], R[_Info_], Order) of
                Ok: begin Min := L; Max := R; end;
                No: begin Min := R; Max := L; end;
              end;
              //
              Min[_Rt_] := Meld( Min[_Rt_], Max);
              Min[_Rt_][_Parent_] := Min;
              //
              HandleRanks(Min);
              Meld := Min;
            end;
    _1st_ : Meld := L;
    _2nd_ : Meld := R;
    _None_: Meld := NIL;
  end;
end;

(* ----------------------------------------------------------------------- *)

Procedure PQ_SkewHeaps.HandleRanks( E: Elements);
begin
  // No ranks-no checking: just swap always...
  If E<>NIL then
    SwapParts( E, _Lf_, E, _Rt_);
end;

(* ----------------------------------------------------------------------- *)

Procedure PQ_PairingHeaps.Insert( E: Elements);
begin
  If E<>NIL then
    With Tree do begin
      Root := Meld( Root, LinkFor(E));
      Inc(Size);
    end;
end;

Function PQ_PairingHeaps.Delete: Elements;
Var
  Here,
  This,
  P,L,R  : Elements;
  AsWhat : Codes;
begin
  Delete := Tree.Current;
  This   := Tree.CurrentLink;
  If Result<>NIL then begin
    AsWhat := This.IsWhat;
    P      := This[_Parent_];
    L      := This[_Lf_];
    R      := This[_Rt_];
    With Tree do begin
      UnLink(L);
      UnLink(R);
      UnLink(This);
      Case NotNIL(L,R) of
        _Both_,
        _2nd_ : begin
                  Here := R[_AllLf_];
                  Repeat HeapSplayUp(Here) Until Here[_Parent_]=NIL;
                  Here[_Lf_] := L;
                  If L<>NIL then L[_Parent_] := Here;
                end;
        _1st_ : begin
                  Here := L[_AllLf_];
                  Repeat HeapSplayUp(Here) Until Here[_Parent_]=NIL;
                  Here[_Lf_] := R;
                  If R<>NIL then R[_Parent_] := Here;
                end;
        _None_: Here := NIL;
      end;
      //
      Case P=NIL of
        Ok: Root := Here;
        No: begin
              P[AsWhat] := Here;
              If Here<>NIL then Here[_Parent_] := P;
            end;
      end;
      //
    end;
    This.Clear; // garbage collection
    Dec(Size);
  end;
end;

Procedure PQ_PairingHeaps.Promote( P: Elements);
Var
  Up,
  Lf,
  Here: Elements;
  What: Codes;
begin
  With Tree do
    If CurrentLink<>NIL then begin
      Here := CurrentLink;
      Here[_NewInfo_] := P;
      What := Here.IsWhat;
      Up := Here[_Parent_];
      Lf := Here[_Lf_];
      Case What of
        _Lf_ : If Less( Here[_Info_], Root[_Info_], Order) then begin
                 UnLink(Lf);
                 UnLink(Here);
                 Link( Lf, _Lf_, Up);
                 Root := Meld( Root, Here);
               end;
        _Rt_ : If Less( Here[_Info_], Up[_Info_], Order) then begin
                 UnLink(Lf);
                 UnLink(Here);
                 Link( Lf, _Rt_, Up);
                 Root := Meld( Root, Here);
               end;
        // '_Root_' : all OK
      end;
    end;
end;

Procedure PQ_PairingHeaps.ReportLess( Up: Elements; R: Reporter);
//
 (* --- Heap order is partially respected, so... ------------------------- *)
  Procedure Report( This: Elements);
  begin
    If (This<>NIL) then begin
      Report(This[_Lf_]);
      If LessOrEqual( This[_Info_], Up, Order) then begin
        R.Announce( This[_Info_]);
        Report(This[_Rt_]);
      end;
    end;
  end;
//
begin
  Report(Tree.Root);
end;

Procedure PQ_PairingHeaps.HeapSplayUp( E: Elements);
//
  Procedure HeapRotateUp( E: Elements);
  Var
    R: Elements;
  begin
    If E<>NIL then begin
      (E As BinaryTreeLinks).RotateUp;
      R := E[_Rt_];
      If (R<>NIL) then
        Case Less( E[_Info_], R[_Info_], Order) of
          Ok: Done;
          No: begin
                SwapParts( E, _Info_, R, _Info_);
                SwapParts( R, _Lf_,   R, _Rt_);
              end;
        end;
    end;
  end;
//
begin
  HeapRotateUp( E[_Parent_] );
  HeapRotateUp( E );
end;

Function PQ_PairingHeaps.Meld( L,R: Elements): Elements;
Var
 Min,
 Max,
 Child : Elements;
begin
  Case NotNIL( L, R) of
    _Both_: begin
              Case Less( L[_Info_], R[_Info_], Order) of
                Ok: begin Min := L; Max := R; end;
                No: begin Min := R; Max := L; end;
              end;
              //
              Child := Min[_Rt_];
              If Child<>NIL then Child[_Parent_] := Max;
              //
              Max[_Lf_] := Child;
              Max[_Parent_] := Min;
              //
              Min[_Lf_] := NIL;
              Min[_Rt_] := Max;
              Min[_Parent_] := NIL;
              //
              Meld := Min;
            end;
    _1st_ : Meld := L;
    _2nd_ : Meld := R;
    _None_: Meld := NIL;
  end;
end;
(* ----------------------------------------------------------------------- *)

(* ----------------------------------------------------------------------- *)
Constructor DIR_SortedTables.SetUp( T: Tables);
begin
  Inherited SetUp;
  Case T<>NIL of
    Ok: Table := T;
    No: Table := Tables.SetUp(128);
  end;
end;

Function DIR_SortedTables.Scheme: StorageSchemes;
begin
  Scheme := Table;
end;

Procedure DIR_SortedTables.Insert( E: Elements);
Var
  k,Here : Integer;
  Found  : Boolean;
begin
  With Table do
    If (E<>NIL) and (Size<MaximumNmr) then begin
      Here := BinarySearch( E, Found);
     (* --- This implementation does not afford double keys... ----------- *)
      If Not Found then begin
        For k := Size DownTo Here do Item[k+1] := Item[k];
        Item[Here] := E;
        Inc(Size);
      end;
    end;
end;

Function DIR_SortedTables.Delete: Elements;
Var
  k,Here: Integer;
begin
  With Table do begin
    Delete := Current;
    Here   := CurrentNmr;
    If (1<=Here) and (Here<=Size) then begin
      For k := Here+1 To Size do Item[k-1] := Item[k];
      Item[Size] := NIL;
      Dec(Size);
    end;
  end;
end;

Function DIR_SortedTables.Find( C: Codes ): Elements;
begin
  Case C of
    _Min_ : Table.CurrentNmr := 1;
    _Max_ : Table.CurrentNmr := Size;
  Else
    Inherited Find(C);
  end;
  Find := Table.Current;
end;

Function DIR_SortedTables.FindKey( K: Elements): Elements;
Var
  Here : Integer;
  Found: Boolean;
begin
  Here := BinarySearch( K, Found);
  Case Found of
    Ok: begin
          Table.CurrentNmr := Here;
          FindKey := Table.Current;
        end;
    No: begin
          Table.CurrentNmr := 0;
          FindKey := NIL;
        end;
  end;
end;

Procedure DIR_SortedTables.ReportLess( Up: Elements; R: Reporter);
Var
  i,B : Integer;
  Found : Boolean;
begin
  B := BinarySearch( Up, Found); If Not Found then Dec(B);
  For i := 1 to B do
    R. Announce( Table.Item[i] );
end;

Procedure DIR_SortedTables.ReportInRange( Lf,Rt: Elements; R: Reporter);
Var
  i,A,B : Integer;
  Found : Boolean;
begin
  A := BinarySearch( Lf, Found);
  B := BinarySearch( Rt, Found); If Not Found then Dec(B);
  For i := A to B do
    R.Announce( Table.Item[i] );
end;

Function DIR_SortedTables.BinarySearch( K: Elements;
                                        Var Found: Boolean): Integer;
// Returns the position of the first element larger than or equal to K
Var
  L,R,
  Middle: Integer;
begin
  L := 1;
  R := Size;
  Found := No;
  BinarySearch := 0;
  Repeat
    Case (L<=R) of
      Ok : begin
             Middle := (L+R) Div 2;
             Case Order( K, Table.Item[Middle] ) of
               _Less_    : R := Middle-1;
               _Equal_   : begin BinarySearch := Middle; Found := Ok; end;
               _Greater_ : L := Middle+1;
             end;
           end;
      No : begin
             BinarySearch := L;
             Found := No;
           end;
    end;
  Until (Result<>0) Or Found;
end;

Procedure DIR_SortedTables.DrawLinks( W: TCanvas; k,l: Integer);
begin
  If Size>0 then begin
  //DrawAnnotation(W, '  (Min)', Table.Location(1), 0,-4);
    DrawAnnotation(W, 'Minimum element (Min)', Table.Location(1), 0,-4);
  //DrawAnnotation(W, '  (Max)', Table.Location(Size), +1,-2);
    DrawAnnotation(W, 'Maximum element (Max)', Table.Location(Size), +1,-2);
  end;
end;

(* ----------------------------------------------------------------------- *)

Constructor DIR_OpenHashTables.SetUp( T: Tables);
begin
  Inherited SetUp;
  Case T<>NIL of
    Ok: HashTable := T;
    No: HashTable := Tables.SetUp(57);
  end;
end;

Function DIR_OpenHashTables.Scheme : StorageSchemes;
begin
  Scheme := HashTable;
end;

Procedure DIR_OpenHashTables.Insert( E: Elements);
Var
  Hash,
  TryCount   : Integer;
  EmptyFound : Boolean;
begin
  If E<>NIL then
    With HashTable do begin
      TryCount := 0;
      Repeat
        Hash       := HashLocation((E As KeyData).Key, MaximumNmr, TryCount);
        EmptyFound := (Item[Hash]=NIL);
        If Not EmptyFound then Location(Hash).Marked := Ok;
        Inc(TryCount);
      Until EmptyFound or (TryCount=HashTable.MaximumNmr);
      If EmptyFound then begin
        Item[Hash] := E;
        Inc(Size);
      end;
    end;
end;

Function DIR_OpenHashTables.FindKey( K: Elements): Elements;
Var
  Hash,
  TryCount : Integer;
  Found    : Boolean;
begin
  With HashTable do begin
    CurrentNmr := 0;
    If K<>NIL then begin
      TryCount := 0;
      Repeat
        Hash  := HashLocation( (K As KeyData).Key, MaximumNmr, TryCount);
        Found := (Item[Hash]<>NIL) and Equal(K, Item[Hash], Order);
        Inc(TryCount);
      Until Found
            or (TryCount=HashTable.MaximumNmr)
            or Not Location(Hash).Marked;
      If Found then
        CurrentNmr := Hash;
    end;
    FindKey := HashTable.Current;
  end;
end;

Function DIR_OpenHashTables.Delete: Elements;
begin
  With HashTable do
    Case (1<=CurrentNmr) and (CurrentNmr<=MaximumNmr) of
      Ok: begin
            Delete := Current;
            Item[CurrentNmr] := NIL;
            Location(CurrentNmr).Marked  := Ok;
            Dec(Size);
          end;
      No: Delete := NIL;
    end;
end;

(* ----------------------------------------------------------------------- *)

Constructor DIR_OpenDynamicHashTables.SetUp( T: Tables);
begin
  Inherited SetUp(T);
 (* --- Just for example --- *)
  LoLimit := 0.30;
  HiLimit := 0.90;
 (* ------------------------ *)
end;

Procedure DIR_OpenDynamicHashTables.Insert( E: Elements);
begin
  CheckLoadFactor( _Insert_);
  Inherited Insert(E);
end;

Function DIR_OpenDynamicHashTables.Delete: Elements;
begin
  Delete := Inherited Delete;
  CheckLoadFactor( _Delete_);
end;

Procedure DIR_OpenDynamicHashTables.CheckLoadFactor( C: Codes);
Var
  LoadFactor : Real;
begin
  LoadFactor := (Size / HashTable.MaximumNmr);
  Case C of
    _Insert_ : If LoadFactor>=HiLimit then
                 ReHashTo( HashTable.MaximumNmr * 2 );
    _Delete_ : If LoadFactor<=LoLimit then
                 ReHashTo( HashTable.MaximumNmr Div 2 );
  end;
end;

Procedure DIR_OpenDynamicHashTables.ReHashTo( N: Integer);
Var
  i,
  Transfer : Integer;
  Temporal : Array Of Elements;  // not necessary, but simpler...
begin
  Transfer := 0;
  SetLength( Temporal, 1+Size);
  With HashTable do begin
    For i := 1 to MaximumNmr do begin
      Location(i).Marked := False;
      If Item[i]<>NIL then begin
        Inc(Transfer);
        Temporal[Transfer] := Item[i];
        Item[i] := NIL;
      end;
    end;
    ReSize(N);
  end;
  //
  Size := 0;
  For i := 1 to Transfer do Inherited Insert( Temporal[i]);
  //
  Temporal := NIL;
end;

(* ----------------------------------------------------------------------- *)

Constructor DIR_ClosedHashTables.SetUp( N: Integer);
begin
  Inherited SetUp;
  HashTable := Buckets.SetUp(N);
end;

Function DIR_ClosedHashTables.Scheme : StorageSchemes;
begin
  Scheme := HashTable;
end;

Procedure DIR_ClosedHashTables.Insert( E: Elements);
Var
  Bucket : Lists;
  Hash   : Integer;
begin
  If E<>NIL then
    With HashTable do begin
      Hash := HashLocation( (E As KeyData).Key, MaximumNmr, 1);
      Bucket := List(Hash);
      With Bucket do Link( LinkFor(E), _After_, NIL);
      Inc(Size);
    end;
end;

Function DIR_ClosedHashTables.Delete: Elements;
begin
  Delete := HashTable.Current;
  If Result<>NIL then begin
    With HashTable, CurrentList do begin
      UnLink(CurrentLink);
      CurrentLink.Clear; // garbage collection
    end;
    Dec(Size);
  end;
end;

Function DIR_ClosedHashTables.FindKey(K: Elements): Elements;
Var
  Here   : Elements;
  Bucket : Lists;
  Hash   : Integer;
begin
  FindKey := NIL;
  With HashTable do begin
    Hash   := HashLocation( (K As KeyData).Key, MaximumNmr, 1);
    Bucket := List(Hash);
    Here   := LookFor( K, Bucket);
    If Here<>NIL then
      With HashTable do begin
        CurrentNmr := Hash;
        CurrentList.CurrentLink := Here;
        FindKey                 := Here[_Info_];
      end;
  end;
end;

Function DIR_ClosedHashTables.LookFor( E: Elements; L: Lists): Elements;
Var
  Here : Elements;
begin
  LookFor := NIL;
  Here    := L.Head;
  While (Result=NIL) and (Here<>NIL) do
    If Equal( E, Here[_Info_], Order) Then LookFor := Here
                                      Else Here := Here[_Next_];
end;

(* ----------------------------------------------------------------------- *)

Constructor DIR_DoubleEntryHashTables.SetUp( N: Integer);
Var
  i : Integer;
begin
  inherited SetUp(N);
  SetLength( BucketSize, 1+N);
  For i := 1 to N do BucketSize[i] := 0;
end;

Destructor DIR_DoubleEntryHashTables.Clear;
begin
  BucketSize := NIL;
  inherited Clear;
end;

Procedure DIR_DoubleEntryHashTables.Insert( E: Elements);
Var
  List1,
  List2 : Lists;
  Hash1,
  Hash2 : Integer;
  Size1,
  Size2 : Integer;
begin
  If E<>NIL then
     With HashTable do begin
       Hash1 := HashLocation( (E As KeyData).Key, MaximumNmr, 000);
       Hash2 := HashLocation( (E As KeyData).Key, MaximumNmr, 193); // eg...
       List1 := Item[Hash1][_Info_] As Lists;
       List2 := Item[Hash2][_Info_] As Lists;
       Size1 := BucketSize[Hash1];
       Size2 := BucketSize[Hash2];
       Case Size1<=Size2 of
         Ok: begin
               With List1 do Link( LinkFor(E), _After_, NIL);
               Inc(BucketSize[Hash1]);
             end;
         No: begin
               With List2 do Link( LinkFor(E), _After_, NIL);
               Inc(BucketSize[Hash2]);
             end;
       end;
       Inc(Size);
     end;
end;

Function DIR_DoubleEntryHashTables.Delete: Elements;
begin
  Delete := HashTable.Current;
  If Result<>NIL then begin
    With HashTable.CurrentList do begin
      UnLink(CurrentLink);
      CurrentLink.Clear; // garbage collection
    end;
    Dec(BucketSize[HashTable.CurrentNmr]);
    Dec(Size);
  end;
end;

Function DIR_DoubleEntryHashTables.FindKey( K: Elements): Elements;
Var
  List1,
  List2 : Lists;
  Hash1,
  Hash2 : Integer;
  Found1,
  Found2 : Boolean;
begin
  FindKey := NIL;
  If K<>NIL then
    With HashTable do begin
      Hash1 := HashLocation( (K As KeyData).Key, MaximumNmr, 0);
      Hash2 := HashLocation( (K As KeyData).Key, MaximumNmr, Hash1); // e.g.
      List1 := Item[Hash1][_Info_] As Lists;
      List2 := Item[Hash2][_Info_] As Lists;
      LookFor( K, List1, List2, Found1, Found2);
      If Found1 then CurrentNmr := Hash1;
      If Found2 then CurrentNmr := Hash2;
      FindKey := Current;
    end;
end;

Procedure DIR_DoubleEntryHashTables.LookFor( E: Elements; L1,L2: Lists;
                                             Var F1,F2: Boolean);
Var
  H1, H2: Elements;
begin
  F1 := No;      F2 := No;
  H1 := L1.Head; H2 := L2.Head;
  While Not (F1 or F2) and ((H1<>NIL) or (H2<>NIL)) do begin
    If (H1<>NIL) then begin
      F1 := F1 or Equal( E, H1[_Info_], Order);
      Case F1 of
        Ok: L1.CurrentLink := H1;
        No: H1 := H1[_Next_];
      end;
    end;
    If (H2<>NIL) then begin
      F2 := F2 or Equal( E, H2[_Info_], Order);
      Case F2 of
        Ok: L2.CurrentLink := H2;
        No: H2 := H2[_Next_]; 
      end;
    end;
  end;
end;

(* ----------------------------------------------------------------------- *)

Constructor DIR_BinaryTrees.SetUp( T: BinaryTrees);
begin
  Inherited SetUp;
  Case T<>NIL of
    Ok : Tree := T;
    No : Tree := BinaryTrees.SetUp;
  end;
end;

Function DIR_BinaryTrees.Scheme : StorageSchemes;
begin
  Scheme := Tree;
end;

Procedure DIR_BinaryTrees.Insert( E: Elements);
Var
  How       : Codes;
  This,Here : Elements;
begin
  If E<>NIL then begin
    With Tree do begin
     (* --- Looking for the correct link: -------------------------------- *)
      Here := Root;
      Case Here=NIL of
        Ok: How := _Root_;
        No: Repeat
              Case Order( E, Here[_Info_]) of
                _Less_    : How := _Lf_;
                _Equal_   : How := _Lf_; (**)
                _Greater_ : How := _Rt_;
              end;
              This := Here[How];
              If This<>NIL then Here := This;
            Until (This=NIL);
      end;
     (* --- Linking & Weighting ------------------------------------------ *)
      InsertLeaf( Tree.LinkFor(E), How, Here);
      Inc(Size);
     (* --- Done --------------------------------------------------------- *)
    end;
  end;
end;

Function DIR_BinaryTrees.Find( C: Codes): Elements;
begin
  Case C of
    _Top_,
    _Min_  : With Tree do
               Case Root<>NIL of
                 Ok: begin
                       CurrentLink := Root[_AllLf_];
                       Tree.FindBalance(Tree.CurrentLink);
                     end;
                 No: CurrentLink := NIL;
               end;
    _Max_  : With Tree do
               Case Root<>NIL of
                 Ok: begin
                       CurrentLink := Root[_AllRt_];
                       Tree.FindBalance(Tree.CurrentLink);
                     end;
                 No: CurrentLink := NIL;
               end;
  Else
    Inherited Find(C);
  end;
  Find := Tree.Current;
end;

Function DIR_BinaryTrees.Delete: Elements;
Var
  This: Elements;
//
  Function BubbleDown( L: Elements): Elements;
  Var
    Here,
    This,
    Lf,Rt : Elements;
  begin
    Here := L;
    Repeat
      Lf := Here[_Lf_];
      Rt := Here[_Rt_];
      Case NotNIL( Lf, Rt) of
        _Both_ : Case Random(2) of   // Just For Fun
                   0: This := Lf[_AllRt_];
                   1: This := Rt[_AllLf_];
                 end;
        _1st_  : This := Lf[_AllRt_];
        _2nd_  : This := Rt[_AllLf_];
        _None_ : This := NIL;
      end;
      If This<>NIL then begin
        SwapParts( This, _Info_, Here, _Info_);
        Here := This;
      end;
    Until (This=NIL); // In Balanced Trees This Will Happen Only Once!
    BubbleDown := Here;
  end;
//
begin
  With Tree do begin
    Delete := Current;
    This   := CurrentLink;
    If This<>NIL then begin
     (* --- Looking for the correct link: -------------------------------- *)
      This := BubbleDown(This);
     (* --- UnLinking & Weighting ---------------------------------------- *)
      DeleteLeaf(This);
      Dec(Size);
     (* --- Done --------------------------------------------------------- *)
    end;
  end;
end;

Function DIR_BinaryTrees.FindKey( K: Elements): Elements;
Var
  How   : Codes;
  Back,
  Here  : Elements;
  Found : Boolean;
begin
  With Tree do begin
   (* --- Looking for the correct link: ---------------------------------- *)
    Back := NIL;
    Here := Root;
    Found := No;
    While (Here<>NIL) and Not Found do begin
      Case Order( K, Here[_Info_] ) of
        _Less_    : How := _Lf_;
        _Equal_   : Found := Ok;
        _Greater_ : How := _Rt_;
      end;
      If Not Found Then begin
        Back := Here;
        Here := Here[How];
      end;
    end;
    CurrentLink := Here;
    Case Found of
      Ok: begin FindBalance(Here); FindKey := Current; end;
      No: begin FindBalance(Back); FindKey := NIL;     end;
    end;
  end;
end;

function DIR_BinaryTrees.IndexOf(E: Elements): Integer;
Var
  How   : Codes;
  Here  : Elements;
  Found : Boolean;
//
  Function WeightOf( S: Elements): Integer;
  begin
    Case S<>NIL of
      Ok: WeightOf := S.Weight;
      No: WeightOf := 0;
    end;
  end;
//
begin
  IndexOf := 0;
  With Tree do begin
    Here := Root;
    Found := No;
    While (Here<>NIL) and Not Found do begin
      Case Order( E, Here[_Info_] ) of
        _Less_    : How := _Lf_;
        _Equal_   : begin
                      Found := Ok;
                      IndexOf := Result + WeightOf(Here[_Lf_]);
                    end;
        _Greater_ : begin
                      How := _Rt_;
                      IndexOf := Result + WeightOf(Here[_Lf_]) + 1;
                    end;
      end;
      If Not Found Then
        Here := Here[How];
    end;
  end;
end;

Procedure DIR_BinaryTrees.ReportLess( Up: Elements; R: Reporter);
Var
  Here : Elements;
//
  Procedure ReportSubTree( E: Elements; R: Reporter);
  begin
    If E<>NIL then begin
      R.Announce( E[_Info_] );
      ReportSubTree( E[_Lf_], R);
      ReportSubTree( E[_Rt_], R);
    end;
  end;
//
begin
  With Tree do
    If Root<>NIL then begin
      Here := Root;
      While Here<>NIL do
        Case Order( Here[_Info_], Up) of
          _Less_    : begin
                        R.Announce(Here[_Info_]);
                        ReportSubTree( Here[_Lf_], R);
                        Here := Here[_Rt_];
                      end;
          _Equal_   : begin
                        R.Announce(Here[_Info_]);
                        ReportSubTree( Here[_Lf_], R);
                        Here := Here[_Rt_];  // handle double keys...?!
                      end;
          _Greater_ : Here := Here[_Lf_];
        end;
    end;
end;

Procedure DIR_BinaryTrees.ReportInRange( Lf,Rt: ELements; R: Reporter);
Var
  How  : Codes;
  Break,
  Here : Elements;
//
  Procedure ReportSubTree( E: Elements; R: Reporter);
  begin
    If E<>NIL then begin
      R.Announce( E[_Info_] );
      ReportSubTree( E[_Lf_], R);
      ReportSubTree( E[_Rt_], R);
    end;
  end;
//
begin
  With Tree do
    If Root<>NIL then begin
      // Find BreakPoint
      Break := Root;
      While (Break<>NIL) and Not InRange( Break[_Info_], Lf,Rt, Order) do
        If Order( Rt, Break[_Info_]) = _Less_ then Break := Break[_Lf_]
        Else
        If Order( Lf, Break[_Info_]) = _Greater_ then Break := Break[_Rt_];
      If Break<>NIL then begin
        R.Announce( Break[_Info_]);
        (* --- Report On Left Path --------------------------------------- *)
        Here := Break[_Lf_];
        While Here<>NIL do begin
          //
          Case Order( Lf, Here[_Info_]) of
            _Less_    : How := _Lf_;
            _Equal_   : How := _Lf_; // to be announced...
            _Greater_ : How := _Rt_;
          end;
          //
          If How = _Lf_ then begin
            R.Announce(Here[_Info_]);
            ReportSubTree( Here[_Rt_], R);
          end;
          Here := Here[How];
        end;
        (* --- Report On Right Path -------------------------------------- *)
        Here := Break[_Rt_];
        While Here<>NIL do begin
          //
          Case Order( Rt, Here[_Info_]) of
            _Less_    : How := _Lf_;
            _Equal_   : How := _Rt_; // to be announced...
            _Greater_ : How := _Rt_;
          end;
          //
          If How=_Rt_ then begin
            R.Announce(Here[_Info_]);
            ReportSubTree( Here[_Lf_], R);
          end;
          Here := Here[How];
        end;
      end;
    end;
end;

Procedure DIR_BinaryTrees.DrawLinks;
begin
  If Tree.Root<>NIL then begin
  //DrawAnnotation( W, '  (Min)', Tree.Root[_AllLf_], 0,-7);
    DrawAnnotation( W, 'Minimum element (Min)', Tree.Root[_AllLf_], 0,-7);
  //DrawAnnotation( W, '  (Max)', Tree.Root[_AllRt_], +1,+3);
    DrawAnnotation( W, 'Maximum element (Max)', Tree.Root[_AllRt_], +1,+3);
  end;
end;

(* ----------------------------------------------------------------------- *)

Constructor DIR_SkipLists.SetUp( SL: SkipLists);
begin
  Inherited SetUp;
  Case SL<>NIL of
    Ok : SkipList := SL;
    No : SkipList := SkipLists.SetUp;
  end;
end;

Function DIR_SkipLists.Scheme : StorageSchemes;
begin
  Scheme := SkipList;
end;

Procedure DIR_SkipLists.Insert( E: Elements);
Const
  _HighestLevel_  = _Last_;
  _OneLevelDown_  = _BackStep_;
var
  Where : Elements;
  Level : Integer;
begin
  With SkipList do
   (* --- This implementation does not afford double keys... ------------- *)
    If (E<>NIL) and (FindKey(E)=NIL) then begin
      ListOfLevels.Go(_HighestLevel_);
      Where := NIL;
      Repeat
        Where := WhereAfter( E, CurrentList, Where);
        ListOfLevels.Go(_OneLevelDown_);
        If (CurrentList<>NIL) and (Where<>NIL) then Where := Where[_DnLink_];
      Until (CurrentList=NIL);
      //
      Level := 0;
      Repeat
        Inc(Level);
      Until (Random(2)=1);
      Link( E, Where, Level); // Links "Level" copies...
      //
      Inc(Size);
      //
    end;
end;

Function DIR_SkipLists.Delete: Elements;
Var
  This : Elements;
begin
  With SkipList do begin
    Delete := Current;
    If Result<>NIL then begin
      This := CurrentList.CurrentLink;
      DeleteLinks(This); // garbage collection included in it 
      Dec(Size);
    end;
  end;
end;

Function DIR_SkipLists.Find( C: Codes): Elements;
begin
  Case C of
    _Top_,
    _Min_  : With SkipList do begin
               ListOfLevels.Go(_First_);
               CurrentList.Go(_First_);
             end;
    _Max_  : With SkipList do begin
               ListOfLevels.Go(_First_);
               CurrentList.Go(_Last_);
             end;
  Else
    Inherited Find(C);
  end;
  Find := SkipList.Current;
end;

Function DIR_SkipLists.FindKey( K: Elements): Elements;
Const
  _HighestLevel_  = _Last_;
  _OneLevelDown_  = _BackStep_;
var
  Where : Elements;
begin
  With SkipList do begin
    ListOfLevels.Go(_HighestLevel_);
    Where := NIL;
    Repeat
      Where := WhereAfter( K, CurrentList, Where);
      ListOfLevels.Go(_OneLevelDown_);
      If (CurrentList<>NIL) and (Where<>NIL) then Where := Where[_DnLink_];
    Until (CurrentList=NIL);
    // Useful for an unsuccessful search
    ListOfLevels.Go(_First_);
    CurrentList.CurrentLink := Where;
    //
    Case (Where<>NIL) and Equal(K,Where[_Info_],Order) of
      Ok: FindKey := Where[_Info_];
      No: FindKey := NIL;
    end;
  end;
end;

procedure DIR_SkipLists.ReportInRange(Lf, Rt: Elements; R: Reporter);
Var
  Here: Elements;
begin
  FindKey(Lf);
  Here := SkipList.CurrentList.CurrentLink;
  If Less( Here[_Info_], Lf, Order) then Here := Here[_Next_];
  While (Here<>NIL) and (LessOrEqual( Here[_Info_], Rt, Order)) do begin
    R.Announce(Here[_Info_]);
    Here := Here[_Next_];
  end;
end;

Procedure DIR_SkipLists.DrawLinks( W: TCanvas; k,l: Integer);
Var
  BottomList : Lists;
begin
  BottomList := SkipList.ListOfLevels.Head[_Info_] As Lists;
//DrawAnnotation( W, '  (Min)', BottomList.Head, +1,+6);
  DrawAnnotation( W, 'Minimum element (Min)', BottomList.Head, +1,+6);
//DrawAnnotation( W, '  (Max)', BottomList.Tail, +1,+4);
  DrawAnnotation( W, 'Maximum element (Max)', BottomList.Tail, +1,+4);
end;

Function DIR_SkipLists.IsAfter( E: Elements;
                                L: Lists;
                                var Here: Elements): Boolean;
Var
  HereNext : Elements;
begin
  Case Here=NIL of
    Ok: HereNext := L.Head;
    No: HereNext := Here[_Next_];
  end;
  Case (HereNext=NIL) Or Less( E, HereNext[_Info_], Order) Of
    Ok: IsAfter := Ok;
    No: begin IsAfter := No; Here := HereNext; end;
  end;
end;

Function DIR_SkipLists.WhereAfter( E: Elements;
                                   L: Lists;
                                   Here: Elements): Elements;
begin
  Case (L<>NIL) Of
    Ok: Repeat WhereAfter := Here; Until IsAfter( E, L, Here);
    No: WhereAfter := NIL;
  end;
end;

(* ----------------------------------------------------------------------- *)

Constructor DIR_BTrees.SetUp( MinS, MaxS: Integer);
begin
  Inherited SetUp;
  BTree := BTrees.SetUp( MinS, MaxS);
end;

Function DIR_BTrees.Scheme : StorageSchemes;
begin
  Scheme := BTree;
end;

Procedure DIR_BTrees.Insert( E: Elements);
Var
  B    : BTreeLinks;
  H    : Integer;
  Here : Elements;
begin
  // Elements Less    than Datum[H] in (H)-th subtree
  // Elements Greater than Datum[H] in (H+1)-th subtree
  // Search path for element E follows the link from B to Child[H]
  If (E<>NIL) then begin
    Here := FindKey(E);
   (* --- This implementation does not afford double keys... ------------- *)
    Case (Here<>NIL) of
      Ok: Here.Color := _InsColor_;
      No: begin
            B := NIL;
            H := 1;
            Repeat
              BTree.StepIfLowSize(B,H);
              H := B.LookFor( E, Order);
            Until B.Child[H]=NIL;
            B.InsData( H, E, NIL, NIL);  // Place H, Element E, No Children
            Inc(Size);
          end;
    end;
  end;
end;

Function DIR_BTrees.Delete: Elements;
Var
  Here  : BTreeLinks; H: Integer;
  Next  : BTreeLinks; N: Integer;
  Found : BTreeLinks; F: Integer;
  C     : Codes;
begin
  Delete := BTree.Current;
  If Result<>NIL then begin
    Found := NIL; F := 1;
    Here  := NIL; H := 1;
    Repeat
      (* --- Step One Level Down From Node "Here", & Child "H" if all Ok - *)
      BTree.StepIfHighSize( Here, H);
      If Here<>NIL then begin
        H := 1;
        Repeat
          C := Order( Here.Datum[H], Result); // Watch out: Result must be OK!
          Case C of
            _Less_    : Inc(H);
            _Equal_   : begin Found := Here; F := H; Inc(H); end; // Next!!
            _Greater_ : begin Next  := Here; N := H; end;
          end;
        Until (C=_Greater_) Or (H>Here.Size);
      end;
    Until Here=NIL;
    Case (Found.Child[F]=NIL) of
      Ok: Found.DelData( F, NIL);
      No: begin
            Found.Datum[F] := Next.Datum[N];
            Next.DelData( N, NIL);
          end;
    end;
    Dec(Size);
  end;
end;

Function DIR_Btrees.Find(C: Codes): Elements;
begin
  Case C of
    _Top_,
    _Min_  : With BTree.Tree do begin
               Case Root<>NIL of
                 Ok: CurrentLink := Root[_AllLf_];
                 No: CurrentLink := NIL;
               end;
               If CurrentLink<>NIL then
                 With (CurrentLink As BTreeLinks) do
                   CurrentData := 1;
             end;
    _Max_  : With BTree.Tree do begin
               Case Root<>NIL of
                 Ok: CurrentLink := Root[_AllRt_];
                 No: CurrentLink := NIL;
               end;
               If CurrentLink<>NIL then
                 With (CurrentLink As BTreeLinks) do
                   CurrentData := Size;
             end;
  Else
    Inherited Find(C);
  end;
  Find := BTree.Current;
end;


Function DIR_Btrees.FindKey(K: Elements): Elements;
Var
  Here : BTreeLinks;
  H : Integer;
begin
  FindKey := NIL;
  With BTree do begin
    Tree.CurrentLink := NIL;
    Here := (Tree.Root As BTreeLinks);
    While (Here<>NIL) and (Result=NIL) do begin
      H := Here.LookFor(K, Order);
      Case (H<=Here.Size) and Equal( Here.Datum[H], K, Order) of
        Ok: begin
              FindKey := Here.Datum[H];
              Tree.CurrentLink := Here;
              Here.CurrentData := H;
            end;
        No: Here := Here.Child[H];
      end;
    end;
  end;
end;

Procedure DIR_Btrees.ReportInRange(Lf, Rt: Elements; R: Reporter);
begin
  inherited ReportInRange(Lf,Rt,R); // Well, a good exercise for the user...
end;

(* ----------------------------------------------------------------------- *)

Constructor DIR_SelfAdjustedLists.SetUp( L: Lists);
begin
  Inherited SetUp;
  Case L<>NIL of
    Ok: List := L;
    No: List := Lists.SetUp;
  end;
end;

Function DIR_SelfAdjustedLists.Scheme : StorageSchemes;
begin
  Scheme := List;
end;

Procedure DIR_SelfAdjustedLists.Insert( E: Elements);
begin
  With List do begin
    Link( LinkFor(E), _Before_, Head);
    Inc(Size);
  end;
end;

Function DIR_SelfAdjustedLists.Delete: Elements;
begin
  With List do begin
    Delete := Current;
    If Result<>NIL then begin
      UnLink(CurrentLink);
      CurrentLink.Clear; // garbage collection
      Dec(Size);
    end;
  end;
end;

Function DIR_SelfAdjustedLists.FindKey( K: Elements): Elements;
Var
  Here : Elements;
begin
  Here := List.Go(_First_);
  While (Here<>NIL) and Not Equal( K, Here[_Info_], Order) do
    Here := List.Go(_Step_);
  Case Here=NIL of
    Ok: FindKey := NIL;
    No: begin
          FindKey := Here[_Info_];
          With List do begin
            UnLink(Here);
            Link( Here, _Before_, Head);
          end;
        end;
  end;
end;

(* ----------------------------------------------------------------------- *)

Constructor PSDIR_BinaryTrees.SetUp( T: BinaryTrees);
begin
  Inherited SetUp;
  Case T<>NIL of
    Ok: Tree := T;
    No: Tree := BinaryTrees.SetUp;
  end;
  HeapOrder := NumericOrder;
end;

Function PSDIR_BinaryTrees.Scheme: StorageSchemes;
begin
  Scheme := Tree;
end;

Procedure PSDIR_BinaryTrees.Insert( E: Elements);
Var
  How   : Codes;
  This,
  Here  : Elements;
  L     : Elements;
  V,H   : Elements;
begin
  (* --- This implementation does not afford double keys... -------------- *)
  If (E<>NIL) and (FindKey(E)=NIL) then begin
    L := PrSearchLinks.SetUp( E, Order, HeapOrder);
    With Tree do begin
      Here := Root;
      Case Here=NIL of
        Ok: How := _Root_;
        No: Repeat
              (* --- Take Care Heap Order -------------------------------- *)
              H := Here[_HeapInfo_];
              V :=    L[_HeapInfo_];
              If (V<>NIL) and ((H=NIL) Or Less( V, H, HeapOrder)) then begin
                (Here As PrSearchLinks).PushDn(V);
                L[_HeapInfo_] := NIL;
              end;
              (* --------------------------------------------------------- *)
              Case Order( E, Here[_Info_]) of
                _Less_    : How := _Lf_;
                // _Equal_  this is avoided here
                _Greater_ : How := _Rt_;
              end;
              (* --------------------------------------------------------- *)
              This := Here[How];
              If This<>NIL then Here := This;
            Until (This=NIL);
      end;
     (* --- Insert: ------------------------------------------------------ *)
      InsertLeaf( L, How, Here);
      Inc(Size);
    end;
  end;
end;

Function PSDIR_BinaryTrees.Delete: Elements;
// With A PS_DIR_ we expect the same type of links
Var
  This     : PrSearchLinks;
  Here,
  HeapPart : Elements;
begin
  With Tree do begin
    Delete := Current;
    This   := Tree.CurrentLink As PrSearchLinks;
    If (Result<>NIL) then begin
      (* --- Prepare deleting heap-part - if UpWards! -------------------- *)
      If This.HasCopy then begin
        Here := This;
        HeapPart := Here[_HeapInfo_];
        While (HeapPart=NIL) or Not Equal( This[_Info_], HeapPart, Order) do
        begin
          Here := Here[_Parent_];
          If Here<>NIL then
            HeapPart := Here[_HeapInfo_];
        end;
        (Here As PrSearchLinks).PullUp;
      end;
      (* --- Prepare deleting dir-part ----------------------------------- *)
      This := This.SwapToLeaf As PrSearchLinks;
      (* --- Delete: ----------------------------------------------------- *)
      DeleteLeaf(This); // garbage collection is included in it
      Dec(Size);
    end;
  end;
end;

Function PSDIR_BinaryTrees.FindKey( K: Elements): Elements;
Var
  How   : Codes;
  Here  : Elements;
  Found : Boolean;
begin
  With Tree do begin
   (* --- Looking for the right item ------------------------------------- *)
    Here := Root;
    Found := No;
    While (Here<>NIL) and Not Found do begin
      Case Order( K, Here[_Info_] ) of
        _Less_    : How := _Lf_;
        _Equal_   : Found := Ok;
        _Greater_ : How := _Rt_;
      end;
      If Not Found Then
        Here := Here[How];
    end;
    CurrentLink := Here;
    Case Found of
      Ok: FindKey := Current;
      No: FindKey := NIL;
    end;
  end;
end;

Function PSDIR_BinaryTrees.Find( C: Codes): Elements;
begin
  Case C of
    _Top_,
    _Min_  : With Tree do
               Case Root<>NIL of
                 Ok: CurrentLink := Root[_AllLf_];
                 No: CurrentLink := NIL;
               end;
    _Max_  : With Tree do
               Case Root<>NIL of
                 Ok: CurrentLink := Root[_AllRt_];
                 No: CurrentLink := NIL;
               end;
  Else
    Inherited Find(C);
  end;
  Find := Tree.Current;
end;

Procedure PSDIR_BinaryTrees.ReportInArea( Up,Lf,Rt: Elements; R: Reporter);
// Reports all elements which are:
// - InRange(Lf,Rt)    with respect to directory (primary) order
// - Less/Equal to Up  with respect to heap (secondary) order
Var
  Break: Elements;
//
  Function Descend( E: Elements; How: Codes): Elements;
  begin
    Case How of
      _Lf_   : Descend := E[_Lf_];
      _Rt_   : Descend := E[_Rt_];
      _Root_ : Descend := Tree.Root;
    end;
    If (Result<>NIL) and (Result[_HeapInfo_]<>NIL) then
      If Not LessOrEqual( Result[_HeapInfo_], Up, HeapOrder) then
        Descend := NIL;
  end;
//
  Procedure CheckAnnounce( L: PrSearchLinks);
  Var
    D,H : Elements;
  begin
    D := L[_Info_];
    H := L[_HeapInfo_];
    If Not L.HasCopy and InRange( D, Lf, Rt, Order)
                     and LessOrEqual( D, Up, HeapOrder) then R.Announce(D);
    If (H<>NIL) and InRange( H, Lf, Rt, Order)          then R.Announce(H);
  end;
//
  Procedure ReportSubTree( E: Elements);
  begin
    If E<>NIL then begin
      CheckAnnounce(E As PrSearchLinks);
      ReportSubTree( Descend( E, _Lf_) );
      ReportSubTree( Descend( E, _Rt_) );
    end;
  end;
//
  Procedure ReportPath( E: Elements; Path, Side: Codes);
  Var
    How  : Codes;
    Here : Elements;
  begin
    Here := Descend( Break, Path);
    While Here<>NIL do begin
      CheckAnnounce( Here As PrSearchLinks);
      (* ----------------------------------------------------------------- *)
      Case Order( E, Here[_Info_]) of
        _Less_    : How := _Lf_;
        _Equal_   : How := Path; // To be announced...
        _Greater_ : How := _Rt_;
      end;
      (* ----------------------------------------------------------------- *)
      If How=Path then
        ReportSubTree( Descend( Here, Side) );
      Here := Descend( Here, How);
    end;
  end;
//
   Procedure GetBreakPoint( var Break: Elements);
   Var
     Found : Boolean;
   begin
     Break := Descend( NIL, _Root_);
     Found := No;
     While Not Found and (Break<>NIL) do begin
       CheckAnnounce( Break As PrSearchLinks );
       If Less( Rt, Break[_Info_], Order) then Break := Descend( Break, _Lf_)
       Else
       If Less( Break[_Info_], Lf, Order) then Break := Descend( Break, _Rt_)
       Else Found := Ok ;
     end;
   end;
//
begin
  GetBreakPoint( Break );
  If Break<>NIL then begin
    ReportPath( Lf, _Lf_, _Rt_);
    ReportPath( Rt, _Rt_, _Lf_);
  end;
end;

(* ======================================================================= *)

begin
end.

