(* ----------------------------------------------------------------------- *)
UNIT ELEMENTS_;
(* ----------------------------------------------------------------------- *)

INTERFACE
(* ----------------------------------------------------------------------- *)

USES

  SysUtils,
  Graphics,

  CONSTANTS_,
  DRAW_;
(* ----------------------------------------------------------------------- *)

TYPE

  Elements = Class
   (* --- Basic Description ---------------------------------------------- *)
                                                                 // Abstract
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp;
    Destructor  Clear;                                              Virtual;
   (* --- Navigation ----------------------------------------------------- *)
    Function    Current: Elements;                                  Virtual;
    Function    GetLink(C: Codes): Elements;                        Virtual;
    Procedure   SetLink(C: Codes; E: Elements);                     Virtual;
    Property    _[C: Codes]: Elements   Read GetLink Write SetLink; DEFAULT;
   (* --- Information ---------------------------------------------------- *)
    Function    GetMark: Boolean;                                   Virtual;
    Procedure   SetMark( B: Boolean);                               Virtual;
    Property    Marked: Boolean                 Read GetMark Write SetMark;
                (* ------------------------------------------------------- *)
    Function    GetNumber: Integer;                                 Virtual;
    Procedure   SetNumber( N: Integer);                             Virtual;
    Property    Number: Integer              Read GetNumber Write SetNumber;
                (* ------------------------------------------------------- *)
    Function    GetRank: Integer;                                   Virtual;
    Procedure   SetRank( R: Integer);                               Virtual;
    Property    Rank: Integer                    Read GetRank Write SetRank;
    Function    RankOf( C: Codes): Integer;
                (* ------------------------------------------------------- *)
    Function    GetWeight: Integer;                                 Virtual;
    Procedure   SetWeight( W: Integer);                             Virtual;
    Property    Weight: Integer              Read GetWeight Write SetWeight;
                (* ------------------------------------------------------- *)
    Function    IsWhat: Codes;                                      Virtual;
   (* --- Visualization -------------------------------------------------- *)
    Function    AsText: String;                                     Virtual;
    Procedure   Draw( W: TCanvas; k,l: Integer);                    Virtual;
    Procedure   DrawLink( W: TCanvas; k,l: Integer; C: Codes);      Virtual;
    Procedure   DrawLinks( W: TCanvas; k,l: Integer);               Virtual;
                (* ------------------------------------------------------- *)
    Function    GetX: Integer;                                      Virtual;
    Procedure   SetX( X: Integer);                                  Virtual;
    Property    X: Integer                             Read GetX Write SetX;
                (* ------------------------------------------------------- *)
    Function    GetY: Integer;                                      Virtual;
    Procedure   SetY( y: Integer);                                  Virtual;
    Property    Y: Integer                             Read GetY Write SetY;
                (* ------------------------------------------------------- *)
    Function    GetColor: TColor;                                   Virtual;
    Procedure   SetColor( C: TColor);                               Virtual;
    Property    Color: TColor                  Read GetColor Write SetColor;
   (* -------------------------------------------------------------------- *)
  end;

  OrderFunctions = Function(E1,E2: Elements): Codes;
(* ======================================================================= *)

Procedure Done;
Function  NotNIL( E1,E2: Elements): Codes;

Function  Equal( E1, E2: Elements; Order: OrderFunctions): Boolean;
Function  Less( E1, E2: Elements; Order: OrderFunctions): Boolean;
Function  LessOrEqual( E1, E2: Elements; Order: OrderFunctions): Boolean;
Function  InRange( E, E1, E2: Elements; Order: OrderFunctions): Boolean;

Function  Minimum( E1,E2: Elements; Order: OrderFunctions): Codes;
Function  Maximum( E1,E2: Elements; Order: OrderFunctions): Codes;

Function  MinOf( E1,E2: Elements; Order: OrderFunctions): Elements;
Function  MaxOf( E1,E2: Elements; Order: OrderFunctions): Elements;

Procedure Swap( Var E1, E2: Elements);
Procedure SwapParts( E1: Elements; C1: Codes; E2: Elements; C2: Codes);
Procedure SwapMinMax( Var E1, E2: Elements; Order: OrderFunctions);

Procedure DrawAnnotation(W: TCanvas; S: String; E: Elements; Nx,Ny: Integer);
(* ======================================================================= *)

IMPLEMENTATION
(* ----------------------------------------------------------------------- *)

Procedure Done;
begin
end;

Function NotNIL( E1,E2: Elements): Codes;
begin
  NotNIL := _Null_;
  Case E1<>NIL of
    Ok: Case E2<>NIL of
          Ok: NotNIL := _Both_;
          No: NotNil := _1st_;
        end;
    No: Case E2<>NIL of
          Ok: NotNIL := _2nd_;
          No: NotNil := _None_;
        end;
  end;
end;

Function Equal( E1, E2: Elements; Order: OrderFunctions): Boolean;
begin
  Equal := (Order(E1,E2)=_Equal_);
end;

Function Less( E1, E2: Elements; Order: OrderFunctions): Boolean;
begin
  Less := (Order(E1,E2)=_Less_);
end;

Function LessOrEqual( E1, E2: Elements; Order: OrderFunctions): Boolean;
begin
  LessOrEqual := (Order(E1,E2) in [_Less_,_Equal_]);
end;

Function InRange( E, E1, E2: Elements; Order: OrderFunctions): Boolean;
begin
  InRange := LessOrEqual( E1, E, Order) and LessOrEqual( E, E2, Order);
end;

Function Minimum( E1,E2: Elements; Order: OrderFunctions): Codes;
begin
  Case NotNIL(E1,E2) of
    _Both_ : Case Order(E1,E2) of
               _Less_,
               _Equal_   : Minimum := _1st_;
               _Greater_ : Minimum := _2nd_;
             end;
    _1st_  : Minimum := _1st_;
    _2nd_  : Minimum := _2nd_;
    _None_ : Minimum := _None_;
  end;
end;

Function Maximum( E1,E2: Elements; Order: OrderFunctions): Codes;
begin
  Case NotNIL(E1,E2) of
    _Both_ : Case Order(E1,E2) of
               _Less_,
               _Equal_   : Maximum := _2nd_;
               _Greater_ : Maximum := _1st_;
             end;
    _1st_  : Maximum := _1st_;
    _2nd_  : Maximum := _2nd_;
    _None_ : Maximum := _None_;
  end;
end;

Function MinOf( E1,E2: Elements; Order: OrderFunctions): Elements;
begin
  Case Minimum( E1,E2, Order) of
    _1st_  : MinOf := E1;
    _2nd_  : MinOf := E2;
  end;
end;

Function MaxOf( E1,E2: Elements; Order: OrderFunctions): Elements;
begin
  Case Maximum( E1,E2, Order) of
    _1st_  : MaxOf := E1;
    _2nd_  : MaxOf := E2;
  end;
end;

Procedure Swap( Var E1, E2: Elements);
Var
  T : Elements;
begin
  T  := E1;
  E1 := E2;
  E2 := T;
end;

Procedure SwapParts( E1: Elements; C1: Codes; E2: Elements; C2: Codes);
Var
  T1, T2: Elements;
begin
  T1 := E1[C1];
  T2 := E2[C2];
  E1[C1] := T2;
  E2[C2] := T1;
end;

Procedure SwapMinMax( Var E1, E2: Elements; Order: OrderFunctions);
begin
  Case NotNIL(E1,E2) of
    _None_,
    _1st_ : Done;
    _2nd_ : Swap( E1, E2);
    _Both_: Case Order(E1,E2) of
              _Less_,
              _Equal_  : Done;
              _Greater_: Swap( E1, E2);
            end;
  end;
end;
(* ----------------------------------------------------------------------- *)

(* ----------------------------------------------------------------------- *)
Constructor Elements.SetUp;
begin
end;

Destructor Elements.Clear;
begin
end;

Procedure Elements.Draw( W: TCanvas; k,l: Integer);
begin
end;

Procedure Elements.DrawLink( W: TCanvas; k,l: Integer; C: Codes);
Var
  Here : Elements;
begin
  Here := Self[C];
  If Here<>NIL then
    DrawLine( W, k+X, l+Y, k+Here.X, l+Here.Y, _LinkColor_);
end;

Function Elements.AsText: String;
begin
  AsText := '';
end;

Procedure Elements.DrawLinks( W: TCanvas; k,l: Integer);
begin
end;

Procedure Elements.SetX( X: Integer);
begin
end;

Function Elements.GetX: Integer;
begin
  GetX := 0;
end;

Procedure Elements.SetY( Y: Integer);
begin
end;

Function Elements.GetY: Integer;
begin
  GetY := 0;
end;

Function Elements.GetColor: TColor;
begin
  GetColor := ClBlack;
end;

Procedure Elements.SetColor( C: TColor);
begin
end;

Function Elements.Current: Elements;
begin
  Current := Self;
end;

Function Elements.GetMark: Boolean;
begin
  GetMark := No;
end;

Procedure Elements.SetMark(B: Boolean);
begin
end;

Function Elements.GetLink(C: Codes): Elements;
begin
  Case C of
    _Info_ : GetLink := Self;
  Else
    GetLink := NIL;
  end;
end;

Procedure Elements.SetLink(C: Codes; E: Elements);
begin
end;

Procedure Elements.SetNumber( N: Integer);
begin
end;

Function Elements.GetNumber: Integer;
begin
  GetNumber := 0;
end;

Procedure Elements.SetRank( R: Integer);
begin
end;

Function Elements.GetRank: Integer;
begin
  GetRank:= 0;
end;

Function Elements.RankOf( C: Codes): Integer;
Var
  E : Elements;
begin
  E := Self[C];
  Case E<>NIL of
    Ok : RankOf := E.Rank;
    No : RankOf := 0;
  end;
end;

Procedure Elements.SetWeight( W: Integer);
begin
end;

Function Elements.GetWeight: Integer;
begin
  GetWeight := 0;
end;

Function  Elements.IsWhat: Codes;
begin
  IsWhat := _Null_;
end;
(* ----------------------------------------------------------------------- *)

Procedure DrawAnnotation(W: TCanvas; S: String; E: Elements; Nx,Ny: Integer);
Var
  k,l,
  Dk,Dl: Integer;
begin
  If E<>NIL then
    With W do begin
      Font.Size := FontSize;
      Dk := W.TextWidth(S);
      Dl := W.TextHeight('S');
      k := E.x + Nx*CellWidth;
      l := E.y + Ny*CellHeight;
      W.TextOut( k, l-Dl, S);
      DrawLine( W, k,l, E.x,E.y, _NoteColor_);
      MoveTo( k, l); LineTo( k+Dk, l);
    end;
end;

(* ======================================================================= *)

begin
end.
