(* ----------------------------------------------------------------------- *)
UNIT LINKS_;
(* ----------------------------------------------------------------------- *)

INTERFACE
(* ----------------------------------------------------------------------- *)

USES

  Windows,
  SysUtils,
  Graphics,

  CONSTANTS_,
  DRAW_,
  ELEMENTS_,
  DATA_;
(* ----------------------------------------------------------------------- *)

TYPE

  Links = Class(Elements)
   (* --- Basic Description ---------------------------------------------- *)
    Info   : Elements;
    Xpos,
    Ypos   : Integer;
    Color  : Tcolor;
    Number : Integer;
    Weight : Integer;
    Marked : Boolean;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp;
    Destructor  Clear;                                             Override;
   (* --- Navigation ----------------------------------------------------- *)
    Function    Current: Elements;                                 Override;
    Function    GetLink(C: Codes): Elements;                       Override;
    Procedure   SetLink(C: Codes; E: Elements);                    Override;
   (* --- Information ---------------------------------------------------- *)
    Function    GetMark: Boolean;                                  Override;
    Procedure   SetMark( B: Boolean);                              Override;
                (* ------------------------------------------------------- *)
    Function    GetX: Integer;                                     Override;
    Procedure   SetX( X: Integer);                                 Override;
                (* ------------------------------------------------------- *)
    Function    GetY: Integer;                                     Override;
    Procedure   SetY( y: Integer);                                 Override;
                (* ------------------------------------------------------- *)
    Function    GetWeight: Integer;                                Override;
    Procedure   SetWeight( W: Integer);                            Override;
   (* --- Visualization -------------------------------------------------- *)
    Function    AsText: String;                                    Override;
    Procedure   Draw( W: TCanvas; k,l: Integer );                  Override;
   (* -------------------------------------------------------------------- *)
  end;

(* ----------------------------------------------------------------------- *)

  ListLinks = Class(Links)
   (* --- Basic Description ---------------------------------------------- *)
    Next,
    Prev  : Elements;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp;
   (* --- Navigation ----------------------------------------------------- *)
    Function    GetLink(C: Codes): Elements;                       Override;
    Procedure   SetLink(C: Codes; E: Elements);                    Override;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   DrawLinks( W: TCanvas; k,l: Integer );             Override;
   (* -------------------------------------------------------------------- *)
  end;

(* ----------------------------------------------------------------------- *)

  SkipListLinks = Class(ListLinks)
   (* --- Basic Description ---------------------------------------------- *)
    Up, Dn : Elements;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp;
   (* --- Navigation ----------------------------------------------------- *)
    Function    GetLink(C: Codes): Elements;                       Override;
    Procedure   SetLink(C: Codes; E: Elements);                    Override;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   DrawLinks( W: TCanvas; k,l: Integer );             Override;
   (* -------------------------------------------------------------------- *)
  end;

(* ------------------------------------------------------------------------ *)

  TreeLinks = Class(Links)
   (* --- Basic Description ---------------------------------------------- *)
    Rank   : Integer;
    Parent : Elements;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp;
   (* --- Navigation ----------------------------------------------------- *)
    Function    GetLink( C: Codes): Elements;                      Override;
    Procedure   SetLink( C: Codes; E: Elements);                   Override;
   (* --- Information ---------------------------------------------------- *)
    Function    GetRank: Integer;                                  Override;
    Procedure   SetRank( R: Integer);                              Override;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   DrawLinks( W: TCanvas; k,l: Integer);              Override;
   (* -------------------------------------------------------------------- *)
  end;

(* ----------------------------------------------------------------------- *)

  FixDegTreeLinks = Class( TreeLinks)
  // reserved for trees where the degree of a node is 'fixed'
  end;

(* ----------------------------------------------------------------------- *)

  AnyDegTreeLinks = Class( TreeLinks)
  // for trees where the degree of a node can become arbitrarily large
   (* --- Basic Description ---------------------------------------------- *)
    NextSibling,
    PrevSibling,
    FirstChild,
    LastChild : Elements;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp;
   (* --- Navigation ----------------------------------------------------- *)
    Function    GetLink( C: Codes): Elements;                      Override;
    Procedure   SetLink( C: Codes; E: Elements);                   Override;
   (* -------------------------------------------------------------------- *)
  end;

(* ----------------------------------------------------------------------- *)

  BinaryTreeLinks = Class(FixDegTreeLinks)
   (* --- Basic Description ---------------------------------------------- *)
    Left,
    Right  : Elements;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp;
   (* --- Navigation ----------------------------------------------------- *)
    Function    GetLink( C: Codes): Elements;                      Override;
    Procedure   SetLink( C: Codes; E: Elements);                   Override;
   (* --- Information ---------------------------------------------------- *)
    Function    IsWhat: Codes;                                     Override;
   (* --- Auxiliary Methods ---------------------------------------------- *)
    Procedure   RotateUp;                                           Virtual;
    Procedure   SwapUpTo( E: Elements);                             Virtual;
    Function    SwapToLeaf: Elements;
   (* -------------------------------------------------------------------- *)
  end;

(* ----------------------------------------------------------------------- *)

  DoubleBinTreeLinks = Class( BinaryTreeLinks)
   (* --- Basic Description ---------------------------------------------- *)
    HeapInfo    : Elements;
    CurrentInfo : Codes;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( E1, E2: Elements);
    Destructor  Clear;                                             Override;
   (* --- Navigation ----------------------------------------------------- *)
    Function    GetLink( C: Codes): Elements;                      Override;
    Procedure   SetLink( C: Codes; E: Elements);                   Override;
   (* -------------------------------------------------------------------- *)
  end;

(* ----------------------------------------------------------------------- *)

  PrSearchLinks = Class(DoubleBinTreeLinks)
   (* --- Basic Description ---------------------------------------------- *)
    Order,
    HeapOrder: OrderFunctions;
    HasCopy  : Boolean;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( E: Elements; O1, O2: OrderFunctions);
   (* --- Auxiliary Methods ---------------------------------------------- *)
    Procedure   SwapUpTo( E: Elements);                            Override;
    Procedure   RotateUp;                                          Override;
    Procedure   PushDn( This: Elements);
    Procedure   PullUp;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   Draw( W: TCanvas; k,l: Integer );                  Override;
   (* -------------------------------------------------------------------- *)
  end;

(* ----------------------------------------------------------------------- *)

  BTreeLinks = Class( TreeLinks )
   (* --- Basic Description ---------------------------------------------- *)
    Degree      : Integer;
    Child       : Array Of BTreeLinks;
                (* ------------------------------------------------------- *)
    Size        : Integer;
    Data        : Array Of Elements;
    CurrentData : Integer;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( MaxDegree: Integer);
    Destructor  Clear;                                             Override;
   (* --- Navigation ----------------------------------------------------- *)
    Function    Current: Elements;                                 Override;
    Function    GetLink( C: Codes): Elements;                      Override;
    Procedure   SetData(k: Integer; E: Elements);
    Function    GetData(k: Integer): Elements;
    Property    Datum[k: Integer]: Elements      Read GetData Write SetData;
   (* --- Auxiliary Methods ---------------------------------------------- *)
    Constructor Meld( E: Elements; L,R: BTreeLinks);
    Procedure   Split( D: Integer; Var E: Elements; Var L,R: BTreeLinks);
    Procedure   InsData( P: Integer; E: Elements; L,R: BTreeLinks);
    Procedure   DelData( P: Integer; B: BTreeLinks);
    Procedure   RotateUp( Which: Codes; H: Integer);
    Function    LookFor( E: Elements; O: OrderFunctions): Integer;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   Draw( W: TCanvas; k,l: Integer );                  Override;
   (* -------------------------------------------------------------------- *)
  end;

(* ======================================================================= *)

IMPLEMENTATION
(* ======================================================================= *)

Constructor Links.SetUp;
begin
  inherited SetUp;
  Info   := NIL;
  Xpos   := 0;
  Ypos   := 0;
  Color  := _LinkColor_;
  Number := 0;
  Weight := 1;
  Marked := False;
end;

Destructor Links.Clear;
begin
 (* --- Allow the system's garbage collector to take effect -------------- *)
  Info := NIL;
end;

Function Links.Current: Elements;
begin
  Current := NIL;
  If Info<>NIL then
    Current := Info.Current;
end;

Function Links.GetLink(C: Codes): Elements;
begin
  Case C=_Info_ of
    Ok : begin
           GetLink := Info;
          (* --- Color relevant element as 'USED' --- *)
           If Info<>NIL then begin
             Info.Color := _UsedColor_;
           end;
          (* ---------------------------------------- *)
         end;
    No : GetLink := NIL;
  end;
end;

Procedure Links.SetLink(C: Codes; E: Elements);
begin
  Case C of
    _Info_    : begin
                  Info := E;
                  (* --- Color relevant element as 'USED' --- *)
                   If Info<>NIL then begin
                     Info.Color := _UsedColor_;
                   end;
                  (* ---------------------------------------- *)
                end;
    _NewInfo_ : begin
                  If (Info<>NIL) and (Info<>E) then Info.Clear;
                  Info := E;
                  (* --- Color relevant element as 'USED' --- *)
                   If Info<>NIL then begin
                     Info.Color := _UsedColor_;
                   end;
                  (* ---------------------------------------- *)
                end;
  end;
end;

Procedure Links.SetX( X: Integer);
begin
  Xpos := X;
end;

Function Links.GetX: Integer;
begin
  GetX := Xpos;
end;

Procedure Links.SetY( Y: Integer);
begin
  Ypos := Y;
end;

Function Links.GetY: Integer;
begin
  GetY := Ypos;
end;

Function Links.AsText: String;
begin
  AsText := '';
 (* ---   ------------------------------------------------------ *)
  If ShowRanks then AsText := Result + Format( '(%d)',[Rank] );
 (* ---   ----------------------------------------------------- *)
  If ShowWeights then AsText := Result + Format( ' #%d',[Weight]);
end;

Procedure Links.Draw( W: TCanvas; k,l: Integer );
var
  S     : String;
  Adjust,
  Dx,Dy,
  Tx,Ty : Integer;
begin
  With W do begin
    Tx := k+X;
    Ty := l+Y;
   (* --- Mark ----------------------------------------------------------- *)
    If Marked then begin
      Font.Size := FontSize-1;
      Font.Color := _NoteColor_;
      Brush.Color := _BkGrColor_;
      TextOut( Tx+CellWidth+3,Ty, ''); // i.e. a "paragraph" sign
    end;
   (* --- Parameters ----------------------------------------------------- *)
    S := AsText;
    If S<>'' then begin
      Font.Size := FontSize-1;
      Dx := TextWidth(S);
      Dy := CellHeight;
      Adjust := (CellWidth-Dx) Div 2;
      DrawFrame( W, Tx+Adjust-1,Ty-Dy-1, Tx+Adjust+Dx+1,Ty+1, _FillColor_);
      Font.Color := _NumberColor_;
      TextOut( Tx+Adjust,Ty-Dy, S);
    end;
   (* --- Content -------------------------------------------------------- *)
    If Info<>NIL then Info.Draw( W, Tx, Ty);
   (* -------------------------------------------------------------------- *)
  end;
end;

function Links.GetMark: Boolean;
begin
  GetMark := Marked;
end;

procedure Links.SetMark(B: Boolean);
begin
  Marked := B;
end;

function Links.GetWeight: Integer;
begin
  GetWeight := Weight;
end;

procedure Links.SetWeight(W: Integer);
begin
  Weight := W;
end;
(* ----------------------------------------------------------------------- *)

(* ----------------------------------------------------------------------- *)
Constructor ListLinks.SetUp;
begin
  Inherited SetUp;
  Next := NIL;
  Prev := NIL;
end;

Function ListLinks.GetLink( C: Codes): Elements;
begin
  Case C of
   (* --- Fundamental ------------------------------------------- *)
    _Next_ : GetLink := Next;
    _Prev_ : GetLink := Prev;
  Else
    GetLink := inherited GetLink(C);
  end;
end;

Procedure ListLinks.SetLink( C: Codes; E: Elements);
begin
  Case C of
    _Next_ : Next := E;
    _Prev_ : Prev := E;
  Else
    inherited SetLink(C,E);
  end;
end;

Procedure ListLinks.DrawLinks( W: TCanvas; k,l: Integer);
begin
  DrawLink( W, k,l, _Prev_);
  DrawLink( W, k,l, _Next_);
end;
(* ----------------------------------------------------------------------- *)

(* ----------------------------------------------------------------------- *)
Constructor SkipListLinks.SetUp;
begin
  Inherited SetUp;
  Up := NIL;
  Dn := NIL;
end;

Function SkipListLinks.GetLink( C: Codes): Elements;
Var
  Here : Elements;
begin
  Case C of
   (* --- Fundamental ---------------------------------------------------- *)
    _Next_   : GetLink := Next;
    _Prev_   : GetLink := Prev;
    _UpLink_ : GetLink := Up;
    _DnLink_ : GetLink := Dn;
   (* --- Derived -------------------------------------------------------- *)
    _UpPrev_ : begin
                 Here := Self;
                 Repeat
                   GetLink := Here[_UpLink_];
                   Here    := Here[_Prev_];
                 Until (Result<>NIL) or (Here=NIL);
               end;
    _UpNext_ : begin
                 Here := Self;
                 Repeat
                   GetLink := Here[_UpLink_];
                   Here    := Here[_Next_];
                 Until (Result<>NIL) or (Here=NIL);
               end;
   (* -------------------------------------------------------------------- *)
  Else
    GetLink := inherited GetLink(C);
  end;
end;

Procedure SkipListLinks.SetLink( C: Codes; E: Elements);
begin
  Case C of
    _Next_   : Next := E;
    _Prev_   : Prev := E;
    _UpLink_ : Up   := E;
    _DnLink_ : Dn   := E;
  Else
    inherited SetLink(C, E);
  end;
end;

Procedure SkipListLinks.DrawLinks( W: TCanvas; k,l: Integer);
begin
  DrawLink( W, k,l, _Next_);
  DrawLink( W, k,l, _DnLink_);
end;
(* ----------------------------------------------------------------------- *)

(* ----------------------------------------------------------------------- *)
Constructor TreeLinks.SetUp;
begin
  Parent := NIL;
  Rank   := 0;
  Inherited SetUp;
end;

Function TreeLinks.GetRank: Integer;
begin
  GetRank := Rank;
end;

Procedure TreeLinks.SetRank( R: Integer);
begin
  Rank := R;
end;

Function TreeLinks.GetLink( C: Codes): Elements;
Var
  Here: Elements;
//
  Function AllTheWay( C: Codes): Elements;
  begin
    Here := Self[C];
    AllTheWay := Self;
    While (Here<>NIL) do begin
      AllTheWay := Here;
      Here := Here[C];
    end;
  end;
//
begin
  Case C of
   (* --- Fundamental ---------------------------------------------------- *)
    _Parent_      : GetLink := Parent;
    _FirstChild_  : GetLink := NIL;
    _LastChild_   : GetLink := NIL;
    _NextSibling_ : GetLink := NIL;
    _PrevSibling_ : GetLink := NIL;
   (* --- Derived -------------------------------------------------------- *)
    _Root_        : begin
                      GetLink := Self;
                      While Result[_Parent_]<>NIL do
                        GetLink := Result[_Parent_];
                    end;
//
    _GrandPa_     : Case Self[_Parent_]<>NIL of
                      Ok : GetLink := Self[_Parent_][_Parent_];
                      No : GetLink := NIL;
                    end;
//
    _PreStep_     : Case (Self[_FirstChild_]<>NIL) of
                      Ok: GetLink := Self[_FirstChild_];
                      No: begin
                            Here := Self;
                            While (Here<>NIL) and (Here[_NextSibling_]=NIL) do
                              Here := Here[_Parent_];
                            Case Here<>NIL of
                              Ok: GetLink := Here[_NextSibling_];
                              No: GetLink := NIL;
                            end;
                          end;
                    end;
//
    _PostStep_    : Case (Self[_NextSibling_]<>NIL) of
                      Ok: GetLink := Self[_NextSibling_][_AllDn_];
                      No: GetLink := Self[_Parent_];
                    end;
//
    _AllUp_       : GetLink := AllTheWay( _Parent_);
    _AllDn_       : GetLink := AllTheWay( _FirstChild_);
    _AllLf_       : GetLink := AllTheWay( _Lf_);
    _AllRt_       : GetLink := AllTheWay( _Rt_);
//
  Else
    GetLink := inherited GetLink(C);
  end;
end;

Procedure TreeLinks.SetLink( C: Codes; E: Elements);
begin
  Case C of
    _Parent_ : Parent := E;
  Else
    Inherited SetLink(C,E);
  end;
end;

Procedure TreeLinks.DrawLinks;
begin
  DrawLink( W, k,l, _Parent_);
end;
(* ----------------------------------------------------------------------- *)

(* ----------------------------------------------------------------------- *)
Constructor AnyDegTreeLinks.SetUp;
begin
  Inherited SetUp;
  NextSibling := NIL;
  PrevSibling := NIL;
  Parent      := NIL;
  FirstChild  := NIL;
  LastChild   := NIL;
end;

Function AnyDegTreeLinks.GetLink( C: Codes): Elements;
begin
  Case C of
   (* --- Fundamental ---------------------------------------------------- *)
    _Parent_      : GetLink := Parent;        // More efficient - Not needed
    _FirstChild_  : GetLink := FirstChild;
    _LastChild_   : GetLink := LastChild;
    _NextSibling_ : Case Parent<>NIL of
                      Ok: GetLink := NextSibling;
                      No: GetLink := NIL;
                    end;
    _PrevSibling_ : Case Parent<>NIL of
                      Ok: GetLink := PrevSibling;
                      No: GetLink := NIL;
                    end;
    // Watch out when tree is part of a forest (??)
    _Next_        : Case Parent<>NIL of
                      Ok: GetLink := NIL;
                      No: GetLink := NextSibling;
                    end;
    _Prev_        : Case Parent<>NIL of
                      Ok: GetLink := NIL;
                      No: GetLink := PrevSibling;
                    end;
   (* -------------------------------------------------------------------- *)
  Else
    GetLink := Inherited GetLink(C);
  end;
end;

Procedure AnyDegTreeLinks.SetLink( C: Codes; E: Elements);
/// Watch out that this way we cannot have a forest of binary trees
begin
  Case C of
   (* --- Fundamental ---------------------------------------------------- *)
    _Parent_      : Parent := E;
    _Next_,
    _NextSibling_ : NextSibling := E;
    _Prev_,
    _PrevSibling_ : PrevSibling := E;
    _LastChild_   : LastChild := E;
    _FirstChild_  : FirstChild := E;
   (* -------------------------------------------------------------------- *)
  Else
    Inherited SetLink(C, E);
  end;
end;
(* ----------------------------------------------------------------------- *)

(* ----------------------------------------------------------------------- *)
Constructor BinaryTreeLinks.SetUp;
begin
  Inherited SetUp;
  Parent := NIL;
  Left   := NIL;
  Right  := NIL;
end;

Function BinaryTreeLinks.GetLink(C: Codes): Elements;
Var
  Here: Elements;
begin
  Case C of
   (* --- Fundamental ---------------------------------------------------- *)
    _Parent_      : GetLink := Parent;
    _Lf_          : GetLink := Left;
    _Rt_          : GetLink := Right;
   (* --- Derived -------------------------------------------------------- *)
    _FirstChild_  : Case NotNil( Left, Right) of
                      _Both_,
                      _1st_ : GetLink := Left;
                      _2nd_ : GetLink := Right;
                      _None_: GetLink := NIL;
                    end;
    _LastChild_   : Case NotNil( Left, Right) of
                      _Both_: GetLink := Right;
                      _1st_ : GetLink := Left;
                      _2nd_ : GetLink := Right;
                      _None_: GetLink := NIL;
                    end;
//
    _NextSibling_ : Case (Parent=NIL) of
                      Ok: GetLink := NIL;
                      No: Case Parent[_Lf_]=Self of
                            Ok: GetLink := Parent[_Rt_];
                            No: GetLink := NIL;
                          end;
                    end;
    _PrevSibling_ : Case (Parent=NIL) of
                      Ok: GetLink := NIL;
                      No: Case Parent[_Rt_]=Self of
                            Ok: GetLink := Parent[_Lf_];
                            No: GetLink := NIL;
                          end;
                    end;
//
    _InStep_      : begin
                      Here := Self;
                      Case Here[_Rt_]<>NIL of
                        Ok: GetLink := Here[_Rt_][_AllLf_];
                        No: begin
                              While (Here<>NIL) and (Here.IsWhat=_Rt_) do
                                Here := Here[_Parent_];
                              Case Here<>NIL of
                                Ok: GetLink := Here[_Parent_];
                                No: GetLink := NIL;
                              end;
                            end;
                      end;
                    end;
   (* -------------------------------------------------------------------- *)
  Else
    GetLink := Inherited GetLink(C);
  end;
end;

Procedure BinaryTreeLinks.SetLink( C: Codes; E: Elements);
begin
  Case C of
    _Parent_ : Parent := E;
    _Lf_     : Left   := E;
    _Rt_     : Right  := E;
  Else
    inherited SetLink(C,E);
  end;
end;

Function BinaryTreeLinks.IsWhat: Codes;
begin
  Case Parent=NIL of
    Ok : IsWhat := _Root_;
    No : If Self=Parent[_Lf_] then IsWhat := _Lf_ else IsWhat := _Rt_;
  end;
end;

Procedure BinaryTreeLinks.RotateUp;
Var
  P,G,S,C,Delta  : Elements;
  WhatP,WhatS,WhatC : Codes;
begin
  S     := Self;
  WhatS := S.IsWhat;
  If WhatS<>_Root_ then begin
    P     := S[_Parent_];
    G     := S[_GrandPa_];
    WhatP := P.IsWhat;
    Case WhatS of
      _Lf_ : WhatC := _Rt_;
      _Rt_ : WhatC := _Lf_;
    end;
    C := S[WhatC];

    (* --- ReLinking! --------------------------------------------------- *)
    S[_Parent_] := G;
    If G<>NIL then G[WhatP] := S;

    P[_Parent_] := S;
    S[WhatC] := P;

    If C<>NIL then C[_Parent_] := P;
    P[WhatS] := C;

    (* --- ReWeighting! ------------------------------------------------- *)
    Weight := P.Weight;
    Delta  := S[WhatS];
    Case Delta<>NIL of
      Ok: P.Weight := P.Weight - (Delta.Weight+1);
      No: P.Weight := P.Weight - 1;
    end;
  end;
end;

Procedure BinaryTreeLinks.SwapUpTo( E: Elements);
begin
  If E<>NIL then
    SwapParts( Self, _Info_, E, _Info_ );
end;

Function BinaryTreeLinks.SwapToLeaf: Elements;
Var
  Here,
  This,
  L,R : Elements;
begin
  Here := Self;
  Repeat // in balanced trees this will be repeated only O(1) times
    L := Here[_Lf_];
    R := Here[_Rt_];
    Case NotNIL( L, R) of
      _Both_ : // Just For Fun...
               Case Random(2) of
                 0: This := L[_AllRt_];
                 1: This := R[_AllLf_];
               end;
      _1st_  : This := L[_AllRt_];
      _2nd_  : This := R[_AllLf_];
      _None_ : This := NIL;
    end;
    If This<>NIL then begin
      (This As BinaryTreeLinks).SwapUpTo(Here);
      Here := This;
    end;
  Until (This=NIL);
  SwapToLeaf := Here;
end;
(* ----------------------------------------------------------------------- *)

(* ----------------------------------------------------------------------- *)
Constructor DoubleBinTreeLinks.SetUp( E1, E2: Elements);
begin
  inherited SetUp;
  Info := E1;
  HeapInfo := E2;
  CurrentInfo := _Null_;
end;

Destructor DoubleBinTreeLinks.Clear;
begin
 (* --- Allow the system's garbage collector to take effect -------------- *)
  Info := NIL;
  HeapInfo := NIL;         
end;

Function DoubleBinTreeLinks.GetLink( C: Codes): Elements;
begin
  Case C of
    _HeapInfo_ : begin
                   GetLink := HeapInfo;
                  (* --- Color relevant element as 'USED' --------- *)
                   If Result<>NIL then Result.Color := _UsedColor_;
                  (* ---------------------------------------------- *)
                 end;
  Else
    GetLink := inherited GetLink(C);
  end;
end;

Procedure DoubleBinTreeLinks.SetLink( C: Codes; E: Elements);
begin
  Case C of
    _HeapInfo_ : begin
                  (* --- Color relevant element as 'USED' --------- *)
                   If E<>NIL then E.Color := _UsedColor_;
                  (* ---------------------------------------------- *)
                   HeapInfo := E;
                 end;
  Else
    inherited SetLink(C,E);
  end;
end;
(* ----------------------------------------------------------------------- *)

(* ----------------------------------------------------------------------- *)
Constructor PrSearchLinks.SetUp( E: Elements; O1, O2: OrderFunctions);
begin
  inherited SetUp(E,E);
  HasCopy   := Ok;
  Order     := O1;
  HeapOrder := O2;
end;

Procedure PrSearchLinks.SwapUpTo( E: Elements);
Var
  Here  : Elements;
  Found : Boolean;
begin
  If E<>Self then begin
    Found := Not HasCopy;
    If HasCopy then begin
      Here  := Self;
      While Not Found and (Here<>E) do begin
        Found := (Self[_Info_]=Here[_HeapInfo_]);
        If Not Found then
          Here := Here[_Parent_];
      end;
      If Found then
        PrSearchLinks(Here).PullUp;
    end;
    (E As PrSearchLinks).HasCopy := HasCopy and Not Found;
    E[_Info_] := Info;
  end;
end;

Procedure PrSearchLinks.RotateUp;
Var
  P: PrSearchLinks;
begin
  P := Self[_Parent_] As PrSearchLinks;
  P.PushDn(NIL);
    PushDn(NIL);
  Inherited RotateUp;
  P.PullUp;
    PullUp;
end;

Procedure PrSearchLinks.PushDn( This: Elements);
begin
  Swap( This, HeapInfo);
  If This<>NIL then
    (* --- Follow Path Of Main Info -------------------------------------- *)
    Case Order( This, Info ) of
      _Less_   : With Self[_Lf_] As PrSearchLinks do PushDn(This);
      _Equal_  : HasCopy  := No;
      _Greater_: With Self[_Rt_] As PrSearchLinks do PushDn(This);
    end;
end;

Procedure PrSearchLinks.PullUp;
Var
  L,R,
  Here: PrSearchLinks;
  More: Elements;
//
  Function Offer( From: Elements): Elements;
  begin
    Case From<>NIL of
      Ok: With (From As PrSearchLinks) do Offer := HeapInfo;
      No: Offer := NIL;
    end;
  end;
//
begin
  L := (Self[_Lf_] As PrSearchLinks);
  R := (Self[_Rt_] As PrSearchLinks);
  Case Minimum( Offer(L), Offer(R), HeapOrder ) of
    _1st_ : Here := L;
    _2nd_ : Here := R;
    _None_: Here := NIL;
  end;
  Case HasCopy of
    Ok: More := NIL;
    No: More := Info;
  end;
  Case Minimum( More, Offer(Here), HeapOrder ) of
    _1st_ : begin HeapInfo := Info;        HasCopy  := Ok; end;
    _2nd_ : begin HeapInfo := Offer(Here); Here.PullUp;    end;
    _None_: HeapInfo := NIL;
  end;
end;

Procedure PrSearchLinks.Draw( W: TCanvas; k,l: Integer );
begin
  Marked := HasCopy;
  inherited Draw(W,k,l);
  (* --- Content --------------------------------------------------------- *)
  If HeapInfo<>NIL then HeapInfo.Draw( W, k+X, l+Y+CellHeight);
end;
(* ----------------------------------------------------------------------- *)

(* ----------------------------------------------------------------------- *)
Constructor BTreeLinks.SetUp( MaxDegree: Integer);
Var
  k : Integer;
begin
  inherited SetUp;
  Size   := 0;
  SetLength( Data, MaxDegree);                // 1..MaxDegree-1 to be used...
  // Here we ignore Oth position for clarity: 1st element in 1st position
  For k := 1 to MaxDegree-1 do Datum[k] := NIL;
  CurrentData := (0);                         // Conventional value
  //
  Degree := Size+1;                           // Watch out...
  SetLength( Child, MaxDegree+1);             // 1..MaxDegree to be used...
  // Here we ignore Oth position for clarity: 1st element in 1st position
  For k := 1 to MaxDegree do Child[k] := NIL;
end;

Destructor BTreeLinks.Clear;
var
  k: Integer;
begin
 (* --- Allow the system's garbage collector to take effect -------------- *)
  For k := 1 to Size   do Datum[k] := NIL;
  For k := 1 to Degree do Child[k] := NIL;
  Data  := NIL;
  Child := NIL;
end;

Function BTreeLinks.Current: Elements;
begin
  Current := NIL;
  If (1<=CurrentData) and (CurrentData<=Size) then
    Current := Datum[CurrentData];
end;

Procedure BTreeLinks.SetData(k: Integer; E: Elements);
begin
 (* --- Color relevant element as 'USED' --- *)
  If E<>NIL then E.Color := _UsedColor_;
 (* ---------------------------------------- *)
  Data[k] := E;
end;

Function BTreeLinks.GetData(k: Integer): Elements;
begin
  GetData := Data[k];
 (* --- Color relevant element as 'USED' --------- *)
  If Result<>NIL then Result.Color := _UsedColor_;
 (* ---------------------------------------------- *)
end;

Constructor BTreeLinks.Meld( E: Elements; L,R: BTreeLinks);
Var
  k: Integer;
begin
  inherited SetUp;
  SetLength( Data, Length(L.Data));
  SetLength( Child, Length(L.Child));

  (*  --- Meld Data ------------------------------------------------------ *)
  For k := 1 to L.Size do Datum[k] := L.Datum[k];
  Datum[L.Size+1] := E;
  For k := 1 to R.Size do Datum[L.Size+1+k] := R.Datum[k];
  Size := L.Size + 1 + R.Size;

  (* --- Meld Children --------------------------------------------------- *)
  For k := 1 to L.Degree do Child[k] := L.Child[k];
  For k := 1 to R.Degree do Child[L.Degree+k] := R.Child[k];
  Degree := L.Degree + R.Degree;

  (* --- Adjust Parent Links --------------------------------------------- *)
  For k := 1 to Degree do
    If Child[k]<>NIL then Child[k].Parent := Self;

  (* --- Clear Old Nodes ------------------------------------------------- *)
  L.Clear;
  R.Clear;
end;

Procedure BTreeLinks.Split( D: Integer; Var E: Elements; Var L,R: BTreeLinks);
// Node is Split to first D Data, Datum[D] and Last (Size-D) Data
Var
  k,
  MaxDegree : Integer;
begin
   MaxDegree := Length(Data);                       // See BTreeLinks.SetUp
 (* --- LEFT Part -------------------------------------------------------- *)
  L := BTreeLinks.SetUp( MaxDegree );

  L.Size := D-1;                                    // Transfer Data
  For k := 1 to L.Size do L.Datum[k] := Datum[k];

  L.Degree := L.Size+1;                             // Transfer Children
  For k := 1 to L.Degree do L.Child[k] := Child[k];

  For k := 1 to L.Degree do                         // Adjust Parent Links
    If L.Child[k]<>NIL then L.Child[k].Parent := L;

 (* --- Separating Element ----------------------------------------------- *)
  E := Datum[D];

 (* --- RIGHT Part ------------------------------------------------------- *)
  R := BTreeLinks.SetUp( MaxDegree );

  R.Size := Size-D;                                 // Transfer Data
  For k := 1 to R.Size do R.Datum[k] := Datum[D+k];

  R.Degree := R.Size+1;                             // Transfer Children
  For k := 1 to R.Degree do R.Child[k] := Child[D+k];

  For k := 1 to R.Degree do                         // Adjust Parent Links
    If R.Child[k]<>NIL then R.Child[k].Parent := R;
end;

Procedure BTreeLinks.InsData( P: Integer; E: Elements; L,R: BTreeLinks);
// P-th subtree is replaced by subtrees L & R w.r. to element E
Var
  k: Integer;
begin
  (* --- Insert Element -------------------------------------------------- *)
  For k := Size DownTo P do Datum[k+1] := Datum[k];
  Datum[P] := E;
  Inc(Size);

  (* --- Insert Children ------------------------------------------------- *)
  For k := Degree DownTo P do Child[k+1] := Child[k];
  Child[P]   := L;
  Child[P+1] := R;
  Degree := Size+1;

  (* --- Adjust Parent Links --------------------------------------------- *)
  If L<>NIL then L.Parent := Self;
  If R<>NIL then R.Parent := Self;
end;

Procedure BTreeLinks.DelData( P: Integer; B: BTreeLinks);
// P-th Element is Deleted, "left" and "right" subtrees replaced by B
Var
  k : Integer;
begin
  (* --- Insert Element -------------------------------------------------- *)
  // Insert Element
  Dec(Size);
  For k := P to Size do Datum[k] := Datum[k+1];

  (* --- Insert Children ------------------------------------------------- *)
  Degree := Size+1;
  Child[P] := B;
  For k := P+1 to Degree do Child[k] := Child[k+1];

  (* --- Adjust Parent Links --------------------------------------------- *)
  If B<>NIL then B.Parent := Self;
end;

Procedure BTreeLinks.RotateUp( Which: Codes; H: Integer);
Var
  E     : Elements;
  C,L,R : BTreeLinks;
begin
  L := Child[H];
  R := Child[H+1];
  Case Which of
    _Lf_:
    begin
      E := Datum[H];                             // Element to Rotate Down
      With L do begin
        C := Child[Degree];                      // Last Child of L
        Self.Datum[H] := Datum[Size];            // Last Datum - Rotated Up
        DelData( Size, Child[Degree-1]);
      end;                                       // Delete Last Data/Child
      With R do
        InsData( 1, E, C, Child[1]);             // E As First of R
    end;
    _Rt_:
    begin
      E := Datum[H];                             // Element to Rotate Down
      With R do begin
        C := R.Child[1];                         // First Child of R
        Self.Datum[H] := Datum[1];               // First Datum - Rotated Up
        DelData( 1, Child[2]);                   // Delete First Data/Child
      end;
      With L do
        InsData( Size+1, E, Child[Degree], C);   // E As Last of L
    end;
  end;
end;

Function BTreeLinks.LookFor( E: Elements; O: OrderFunctions): Integer;
// Returns the Index of the link one should follow when searching for E
Var
  C : Integer;
begin
  C := 1;
  While (C<=Size) and Less( Datum[C], E, O) do
    Inc(C);
  LookFor := C;
end;

Procedure BTreeLinks.Draw( W: TCanvas; k,l: Integer);
var
  i : Integer;
begin
  If Size>0 then inherited Draw(W,k,l);
  (* --- Content --------------------------------------------------------- *)
  For i := 1 to Size do
    If Datum[i]<>NIL then Datum[i].Draw( W, k+X,l+Y+(i-1)*CellHeight);
end;

Function BTreeLinks.GetLink( C: Codes ): Elements;
Var
  k    : Integer;
begin
  Case C of
   (* --- Fundamental ---------------------------------------------------- *)
    _Parent_      : GetLink := Parent;
    _Lf_,
    _FirstChild_  : Case Degree<2 of
                      Ok: GetLink := NIL;
                      No: GetLink := Child[1];
                    end;
    _Rt_,
    _LastChild_   : Case Degree<2 of
                      Ok: GetLink := NIL;
                      No: GetLink := Child[Degree];
                    end;
    _NextSibling_ : Case Parent=NIL of
                      Ok: GetLink := NIL;
                      No: With Parent As BTreeLinks do begin
                            k := 1;
                            While (k<=Degree) and (Child[k]<>Self) do Inc(k);
                            Case (k+1)<=Degree of
                              Ok: Result := Child[k+1];
                              No: Result := NIL;
                            end;
                         end;
                    end;
    _PrevSibling_ : Case Parent=NIL of
                      Ok: GetLink := NIL;
                      No: With Parent As BTreeLinks do begin
                            k := 1;
                            While (k<=Degree) and (Child[k]<>Self) do Inc(k);
                            Case (k>1) of
                              Ok: Result := Child[k-1];
                              No: Result := NIL;
                            end;
                         end;
                    end;

   (* -------------------------------------------------------------------- *)
  Else
    GetLink := inherited GetLink(C);
  end;
end;

(* ======================================================================= *)

begin
end.


