(* ----------------------------------------------------------------------- *)
UNIT SSCHEMES_;
(* ----------------------------------------------------------------------- *)

INTERFACE
(* ----------------------------------------------------------------------- *)

USES

  SysUtils,
  Graphics,

  CONSTANTS_,
  DRAW_,
  ELEMENTS_,
  LINKS_,
  REPORTING_;
(* ----------------------------------------------------------------------- *)

TYPE

  StorageSchemes = Class( Elements )
   (* --- Basic Description ---------------------------------------------- *)
                                                                 // Abstract
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp;
    Function    LinkFor( E: Elements): Elements;                    Virtual;
   (* --- Navigation ----------------------------------------------------- *)
    Function    Current: Elements;                                 Override;
    Function    Go( C: Codes): Elements;                            Virtual;
    Function    GoNmr( N: Integer): Elements;                       Virtual;
   (* --- Information ---------------------------------------------------- *)
    Function    NmrOfElements: Integer;                             Virtual;
    Function    MaxCapacity: Integer;                               Virtual;
    Function    Enumerate(F,S: Codes; From,Step: Integer): Integer; Virtual;
    Procedure   ColorAll( C: Codes);
   (* --- Visualization -------------------------------------------------- *)
    Procedure   Arrange( k,l, Dk,Dl: Integer);                      Virtual;
    Procedure   Frame( Var x1,y1, x2,y2: Integer);                  Virtual;
   (* -------------------------------------------------------------------- *)
  end;

  SequentialSchemes = Class(StorageSchemes)
   (* --- Basic Description ---------------------------------------------- *)
    CurrentNmr : Integer;  // Index of current element
    MaximumNmr : Integer;  // Maximum number of elements
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp;
   (* --- Navigation ----------------------------------------------------- *)
    Function    Current: Elements;                                 Override;
    Function    Go( C: Codes): Elements;                           Override;
   (* --- Information ---------------------------------------------------- *)
    Function    MaxCapacity: Integer;                              Override;
   (* --- Handling Operations -------------------------------------------- *)
    Function    Get( i: Integer): Elements;                         Virtual;
    Procedure   Put( i: Integer; E: Elements);                      Virtual;
    Property    Item[i: Integer]: Elements               Read Get Write Put;
    Function    Location( i: Integer): Links;                       Virtual;
    Procedure   SwapItems( i,j: Integer);
   (* --- Visualization -------------------------------------------------- *)
    Procedure   Arrange( k,l, Dk,Dl: Integer);                     Override;
    Procedure   Draw( W: TCanvas; k,l: Integer);                   Override;
    Procedure   DrawLinks( W: TCanvas; k,l: Integer);              Override;
   (* -------------------------------------------------------------------- *)
  end;

  ArrayOfLinks = Array Of Links;

  Tables = Class(SequentialSchemes)
   (* --- Basic Description ---------------------------------------------- *)
    Table  : ArrayOfLinks;     //For Making "Arrange" and "Marking" Possible
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( Max: Integer);
    Destructor  Clear;                                             Override;
   (* --- Handling Operations -------------------------------------------- *)
    Function    Location(i: Integer): Links;                       Override;
    Procedure   ReSize( NewSize: Integer);                          Virtual;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   DrawLinks( W: TCanvas; k,l: Integer);              Override;
   (* -------------------------------------------------------------------- *)
  end;

  DynamicTables = Class(Tables)
   (* --- Information ---------------------------------------------------- *)
    Function    MaxCapacity: Integer;                              Override;
   (* --- Handling Operations -------------------------------------------- *)
    Procedure   ReSize( NewSize: Integer);                         Override;
   (* -------------------------------------------------------------------- *)
  end;

  CircularTables = Class(Tables)
   (* --- Basic Description ---------------------------------------------- *)
    First : Integer;
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( Max: Integer);
   (* --- Handling Operations -------------------------------------------- *)
    Function    Location(i: Integer): Links;                       Override;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   Arrange( k,l, Dk,Dl: Integer);                     Override;
   (* -------------------------------------------------------------------- *)
  end;

  LinkedSchemes = Class( StorageSchemes)
   (* --- Basic Description ---------------------------------------------- *)
    CurrentLink : Elements;  // Link to the Current Element
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp;
   (* --- Navigation ----------------------------------------------------- *)
    Function    Current: Elements;                                 Override;
   (* --- Information ---------------------------------------------------- *)
    Function    MaxCapacity: Integer;                              Override;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   Arrange( k,l, Dk,Dl: Integer);                     Override;
    Procedure   Draw( W: TCanvas; k,l: Integer);                   Override;
   (* -------------------------------------------------------------------- *)
  end;

  Lists = Class( LinkedSchemes)
   (* --- Basic Description ---------------------------------------------- *)
    Head,               // First element in the list
    Tail   : Elements;  // Last element in the list
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp;
    Destructor  Clear;                                             Override;
    Function    LinkFor( E: Elements): Elements;                   Override;
   (* --- Navigation ----------------------------------------------------- *)
    Function    Go( C: Codes): Elements;                           Override;
   (* --- Handling Operations -------------------------------------------- *)
    Procedure   Link( This: Elements; How: Codes; Where: Elements);
    Procedure   UnLink( E: Elements);
    Procedure   Merge( L,R: Lists; Order: OrderFunctions);
    Procedure   Split( Var L,R: Lists);
    Procedure   Append( L: Lists);
   (* --- Visualization -------------------------------------------------- *)
    Procedure   DrawLinks( W: TCanvas; k,l: Integer);              Override;
   (* -------------------------------------------------------------------- *)
  end;

  RootedTrees = Class(LinkedSchemes)
   (* --- Basic Description ---------------------------------------------- *)
    Root : Elements;  // The root of the tree
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp;
    Destructor  Clear;                                             Override;
    Function    LinkFor( E: Elements): Elements;                   Override;
   (* --- Navigation ----------------------------------------------------- *)
    Function    Go( C: Codes): Elements;                           Override;
   (* --- Information ---------------------------------------------------- *)
    Function    Enumerate(F,S: Codes; From,Step: Integer): Integer;Override;
   (* --- Handling Operations -------------------------------------------- *)
    Procedure   Link( This: Elements; How: Codes; Where: Elements);
    Procedure   UnLink( E: Elements);
   (* --- Visualization -------------------------------------------------- *)
    Procedure   Arrange( k,l,Dk,Dl: Integer);                      Override;
    Procedure   DrawLinks( W: TCanvas; k,l: Integer);              Override;
   (* -------------------------------------------------------------------- *)
  end;

  BinaryTrees = Class( RootedTrees )
   (* --- Construction --------------------------------------------------- *)
    Function    LinkFor( E: Elements): Elements;                   Override;
   (* --- Navigation ----------------------------------------------------- *)
    Function    GoBinary( N: Integer): Elements;
   (* --- Handling Operations -------------------------------------------- *)
    Procedure   Link( This: Elements; How: Codes; Where: Elements);
    Procedure   UnLink( E: Elements);
    Procedure   InsertLeaf( E: Elements; How: Codes; P: Elements );
    Procedure   DeleteLeaf( E: Elements);
   (* --- Balance Operations --------------------------------------------- *)
    Procedure   RotateUp( E: Elements);
    Procedure   DoubleRotateUp( E: Elements);
    Procedure   SplayUp( E: Elements );
    Procedure   InsBalance( P: Elements; Which: Codes);             Virtual;
    Procedure   DelBalance( P: Elements; Which: Codes);             Virtual;
    Procedure   FindBalance( P: Elements);                          Virtual;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   Arrange( k,l,Dk,Dl: Integer);                      Override;
   (* -------------------------------------------------------------------- *)
  end;

  HalfBalancedTrees = Class(BinaryTrees)
   (* --- Balance Operations --------------------------------------------- *)
    Procedure   InsBalance( P: Elements; Which: Codes);            Override;
    Procedure   DelBalance( P: Elements; Which: Codes);            Override;
   (* -------------------------------------------------------------------- *)
  end;

  AVLtrees = Class(BinaryTrees)
   (* --- Balance Operations --------------------------------------------- *)
    Procedure   InsBalance( P: Elements; Which: Codes);            Override;
    Procedure   DelBalance( P: Elements; Which: Codes);            Override;
   (* -------------------------------------------------------------------- *)
  end;

  SplayTrees = Class(BinaryTrees)
   (* --- Balance Operations --------------------------------------------- *)
    Procedure   InsBalance( P: Elements; Which: Codes);            Override;
    Procedure   DelBalance( P: Elements; Which: Codes);            Override;
    Procedure   FindBalance( P: Elements);                         Override;
   (* -------------------------------------------------------------------- *)
  end;

  RotateToRootTrees = Class(BinaryTrees)
   (* --- Balance Operations --------------------------------------------- *)
    Procedure   InsBalance( P: Elements; Which: Codes);            Override;
    Procedure   DelBalance( P: Elements; Which: Codes);            Override;
    Procedure   FindBalance( P: Elements);                         Override;
   (* -------------------------------------------------------------------- *)
  end;

  RandomizedTrees = Class(BinaryTrees)
   (* --- Balance Operations --------------------------------------------- *)
    Procedure   InsBalance( P: Elements; Which: Codes);            Override;
   (* -------------------------------------------------------------------- *)
  end;

  RandomizedSearchTrees = Class( BinaryTrees )
   (* --- Balance Operations --------------------------------------------- *)
    Procedure   InsBalance( P: Elements; Which: Codes);            Override;
   (* -------------------------------------------------------------------- *)
  end;

  AnyDegTrees = Class(RootedTrees)
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp;
    Function    LinkFor( E: Elements): Elements;                   Override;
   (* --- Handling Operations -------------------------------------------- *)
    Procedure   Link( This: Elements; How: Codes; Where: Elements);
    Procedure   UnLink( E: Elements);
   (* -------------------------------------------------------------------- *)
  end;

  Btrees = Class( LinkedSchemes)
   (* --- Basic Description ---------------------------------------------- *)
    Tree     : RootedTrees;  // A TREE of TABLES
    MinSize,                 // minimum number of elements per node
    MaxSize : Integer;       // maximum number of elements per node
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( MinS, MaxS: Integer);
    Destructor  Clear;                                             Override;
    Function    LinkFor( E: Elements): Elements;                   Override;
   (* --- Navigation ----------------------------------------------------- *)
    Function    Go(C: Codes): Elements;                            Override;
    Function    Current: Elements;                                 Override;
    Function    CurrentNode: BTreeLinks;
   (* --- Handling Operations -------------------------------------------- *)
    Procedure   StepIfHighSize( var B: BTreeLinks; H: Integer);
    Procedure   StepIfLowSize( var B: BTreeLinks; H: Integer);
   (* --- Visualization -------------------------------------------------- *)
    Procedure   Arrange( k,l,Dk,Dl: Integer);                      Override;
    Procedure   Draw( W: TCanvas; k,l: Integer);                   Override;
    Procedure   DrawLinks( W: TCanvas; k,l: Integer);              Override;
    Procedure   Frame( Var x1,y1, x2,y2: Integer);                 Override;
   (* -------------------------------------------------------------------- *)
  end;

  SkipLists = Class( LinkedSchemes)
   (* --- Basic Description ---------------------------------------------- *)
    ListOfLevels : Lists;    // A LIST of LISTS - each being a "level"
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp;
    Destructor  Clear;                                             Override;
    Function    LinkFor( E: Elements): Elements;                   Override;
   (* --- Navigation ----------------------------------------------------- *)
    Function    Go( C: Codes): Elements;                           Override;
    Function    Current: Elements;                                 Override;
    Function    CurrentList: Lists;
   (* --- Handling Operations -------------------------------------------- *)
    Procedure   Link( This, Where: Elements; Levels: Integer);
    Procedure   DeleteLinks( L: Elements);
    Procedure   AdjustHeight( Levels: Integer);
   (* --- Information ---------------------------------------------------- *)
    Function    Enumerate(F,S: Codes; From,Step: Integer): Integer;Override;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   Arrange( k,l,Dk,Dl: Integer);                      Override;
   (* -------------------------------------------------------------------- *)
  end;

  Forests = Class(LinkedSchemes)
   (* --- Basic Description ---------------------------------------------- *)
    ListOfTrees : Lists;     // A LIST Of TREES
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp;
    Function    LinkFor( E: Elements): Elements;                   Override;
    Destructor  Clear;                                             Override;
   (* --- Navigation ----------------------------------------------------- *)
    Function    Go( C: Codes): Elements;                           Override;
    Function    Current: Elements;                                 Override;
    Function    CurrentTree: AnyDegTrees;
    Function    RootOf( E: Elements): Elements;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   Arrange( k,l, Dk, Dl: Integer);                    Override;
   (* -------------------------------------------------------------------- *)
  end;

  Buckets = Class(Tables)
   (* --- Basic Description ---------------------------------------------- *)
                             // A TABLE of LISTS
   (* --- Construction --------------------------------------------------- *)
    Constructor SetUp( N: Integer);
    Destructor  Clear;                                             Override;
   (* --- Navigation ----------------------------------------------------- *)
    Function    Go(C: Codes): Elements;                            Override;
    Function    Current: Elements;                                 Override;
    Function    CurrentList: Lists;
    Function    List(Nmr: Integer): Lists;
   (* --- Information ---------------------------------------------------- *)
    Function    MaxCapacity: Integer;                              Override;
   (* --- Visualization -------------------------------------------------- *)
    Procedure   Arrange( k,l, Dk, Dl: Integer);                    Override;
   (* -------------------------------------------------------------------- *)
  end;
(* ----------------------------------------------------------------------- *)

IMPLEMENTATION
(* ----------------------------------------------------------------------- *)

(* --- STORAGE SCHEMES --------------------------------------------------- *)
Constructor StorageSchemes.SetUp;
begin
end;

Function StorageSchemes.LinkFor( E: Elements): Elements;
begin
  LinkFor := E;
end;

Function StorageSchemes.MaxCapacity: Integer;
begin
  MaxCapacity := 0;
end;

Procedure StorageSchemes.Arrange( k,l,Dk,Dl: Integer);
begin
end;

Procedure StorageSchemes.Frame( Var x1,y1, x2,y2: Integer);
var
  Here : Elements;
begin
  Here := Go(_First_);
  Case Here<>NIL of
    Ok: begin x1 := Here.x; y1 := Here.y; x2 := x1; y2 := y1; end;
    No: begin x1 := 0; y1 := 0; x2 := 0; y2 := 0; end;
  end;
  Here := Go(_Step_);
  While (Here<>NIL) do begin
    If Here.x < x1 then x1 := Here.x;
    If Here.x > x2 then x2 := Here.x;
    If Here.y < y1 then y1 := Here.y;
    If Here.y > y2 then y2 := Here.y;
    Here := Go(_Step_);
  end;
end;

Procedure StorageSchemes.ColorAll( C: Codes);
Var
  Here : Elements;
begin
  Here := Go(_First_);
  While Here<>NIL do begin
    If Here[_Info_]<>NIL then
      Here[_Info_].Color := C;
    Here := Go(_Step_);
  end;
end;

Function StorageSchemes.Current: Elements;
begin
  Current := NIL;
end;

Function StorageSchemes.NmrOfElements: Integer;
Var
  Count : Integer;
  Here  : Elements;
begin
  Count := 0;
  Here  := Go(_First_);
  While (Here<>NIL) do begin
    Inc(Count);
    Here := Go(_Step_);
  end;
  NmrOfElements := Count;
end;

Function StorageSchemes.Go(C: Codes): Elements;
begin
  Case C of
    _Any_ : Go := GoNmr(1+Random(NmrOfElements))
  Else
    Go := NIL;
  end;
end;

Function StorageSchemes.GoNmr( N: Integer): Elements;
var
  Count : Integer;
begin
  Count := 1;
  Go(_First_);
  While (Count<N) and (Current<>NIL) do begin
    Inc(Count);
    Go(_Step_);
  end;
  GoNmr := Current;
end;

Function StorageSchemes.Enumerate( F, S: Codes; From,Step: Integer): Integer;
Var
  Here : Elements;
  Count: Integer;
begin
  Count := From;
  Here  := Go(F);
  While Here<>NIL do begin
    Inc(Count, Step);
    If Current<>NIL then Current.Number := Count;
    Here := Go(S);
  end;
  Enumerate := Count;
end;
(* ----------------------------------------------------------------------- *)

(* --- LINKED SCHEMES ---------------------------------------------------- *)
Constructor LinkedSchemes.SetUp;
begin
  inherited SetUp;
  CurrentLink := NIL;
end;

Function LinkedSchemes.Current: Elements;
begin
  Case CurrentLink<>NIL of
    Ok: Current := CurrentLink.Current;
    No: Current := NIL;
  end;
end;

Function LinkedSchemes.MaxCapacity: Integer;
begin
(* --- Restricted only for visualization purposes ------------------------ *)
   MaxCapacity := _MaximumSize_;
(* ----------------------------------------------------------------------- *)
end;

Procedure LinkedSchemes.Arrange( k,l,Dk,Dl: Integer);
var
  Here  : Elements;
begin
  Here := Go(_First_);
  While (Here<>NIL) do begin
    Here.x := k;
    Here.y := l;
    Inc(k,Dk);
    Inc(l,Dl);
    Here := Go(_Step_);
  end;
end;

Procedure LinkedSchemes.Draw( W: TCanvas; k,l: Integer);
Var
  Here : Elements;
begin
  Here := Go(_First_);
  While (Here<>NIL) do begin
    Here.DrawLinks(W, k+(CellWidth Div 2),l+(CellHeight Div 2));
    Here := Go(_Step_);
  end;
  Here := Go(_First_);
  While (Here<>NIL) do begin
    Here.Draw(W,k,l);
    Here := Go(_Step_);
  end;
end;
(* ----------------------------------------------------------------------- *)

(* --- SEQUENTIAL SCHEMES ------------------------------------------------ *)
Constructor SequentialSchemes.SetUp;
begin
  inherited SetUp;
  CurrentNmr := 0;
end;

Function SequentialSchemes.MaxCapacity: Integer;
begin
(* --- Restricted only for visualization purposes ------------------------ *)
   MaxCapacity := MaximumNmr;
(* ----------------------------------------------------------------------- *)
end;

Procedure SequentialSchemes.Arrange( k,l,Dk,Dl: Integer);
var
  Here : Elements;
  i    : Integer;
begin
  For i := 1 to MaximumNmr do begin
    Here := Location(i);
    If Here<>NIL then begin
      Here.x := k + (i-1)*Dk;
      Here.y := l;
    end;
  end;
end;

Procedure SequentialSchemes.DrawLinks( W: TCanvas; k,l: Integer);
begin
end;

Procedure SequentialSchemes.Draw( W: TCanvas; k,l: Integer);
Var
  i : Integer;
begin
  For i := 1 to MaximumNmr do Location(i).Draw(W,k,l);
end;

Function SequentialSchemes.Current: Elements;
begin
  Case (1<=CurrentNmr) and (CurrentNmr<=MaximumNmr) of
    Ok: Case Item[CurrentNmr]<>NIL of
          Ok: Current := Item[CurrentNmr].Current;
          No: Current := NIL;
        end;
    No: Current := NIL;
  end;
end;

Function SequentialSchemes.Go( C: Codes): Elements;
//
  Function GetNotNIL( C: Integer): Integer;
  begin
    While (C<=MaximumNmr) and (Item[C]=NIL) do Inc(C);
    GetNotNIL := C;
  end;
//
begin
  Case C of
    _First_ : CurrentNmr := GetNotNIL(1);
    _Step_  : CurrentNmr := GetNotNIL(CurrentNmr+1);
  Else
    CurrentNmr := 0;
  end;
  Case (1<=CurrentNmr) and (CurrentNmr<=MaximumNmr) of
    Ok: Go := Location(CurrentNmr);
    No: Go := NIL;
  end;
end;

Function SequentialSchemes.Location(i: Integer): Links;
begin
  Location := NIL;
end;

Function SequentialSchemes.Get( i: Integer): Elements;
Var
  Here : Links;
begin
  Here := Location(i);
  Case Here<>NIL of
    Ok: Get := Here.Info;
    No: Get := NIL;
  end;
 (* --- Color relevant element as 'USED' --------- *)
  If Result<>NIL then Result.Color := _UsedColor_;
 (* ---------------------------------------------- *)
end;

Procedure SequentialSchemes.Put( i: Integer; E: Elements);
Var
  Here : Links;
begin
 (* --- Color relevant element as 'USED' --------- *)
  If E<>NIL then E.Color := _UsedColor_;
 (* ---------------------------------------------- *)
  Here := Location(i);
  Case Here<>NIL of
    Ok: Here.Info := E;
    No: begin (* not possible *) end;
  end;
end;

Procedure SequentialSchemes.SwapItems(i,j: Integer);
Var
  E : Elements;
begin
  E       := Item[i];
  Item[i] := Item[j];
  Item[j] := E;
end;
(* ----------------------------------------------------------------------- *)

(* --- TABLES ------------------------------------------------------------ *)
Constructor Tables.SetUp( Max: Integer);
Var
  k : Integer;
begin
  inherited SetUp;
  MaximumNmr := Max;
  SetLength( Table, MaximumNmr); (**)
  For k := 0 to MaximumNmr-1 do Table[k] := Links.SetUp;
end;

Destructor Tables.Clear;
Var
  k : Integer;
begin
  For k := 0 to MaximumNmr-1 do Table[k] := NIL;
  Table := NIL;
end;

Function Tables.Location(i: Integer): Links;
begin
  Case (1<=i) and (i<=MaximumNmr) of
    Ok: Location := Table[i-1];
    No: Location := NIL;
  end;
end;

Procedure Tables.ReSize( NewSize: Integer);
begin
 (* --- fails for static... --- *)
end;

Procedure Tables.DrawLinks( W: TCanvas; k,l: Integer);
Var
  i : Integer;
begin
  k := k + (CellWidth  Div 2);
  l := l + (CellHeight Div 2);
  For i := 1 to MaximumNmr do
    With Location(i) do begin
      If i>1 then
        DrawLine( W, k+X, l+Y,
                     k+Location(i-1).X, l+Location(i-1).Y, _LinkColor_);
      If Location(i).Info=NIL then
        DrawFrame( W, k+X-3, l+Y-3, k+X+3, l+Y+3, _FrameColor_);
    end;
end;
(* ----------------------------------------------------------------------- *)

(* --- DYNAMIC TABLES ---------------------------------------------------- *)
Function DynamicTables.MaxCapacity: Integer;
begin
(* --- since they are dynamic.. /restricted for visualization purposes --- *)
   MaxCapacity := _MaximumSize_;
(* ----------------------------------------------------------------------- *)
end;

Procedure DynamicTables.ReSize( NewSize : Integer);
Var
  i,
  OldSize  : Integer;
  OldLinks,
  NewLinks : ArrayOfLinks;
begin
 (* --- succeeds for dynamic tables -------------------------------------- *)
  OldLinks := Table;
  OldSize  := MaximumNmr;

  SetLength( NewLinks, NewSize);
  MaximumNmr := NewSize;

 (* --- Case expanded ------- *)
  If NewSize>=OldSize then begin
    For i := 1         to OldSize do NewLinks[i-1] := OldLinks[i-1];
    For i := OldSize+1 to NewSize do NewLinks[i-1] := Links.SetUp;
  end
  Else
 (* --- Case contracted ----- *)
  If NewSize<OldSize then begin
    For i := 1 to NewSize do NewLinks[i-1] := OldLinks[i-1];
    For i := NewSize+1 to OldSize do OldLinks[i-1].Clear;
  end;

 (* --- switch and clear old array --------------------------------------- *)
  Table    := NewLinks;
  OldLinks := NIL;
end;
(* ----------------------------------------------------------------------- *)

(* --- CIRCULAR TABLES --------------------------------------------------- *)
Constructor CircularTables.SetUp( Max: Integer);
begin
  inherited SetUp(Max);
  First := 0;
end;

Function CircularTables.Location(i: Integer): Links;
begin
  Location := Inherited Location((First+i) Mod MaximumNmr + 1);
end;

procedure CircularTables.Arrange(k, l, Dk, Dl: Integer);
Var
  Radius : Real;
  i      : Integer;
begin
  Radius := 200.0;
  For i := 0 to MaximumNmr-1 do
    If Table[i]<>NIL then begin
      Table[i].x := Trunc(400.0+1.5*Radius*Cos(2.0*Pi*(i/MaximumNmr)));
      Table[i].y := Trunc(300.0+    Radius*Sin(2.0*Pi*(i/MaximumNmr)));
    end;
end;
(* ----------------------------------------------------------------------- *)

(* --- TABLES OF LISTS --------------------------------------------------- *)
Constructor Buckets.SetUp( N: Integer);
Var
  i: Integer;
begin
  inherited SetUp(N);  // Just as an example
  For i := 1 to MaximumNmr do
    Item[i] := Lists.SetUp;
end;

Destructor Buckets.Clear;
Var
  i: Integer;
begin
  For i := 1 to MaximumNmr do
    If Item[i]<>NIL then Item[i].Clear;
  inherited Clear;
end;

Function Buckets.Go( C: Codes): Elements;
//
  Function TryStep( N: Integer; C: Codes): Elements;
  begin
    CurrentNmr := N;
    TryStep := (Item[CurrentNmr] As Lists).Go(C);
    If (Result=NIL) then
      If (CurrentNmr<MaximumNmr) then
        TryStep := TryStep( CurrentNmr+1, _First_);
  end;
//
begin
  Case C of
    _First_ : Go := TryStep( 1, _First_);
    _Step_  : Go := TryStep( CurrentNmr, _Step_);
  Else
    Go := inherited Go(C);
  end;
end;

Function Buckets.List(Nmr: Integer): Lists;
begin
  List := (Item[Nmr] As Lists);
end;

Function Buckets.CurrentList: Lists;
begin
  Case (1<=CurrentNmr) and (CurrentNmr<=MaximumNmr) of
    Ok : CurrentList := List(CurrentNmr);
    No : CurrentList := NIL;
  end;
end;

Function Buckets.Current: Elements;
begin
  Case CurrentList<>NIL of
    Ok : Current := CurrentList.Current;
    No : Current := NIL;
  end;
end;

Function Buckets.MaxCapacity: Integer;
begin
(* --- Restricted only for visualization purposes ------------------------ *)
   MaxCapacity := _MaximumSize_;
(* ----------------------------------------------------------------------- *)
end;

Procedure Buckets.Arrange( k,l, Dk, Dl: Integer);
Var
  i : Integer;
begin
  For i := 1 to MaximumNmr do
    (Item[i] As Lists).Arrange( k+i*Dk,l, 0, Dl);
end;
(* ----------------------------------------------------------------------- *)

(* --- LISTS ------------------------------------------------------------- *)
Constructor Lists.SetUp;
begin
  inherited SetUp;
  Head   := NIL;
  Tail   := NIL;
end;

Destructor Lists.Clear;
Var
  Here,
  This : Elements;
begin
  Here := Head;
  While Here<>NIL do begin
    This := Here;
    Here := Here[_Next_];
    This.Clear;
  end;
end;

Function Lists.LinkFor( E: Elements): Elements;
begin
  LinkFor := ListLinks.SetUp;
  Result[_Info_] := E;
end;

Function Lists.Go( C: Codes ): Elements;
begin
  Case C of
    _First_    : Go := Head;
    _Next_,
    _Step_     : Case CurrentLink<>NIL of
                  Ok: Go := CurrentLink[_Next_];
                  No: Go := NIL;
                 end;
    _BackStep_ : Case CurrentLink<>NIL of
                   Ok: Go := CurrentLink[_Prev_];
                   No: Go := NIL;
                 end;
    _Last_     : Go := Tail;
  Else
    Go := Inherited Go(C);
  end;
  CurrentLink := Result;
end;

Procedure Lists.Link( This: Elements; How: Codes; Where: Elements);
Var
  Prev, Next : Elements;
begin
  If This<>NIL then
    Case Where=NIL of
      Ok: Case How of
            _After_ : begin // link as New "Head" taking NIL as -00 element
                         This[_Next_] := Head;
                         Case Head<>NIL of
                           Ok : Head[_Prev_] := This;
                           No : Tail := This;
                         end;
                         Head := This;
                      end;
            _Before_: begin // link as New "Tail" taking NIL as +00 element
                        This[_Prev_] := Tail;
                        Case Tail<>NIL of
                          Ok : Tail[_Next_] := This;
                          No : Head := This;
                        end;
                        Tail := This;
                      end;
          end;
      No: Case How of
           _After_ : begin
                       Prev := Where;
                       Next := Where[_Next_];
                       This[_Next_] := Next;
                       This[_Prev_] := Prev;
                       Prev[_Next_] := This;
                       Case Next=NIL of
                         Ok : Tail := This;
                         No : Next[_Prev_] := This;
                       end;
                     end;
          _Before_ : begin
                       Prev := Where[_Prev_];
                       Next := Where;
                       This[_Next_] := Next;
                       This[_Prev_] := Prev;
                       Next[_Prev_] := This;
                       Case Prev=NIL of
                         Ok : Head := This;
                         No : Prev[_Next_] := This;
                       end;
                     end;
          end;
  end;
end;

Procedure Lists.UnLink( E: Elements);
Var
  Prev, Next : Elements;
begin
  If E<>NIL then begin
    Prev := E[_Prev_];
    Next := E[_Next_];
    Case Prev=NIL of
      Ok : Head := Next;
      No : Prev[_Next_] := Next;
    end;
    Case Next=NIL of
      Ok : Tail := Prev;
      No : Next[_Prev_] := Prev;
    end;
  end;
end;

Procedure Lists.Split( Var L,R: Lists);
Var
  Here,
  This   : Elements;
  Toggle : Boolean;
begin
  L := Lists.SetUp;
  R := Lists.SetUp;
  Case Head=NIL of
    Ok: Done;
    No: Case Head=Tail of
          Ok: L.Link( Head, _After_, Head);
          No: begin
                Toggle := True;
                Here   := Head;
                While Here<>NIL do begin
                  This := Here;
                  Here := Here[_Next_];
                  Case Toggle of
                    Ok: With L do Link( This, _After_, Tail);
                    No: With R do Link( This, _After_, Tail);
                  end;
                  Toggle := Not Toggle;
                end;
              end;
        end;
  end;
  Head := NIL;
  Tail := NIL;
end;

Procedure Lists.Merge( L,R: Lists; Order: OrderFunctions);
Var
  Lf,Rt : Elements;
  More  : Boolean;
begin
  Lf := L.Head;
  Rt := R.Head;
  More := True;
  Repeat
    Case NotNIL( Lf, Rt) of
      _Both_: Case Order( Lf[_Key_], Rt[_Key_]) of
                _Less_,
                _Equal_  : begin
                             L.UnLink(Lf);
                             Link( Lf, _After_, Tail);
                             Lf := L.Head;
                           end;
                _Greater_: begin
                             R.UnLink(Rt);
                             Link( Rt, _After_, Tail);
                             Rt := R.Head;
                           end;
              end;
      _1st_ : begin Append(L); More := False; end;
      _2nd_ : begin Append(R); More := False; end;
      _None_: More := False;
    end;
  Until Not More;
end;

Procedure Lists.Append( L: Lists);
begin
  Case NotNil( Tail, L.Tail) of
    _None_,
    _1st_  : Done;
    _2nd_  : begin
               Head := L.Head;
               Tail := L.Tail;
             end;
    _Both_ : begin
               Tail[_Next_]   := L.Head;
               L.Head[_Prev_] := Tail;
               Tail           := L.Tail;
             end;
  end;
  L.Clear;
end;

Procedure Lists.DrawLinks( W: TCanvas; k,l: Integer);
begin
//DrawAnnotation( W, ' ', Head, -1,+3);
  DrawAnnotation( W, ' Head', Head, -1,+3);
//DrawAnnotation( W, '', Tail, +2,+5);
  DrawAnnotation( W, 'Tail', Tail, +2,+5);
end;
(* ----------------------------------------------------------------------- *)

(* --- SKIP LISTS -------------------------------------------------------- *)
Constructor SkipLists.SetUp;
begin
  inherited SetUp;
  ListOflevels := Lists.SetUp;
  AdjustHeight(1);
end;

Destructor SkipLists.Clear;
Var
  Here,
  This : Lists;
begin
  ListOfLevels.Go(_First_);
  Here := CurrentList;
  While Here<>NIL do begin
    This := Here;
    ListOfLevels.Go(_Step_);
    Here := CurrentList;
    This.Clear;
  end;
end;

Function SkipLists.LinkFor( E: Elements): Elements;
begin
  LinkFor := SkipListLinks.SetUp;
  Result[_Info_] := E;
end;

Function SkipLists.Go( C: Codes): Elements;
//
  Function TryStep( L,T: Codes): Elements;
  begin
    If L<>_Stay_ then ListOfLevels.Go(L);
    Case CurrentList<>NIL of
      Ok : TryStep := CurrentList.Go(T);
      No : TryStep := NIL;
    end;
    If (Result=NIL) and (CurrentList<>NIL) then
      TryStep := TryStep(_Step_, _First_);
  end;
//
begin
  Case C of
    _First_ : Go := TryStep( _First_, _First_);
    _Step_  : Go := TryStep( _Stay_, _Step_);
  Else
    Go := inherited Go(C);
  end;
end;

Function SkipLists.CurrentList: Lists;
begin
  With ListOfLevels do
    Case CurrentLink<>NIL of
      Ok : CurrentList := (CurrentLink[_Info_] As Lists);
      No : CurrentList := NIL;
    end;
end;

Function SkipLists.Current: Elements;
begin
  Case CurrentList<>NIL of
    Ok: Current := CurrentList.Current;
    No: Current := NIL;
  end;
end;

Procedure SkipLists.AdjustHeight( Levels: Integer);
Var
  L : Integer;
begin
  For L := ListOfLevels.NmrOfElements+1 to Levels do
    With ListOfLevels do
      Link( LinkFor(Lists.SetUp), _After_, Tail);
end;

Procedure SkipLists.Link( This: Elements; Where: Elements; Levels: Integer);
Var
  L    : Integer;
  Here,
  Down : Elements;
begin
  AdjustHeight(Levels);
  Down := NIL;
  ListOfLevels.Go(_First_);
  For L := 1 to Levels do begin
    Here := LinkFor(This);
    CurrentList.Link( Here, _After_, Where);
//
    If Down<>NIL then begin
      Down[_UpLink_] := Here;
      Here[_DnLink_] := Down;
    end;
    Down := Here;
//
    ListOfLevels.Go(_Step_);
    If Where<>NIL then
      Where := Where[_UpPrev_];
  end;
end;

Procedure SkipLists.DeleteLinks( L: Elements);
var
  Here: Elements;
begin
  ListOfLevels.Go(_First_);
  While (L<>NIL) and (CurrentList<>NIL) do begin
    CurrentList.UnLink(L);
    Here := L;
    L := L[_UpLink_];
    Here.Clear; // garbage collection
    ListOfLevels.Go(_Step_);
  end;
end;

Function SkipLists.Enumerate( F,S: Codes; From,Step: Integer): Integer;
begin
  ListOfLevels.Go(_First_);
  Case CurrentList=NIL of
    Ok: Enumerate := From;
    No: Enumerate := CurrentList.Enumerate( F,S, From,Step);
  end;
end;

Procedure SkipLists.Arrange( k,l,Dk,Dl: Integer);
Var
  Height : Integer;
  Here,
  This   : Elements;
begin
  With ListOfLevels do begin
    Height := NmrOfElements;
    Go(_First_);
    With CurrentList do begin
      Arrange( k, l+Height*Dl, Dk, 0);
      Here := Go(_First_);
      While Here<>NIL do begin
        This := Here[_UpLink_];
        While This<>NIL do begin
          This.x := This[_DnLink_].x;
          This.y := This[_DnLink_].y - Dl;
          This := This[_UpLink_];
        end;
        Here := Go(_Step_);
      end;
    end;
  end;
end;
(* ----------------------------------------------------------------------- *)

(* --- ROOTED TREES ------------------------------------------------------ *)
Constructor RootedTrees.SetUp;
begin
  inherited SetUp;
  Root    := NIL;
end;

Destructor RootedTrees.Clear;
Var
  Here,This : Elements;
begin
  Here := Go(_PostFirst_);
  While Here<>NIL do begin
    This := Here;
    Here := Go(_PostStep_);
    This.Clear;
  end;
  Root := NIL;
end;

Function RootedTrees.LinkFor( E: Elements): Elements;
begin
  LinkFor := AnyDegTreeLinks.SetUp;
  Result[_Info_] := E;
end;

Procedure RootedTrees.Link( This: Elements; How: Codes; Where: Elements);
begin
  Case How of
    _Root_      : begin
                    Root := This;
                    If This<>NIL then This[_Parent_] := NIL;
                  end;
    _Child_,
    _FirstChild_,
    _LastChild_ : Case (Where<>NIL) of
                    Ok : If This<>NIL then This[_Parent_] := Where;
                    No:  Link( This, _Root_, NIL);
                  end;
  end;
end;

Procedure RootedTrees.UnLink( E: Elements);
begin
  If E<>NIL then begin
    E[_Parent_] := NIL;
    If E=Root then Root := NIL;
  end;
end;

Function RootedTrees.Go( C: Codes): Elements;
begin
  Case C of
    _Root_     : Go := Root;
    //
    _First_    : Go := Root;
    _Step_     : If CurrentLink<>NIL then
                   Go := CurrentLink[_PreStep_];
    //
    _PreFirst_ : Go := Root;
    _InFirst_  : Case Root<>NIL of
                   Ok: Go := Root[_AllLf_];
                   No: Go := NIL;
                 end;
    _PostFirst_: Case Root<>NIL of
                   Ok: Go := Root[_AllDn_];
                   No: Go := NIL;
                 end;
    //
    _PreStep_,
    _InStep_,
    _PostStep_ : If CurrentLink<>NIL then
                   Go := CurrentLink[C];
    //
    _BFS_      : Go := Root;
    // _BFSStep_ : cannot be implemented efficiently
  Else
    Go := inherited Go(C);
  end;
  CurrentLink := Result;
end;

Procedure RootedTrees.Arrange;
var
  Last_k: Integer;
//
  Procedure ArrangeLink( This: Elements; k,l: Integer);
  Var
    Child : Elements;
  begin
    Child := This[_FirstChild_];
    While Child<>NIL do begin
      ArrangeLink( Child, Last_k, l+Dl);
      Child := Child[_NextSibling_];
      If Child<>NIL then Inc(Last_k, Dk);
    end;
    Case This[_FirstChild_]<>NIL of
      Ok : This.x := (This[_FirstChild_].x + This[_LastChild_].x) div 2;
      No : This.x := k;
    end;
    This.y := l;
  end;
//
begin
  (* --- In the standard Lf_To_Rt, Up_To_Dn way -------------------------- *)
  Last_k :=  k;
  If Root<>NIL then ArrangeLink( Root, k,l);
end;

Procedure RootedTrees.DrawLinks( W: TCanvas; k,l: Integer);
begin
//DrawAnnotation( W, '', Root, +1,-2);
  DrawAnnotation( W, 'Root', Root, +1,-2);
end;

Function RootedTrees.Enumerate( F,S: Codes; From,Step: Integer): Integer;
Var
  Queue : Array Of Elements;
  First,
  Last  : Integer;
//
  Procedure SetUpQueue;
  begin
    SetLength(Queue, NmrOfElements);  // a compromise for simplicity
    First := 0;
    Last  := (-1);
  end;
//
  Function QueueNotEmpty: Boolean;
  begin
    QueueNotEmpty := (First<=Last);
  end;
//
  Procedure EnQueue( P: Elements);
  begin
    If P<>NIL then begin Inc(Last); Queue[Last] := P; end;
  end;
//
  Function DeQueue: Elements;
  begin
    Case QueueNotEmpty of
      Ok: begin DeQueue := Queue[First]; Inc(First); end;
      No: DeQueue := NIL;
    end;
  end;
//
Procedure BFS_Enumerate;
Var
  Here,
  Child : Elements;
begin
  SetUpQueue;
  Here := Root;
  EnQueue(Here);
  While QueueNotEmpty do begin
    Here := DeQueue;
    If Here[_Info_]<>NIL then begin
      Inc(From,Step);
      Here[_Info_].Number := From;
    end;
    //
    Child := Here[_FirstChild_];
    While Child<>NIL do begin
      EnQueue(Child);
      Child := Child[_NextSibling_];
    end;
    //
  end;
  Queue := NIL;
end;

begin
  If (F=_BFS_) then begin
                      BFS_Enumerate;
                      Enumerate := From;
                    end
               else Enumerate := inherited Enumerate( F,S, From,Step);
end;
(* ----------------------------------------------------------------------- *)

(* ---  BINARY TREES ----------------------------------------------------- *)
Function BinaryTrees.LinkFor( E: Elements): Elements;
begin
  LinkFor := BinaryTreeLinks.SetUp;
  Result[_Info_] := E;
end;

Function BinaryTrees.GoBinary( N: Integer): Elements;
Var
  C : Array[1..64] of Codes; // Enough For 2^64 Elements}, N>=1
  k : Integer;
begin
  k := 0;
  While N>1 do begin Inc(k); C[k] := Child[N Mod 2]; N := N Div 2; end;
  GoBinary := Root;
  While (Result<>NIL) and (k>0) do begin
    GoBinary  := Result[ C[k] ];
    Dec(k);
  end;
  CurrentLink := Result;
end;

Procedure BinaryTrees.RotateUp( E: Elements);
begin
  If (E<>NIL) then begin
    (E As BinaryTreeLinks).RotateUp;
    If E[_Parent_]=NIL then Root := E;
  end;
end;

Procedure BinaryTrees.DoubleRotateUp( E: Elements);
begin
  RotateUp(E);
  RotateUp(E);
end;

Procedure BinaryTrees.SplayUp( E: Elements);
Var
  P,G : Elements;
//
  Procedure Zig( E: Elements);    begin
                                    RotateUp(E);
                                  end;
  Procedure ZigZig( E: Elements); begin
                                    RotateUp(E[_Parent_]); RotateUp(E);
                                  end;
  Procedure ZigZag( E: Elements); begin
                                    RotateUp(E); RotateUp(E);
                                  end;
//
begin
  If E<>NIL then begin
    P := E[_Parent_];
    If P<>NIL then begin
      G := P[_Parent_];
      Case G = NIL of
        Ok: Zig(E);
        No: Case E.IsWhat of
              _Lf_: Case P.IsWhat of
                      _Lf_: ZigZig(E);
                      _Rt_: ZigZag(E);
                    end;
              _Rt_: Case P.IsWhat of
                      _Lf_: ZigZag(E);
                      _Rt_: ZigZig(E);
                    end;
            end;
      end;
    end;
  end;
end;

Procedure BinaryTrees.InsBalance( P: Elements; Which: Codes);
begin
end;

Procedure BinaryTrees.DelBalance( P: Elements; Which: Codes);
begin
end;

procedure BinaryTrees.FindBalance(P: Elements);
begin
end;

Procedure BinaryTrees.Link( This: Elements; How: Codes; Where: Elements);
begin
  Case How of
    _Root_ : begin
               Root := This;
               If This<>NIL then This[_Parent_]  := NIL;
             end;
    _Lf_   : begin
               If This<>NIL  then This[_Parent_] := Where;
               If Where<>NIL then Where[_Lf_]    := This;
             end;
    _Rt_   : begin
               If This<>NIL  then This[_Parent_] := Where;
               If Where<>NIL then Where[_Rt_]    := This;
             end;
  end;
end;

Procedure BinaryTrees.UnLink( E: Elements);
Var
  P: Elements;
begin
  If E<>NIL then begin
    P := E[_Parent_];
    Case P=NIL of
      Ok : Root := NIL;
      No : Case E.IsWhat of
             _Lf_ : P[_Lf_]:= NIL;
             _Rt_ : P[_Rt_]:= NIL;
           end;
    end;
    E[_Parent_] := NIL;
  end;
end;

Procedure BinaryTrees.InsertLeaf( E: Elements; How: Codes; P: Elements );
Var
  Here : Elements;
begin
  If E<>NIL then begin
    (* --- Linking ------------------------------------------------------- *)
     Link( E,How, P);
     E.Rank := 1;
    (* --- Weighting ----------------------------------------------------- *)
     Here := P;
     While Here<>NIL do begin
       Here.Weight := Here.Weight+1;
       Here := Here[_Parent_];
     end;
    (* --- Balancing ----------------------------------------------------- *)
     InsBALANCE( P, How);
    (* --- Done ---------------------------------------------------------- *)
  end;
end;

Procedure BinaryTrees.DeleteLeaf( E: Elements);
Var
  Here,P : Elements;
  Edge   : Codes;
begin
  If E<>NIL then begin
    P := E[_Parent_];
    Edge := E.IsWhat;
    (* --- Linking ------------------------------------------------------- *)
     UnLink(E);
     E.Clear; // garbage collection
    (* --- Weighting ----------------------------------------------------- *)
     Here := P;
     While Here<>NIL do begin
       Here.Weight := Here.Weight-1;
       Here := Here[_Parent_];
     end;
   (* --- Balancing ------------------------------------------------------ *)
     DelBALANCE( P, Edge );
   (* --- Done ----------------------------------------------------------- *)
  end;
end;

Procedure BinaryTrees.Arrange;
var
  Last_k: Integer;
//
  Procedure ArrangeLink( This: Elements; k,l: Integer );
  Var
    Lk,Rk : Integer;
    Lf,Rt : Elements;
  begin
    If This<>NIL then begin
      Lf := This[_Lf_];
      Case Lf<>NIL of
        Ok: begin ArrangeLink( Lf, Last_k,l+Dl); Lk := Lf.x; end;
        No: Lk := Last_k;
      end;
      Last_k := Last_k + Dk div 2;
      Rt := This[_Rt_];
      Case Rt<>NIL of
        Ok: begin ArrangeLink( Rt, Last_k, l+Dl); Rk := Rt.x; end;
        No: Rk := Last_k;
      end;
      This.x := (Lk+Rk) Div 2;
      This.y := l;
    end;
  end;
//
begin
  (* --- In the standard Lf_To_Rt, Up_To_Dn way -------------------------- *)
  Last_k := k;
  ArrangeLink( Root, k,l);
end;
(* ----------------------------------------------------------------------- *)

(* --- HALF BALANCED TREES ----------------------------------------------- *)
Procedure HalfBalancedTrees.InsBalance( P: Elements; Which: Codes);
Var
  S,G : Elements;
  MirrorChild: Codes;
begin
  If (P<>NIL) then begin
    G := P[_Parent_];
    S := P[Which];      // S is not NIL in INSERT-Balance
    If (G<>NIL) and (G.Rank=P.Rank) and (P.Rank=S.Rank) then begin
      MirrorChild := Mirror(Which);
      Case (Which=P.IsWhat) of
        Ok: Case S.Rank-G.RankOf(MirrorChild) of
              0 : begin
                    G.Rank := G.Rank+1;
                    InsBALANCE(G[_Parent_], G.IsWhat);
                  end;
              1 : RotateUp(P);
            end;
        No: Case S.Rank-G.RankOf(Which) of
              0 : begin
                    G.Rank := G.Rank+1;
                    InsBALANCE(G[_Parent_], G.IsWhat);
                  end;
              1 : DoubleRotateUp(S);
            end;
      end;
    end;
  end;
end;

Procedure HalfBalancedTrees.DelBalance( P: Elements; Which: Codes);
Var
  C,M : Elements;
  T   : Integer;
  MirrorChild : Codes;
begin
  If P<>NIL then begin
    T := P.RankOf(Which);
    MirrorChild := Mirror(Which);
    If (P<>NIL) and ((P.Rank-T)=2) then begin
      Case P.RankOf(MirrorChild)-T of
        1 : begin
              M := P[MirrorChild];               // Not NIL!
              If (M.RankOf(MirrorChild)-T)=1 then begin
                P.Rank := P.Rank-1;
                M.Rank := M.Rank+1;
                RotateUp(M);
              end
              Else
              If (M.RankOf(Which)-T)=1 then begin
                C := M[Which];                   // Not NIL!
                P.Rank := P.Rank-1;
                C.Rank := C.Rank+1;
                DoubleRotateUp(C);
              end
              Else begin
                P.Rank := P.Rank-1;
                DelBALANCE( P[_Parent_], P.IsWhat);
              end;
            end;
        2 : begin
              M := P[MirrorChild];               // Not NIL!
              C := M[Which];                     // Not NIL!
              If (C.RankOf(MirrorChild)-T)=1 then begin
                P.Rank := P.Rank-1;
                C.Rank := C.Rank+1;
                DoubleRotateUp(C);
              end
              Else
              If (C.RankOf(Which)-T)=1 then begin
                P.Rank := P.Rank-1;
                RotateUp(M);
                DoubleRotateUp(C[Which]);
              end
              Else begin
                P.Rank := P.Rank-1;
                RotateUp(M);
              end;
           end;
      end;
    end;
  end;
end;
(* ----------------------------------------------------------------------- *)

(* --- AVL TREES --------------------------------------------------------- *)
Procedure AVLtrees.InsBalance( P: Elements; Which: Codes);
Var
  Delta : Integer;
  S,E   : Elements;
begin
  If P<>NIL then begin
    S := P[Which];
    Delta := P.RankOf(Which) - P.RankOf(Mirror(Which));
    Case Delta of
       0: Done;
       1: begin
            P.Rank := P.Rank+1;
            InsBALANCE( P[_Parent_], P.IsWhat);
          end;
       2: begin
            Delta := S.RankOf(Which)-S.RankOf(Mirror(Which));
            Case Delta of
              +1: begin
                    P.Rank := P.Rank-1;
                    RotateUp(S);
                  end;
               0: begin
                    S.Rank := S.Rank+1;
                    RotateUp(S);
                    InsBALANCE( S[_Parent_], S.IsWhat);
                  end;
              -1: begin
                    S.Rank := S.Rank-1;
                    P.Rank := P.Rank-1;
                    E := S[Mirror(Which)];
                    E.Rank := E.Rank+1;
                    DoubleRotateUp(E);
                  end;
            end;
         end;
    end;
  end;
end;

Procedure AVLtrees.DelBalance( P: Elements; Which: Codes);
Var
  Delta : Integer;
  E     : Elements;
begin
  If P<>NIL then begin
    Delta := P.RankOf(Mirror(Which)) - P.RankOf(Which);
    Case Delta of
      0: begin
            P.Rank := P.Rank-1;
            DelBALANCE( P[_Parent_], P.IsWhat);
          end;
      1: Done;
      2: begin
           E := P[Mirror(Which)];  // E Not NIL
           Delta := E.RankOf(Mirror(Which)) - E.RankOf(Which);
           Case Delta of
             -1: begin
                   P.Rank := P.Rank-2;
                   E.Rank := E.Rank-1;
                   E := E[Which];
                   E.Rank := E.Rank+1;
                   DoubleRotateUp(E);
                   DelBALANCE( E[_Parent_], E.IsWhat);
                 end;
              0: begin
                   P.Rank := P.Rank-1;
                   E.Rank := E.Rank+1;
                   RotateUp(E);
                 end;
             +1: begin
                   P.Rank := P.Rank-2;
                   RotateUp(E);
                   DelBALANCE( E[_Parent_], E.IsWhat);
                 end;
            end;
          end;
    end;
  end;
end;
(* ----------------------------------------------------------------------- *)

(* --- SPLAY TREES ------------------------------------------------------- *)
Procedure SplayTrees.InsBalance( P: Elements; Which: Codes);
begin
  If P<>NIL then begin
    P := P[Which];
    Repeat SplayUp(P) Until P[_Parent_]=NIL;
  end;
end;

Procedure SplayTrees.DelBalance( P: Elements; Which: Codes);
begin
  If P<>NIL then
    Repeat SplayUp(P) Until P[_Parent_]=NIL;
end;

Procedure SplayTrees.FindBalance( P: Elements);
begin
  If P<>NIL then
    Repeat SplayUp(P) Until P[_Parent_]=NIL;
end;
(* ----------------------------------------------------------------------- *)

(* --- ROTATE TO ROOT TREES ---------------------------------------------- *)
Procedure RotateToRootTrees.InsBalance( P: Elements; Which: Codes);
begin
  If P<>NIL then begin
    P := P[Which];
    Repeat RotateUp(P) Until P[_Parent_]=NIL;
  end;
end;

Procedure RotateToRootTrees.DelBalance( P: Elements; Which: Codes);
begin
  If P<>NIL then
    Repeat RotateUp(P) Until P[_Parent_]=NIL;
end;

Procedure RotateToRootTrees.FindBalance( P: Elements);
begin
  If P<>NIL then
    Repeat RotateUp(P) Until P[_Parent_]=NIL;
end;
(* ----------------------------------------------------------------------- *)

(* --- RANDOM ROTATE-UP TREES -------------------------------------------- *)
Procedure RandomizedTrees.InsBalance( P: Elements; Which: Codes);
Var
  H,
  Height : Integer;
  Here   : Elements;
begin
  If P<>NIL then begin
    P := P[Which];
    P.Rank := 0;  // No Ranks Are Needed Here
    Height := 0;
    Here   := P;
    Repeat Inc(height); Here := Here[_Parent_] Until (Here=NIL);
    Height := Random(Height);
    For H := 1 to Height do RotateUp(P);
  end;
end;
(* ----------------------------------------------------------------------- *)

(* --- RANDOMIZED TREES -------------------------------------------------- *)
Procedure RandomizedSearchTrees.InsBalance( P: Elements; Which: Codes);
Var
  Here : Elements;
begin
  Case P=NIL of
    Ok: Here := Root;
    No: Here := P[Which];
  end;
  Here.Rank := 1+Random(9999);  // Random Ranks Are Needed Here
  P := Here[_Parent_];
  While (P<>NIL) and (Here.Rank < P.Rank) do begin
    RotateUp(Here);
    P := Here[_Parent_];
  end;
end;
(* ----------------------------------------------------------------------- *)

(* --- ANY DEGREE TREES -------------------------------------------------- *)
Constructor AnyDegTrees.SetUp;
begin
  inherited SetUp;
end;

Function AnyDegTrees.LinkFor( E: Elements): Elements;
begin
  LinkFor := AnyDegTreeLinks.SetUp;
  Result[_Info_] := E;
end;

Procedure AnyDegTrees.Link( This: Elements; How: Codes; Where: Elements);
Var
  Prev : Elements;
begin
  If This<>NIL then
    Case How of
      _Root_  : begin
                  Root := This;
                  This[_Parent_] := NIL; // Just to be sure...
                end;
      _Child_ : Case Where[_LastChild_]=NIL of
                  Ok: begin
                        Where[_FirstChild_] := This;
                        Where[_LastChild_]  := This;
                        This[_PrevSibling_] := NIL;
                        This[_NextSibling_] := NIL;
                        This[_Parent_]      := Where;
                      end;
                  No: begin
                        Prev := Where[_LastChild_];
                        This[_NextSibling_] := NIL;
                        This[_PrevSibling_] := Prev;
                        Prev[_NextSibling_] := This;
                        Where[_LastChild_]  := This;
                        This[_Parent_]      := Where;
                      end;
                end;
    end;
end;

Procedure AnyDegTrees.UnLink( E: Elements);
Var
  P, Prev, Next : Elements;
begin
  If E<>NIL then begin
    P := E[_Parent_];
    Case P<>NIL of
      Ok: begin
            Prev := E[_PrevSibling_];
            Next := E[_NextSibling_];
            Case Prev=NIL of
              Ok : P[_FirstChild_] := Next;
              No : Prev[_NextSibling_] := Next;
            end;
            Case Next=NIL of
              Ok : P[_LastChild_] := Prev;
              No : Next[_Prev_]   := Prev;
            end;
          end;
      No: Root := NIL;
    end;
    E[_Parent_] := NIL;
  end;
end;
(* ----------------------------------------------------------------------- *)

(* --- B-TREES ----------------------------------------------------------- *)
Constructor Btrees.SetUp( MinS, MaxS: Integer);
begin
  inherited SetUp;
  Tree := RootedTrees.SetUp;
  Tree.Root := BTreeLinks.SetUp(MaxS+1);
  MinSize := MinS;
  MaxSize := MaxS;
end;

Destructor Btrees.Clear;
begin
  Tree.Clear;
end;

Function BTrees.LinkFor( E: Elements): Elements;
Var
  BTreeLink: BTreeLinks;
begin
  BTreeLink := BTreeLinks.SetUp(MaxSize);
  If E<>NIL then
    With BTreeLink do begin
      Size := 1;
      Datum[1] := E;
      Degree := 2;
    end;
  LinkFor := BTreeLink;
end;

Function BTrees.CurrentNode: BTreeLinks;
begin
  With Tree do
    Case CurrentLink<>NIL of
      Ok : CurrentNode := (CurrentLink As BTreeLinks);
      No : CurrentNode := NIL;
    end;
end;

Function BTrees.Current: Elements;
begin
  Case CurrentNode<>NIL of
    Ok : Current := CurrentNode.Current;
    No : Current := NIL;
  end;
end;

Procedure Btrees.StepIfHighSize( var B: BTreeLinks; H: Integer);
Var
  Here,
  L,R : BTreeLinks;
begin
  // Data: 1..Size-1, Children: 1..Degree
  // Invariant: Element under Deletion is in the H-th Subtree of B Node
  // After StepIfHighSize "Here"-node has at least MinSize+1 data
  Case B<>NIL of
    Ok: Here := B.Child[H];
    No: Here := BTreeLinks(Tree.Root);
  end;
  If (Here<>NIL) then
    If (Here<>Tree.Root) and (Here.Size=MinSize) then begin
      // Select a sibling
      If (H=B.Degree) then Dec(H);
      L := B.Child[H];
      R := B.Child[H+1];
      // Meld or Rotate
      If (L.Size=MinSize) and (R.Size=MinSize) then begin
        Here := BTreeLinks.Meld( B.Datum[H], L,R);
        B.DelData( H, Here);
        // Check For Old Root Destruction
        If B.Size=0 then begin
          B.Clear;
          Here.Parent := NIL;
          Tree.Root := Here;
        end;
      end
      Else If (L.Size>MinSize) then B.RotateUp( _Lf_, H)  // L<>B.Child[H]!
      Else If (R.Size>MinSize) then B.RotateUp( _Rt_, H); // R<>B.Child[H]!
  end;
  B := Here;
end;

Procedure Btrees.StepIfLowSize( Var B: BTreeLinks; H: Integer);
Var
  E    : Elements;
  Here,
  L,R  : BTreeLinks;
begin
  // Invariant: Element will be Inserted in the Hth Subtree of B Node
  // Data: 1..Size, Children: 1..Degree
  Case B<>NIL of
    Ok: Here := B.Child[H];
    No: Here := (Tree.Root As BTreeLinks);
  end;
  If (Here<>NIL) and (Here.Size=MaxSize) then begin
    Here.Split( (MaxSize Div 2)+1, E, L, R);
    Here.Clear;
    // Check For New Root Construction
    If (B=NIL) then begin
      B := BTreeLinks.SetUp(MaxSize+1);
      Tree.Root := B;
      H := 1;
    end;
    B.InsData( H, E, L, R);
    Here := B;
  end;
  B := Here;
end;

Procedure BTrees.Draw( W: TCanvas; k,l: Integer);
begin
  Tree.Draw( W, k,l);
end;

Procedure BTrees.DrawLinks( W: TCanvas; k,l: Integer);
begin
  With Tree.Root as BTreeLinks do
    If (Size>0) then
    //DrawAnnotation( W, '', Tree.Root, +1,-2);
      DrawAnnotation( W, 'Root', Tree.Root, +1,-2);
end;

Procedure Btrees.Arrange( k,l,Dk,Dl: Integer);
begin
  Tree.Arrange(k,l,Dk,Dl);
end;

Procedure Btrees.Frame( Var x1,y1, x2,y2: Integer);
begin
  Tree.Frame(x1,y1,x2,y2);
end;

Function Btrees.Go(C: Codes): Elements;
//
  Function TryStep( N,D: Codes): Elements;
  begin
    If N<>_Stay_ then Go := Tree.Go(N);
    Case CurrentNode<>NIL of
      Ok : Case D of
             _First_,
             _PreFirst_,
             _PostFirst_: With CurrentNode do begin
                            CurrentData := 1;
                            TryStep := Datum[CurrentData];
                          end;
             _Step_,
             _PreStep_,
             _PostStep_:  With CurrentNode do begin
                            Inc(CurrentData);
                            Case CurrentData<=Size of
                              Ok: TryStep := Datum[CurrentData];
                              No: TryStep := NIL;
                            end;
                          end;
           end;
      No : TryStep := NIL;
    end;
    If (Result=NIL) and (CurrentNode<>NIL) then
      TryStep := TryStep( C, _First_);
  end;
//
begin
  Case C of
    _First_,
    _PreFirst_,
    _PostFirst_: Go := TryStep( C, C);
    //
    _Step_,
    _PreStep_,
    _PostStep_ : Go := TryStep( _Stay_, C);
    //
    _BFS_      : Go := NIL; // not implemented
  Else
    Go := inherited Go(C);
  end;
end;
(* ----------------------------------------------------------------------- *)

(* --- FORESTS ----------------------------------------------------------- *)
Constructor Forests.SetUp;
begin
  inherited SetUp;
  ListOfTrees := Lists.SetUp;
end;

Destructor Forests.Clear;
Var
  Here,
  This : AnyDegTrees;
begin
  ListOfTrees.Go(_First_);
  Here := CurrentTree;
  While Here<>NIL do begin
    This := Here;
    ListOfTrees.Go(_Step_);
    Here := CurrentTree;
    This.Clear;
  end;
end;

Function Forests.LinkFor( E: Elements): Elements;
Var
  Tree : AnyDegTrees;
begin
  Tree      := AnyDegTrees.SetUp;
  Tree.Root := Tree.LinkFor(E);
  LinkFor   := ListOfTrees.LinkFor(Tree);
end;

Function Forests.RootOf(E: Elements): Elements;
begin
  Case (E<>NIL) and (E[_Info_]<>NIL) of
    Ok: RootOf := RootedTrees(E[_Info_]).Root;
    No: RootOf := NIL;
  end;
end;

Function Forests.CurrentTree: AnyDegTrees;
begin
  With ListOfTrees do
    Case CurrentLink<>NIL of
      Ok : CurrentTree := (CurrentLink[_Info_] As AnyDegTrees);
      No : CurrentTree := NIL;
    end;
end;

Function Forests.Current: Elements;
begin
  Case CurrentTree<>NIL of
    Ok: Current := CurrentTree.Current;
    No: Current := NIL;
  end;
end;

Function Forests.Go( C: Codes): Elements;
//
  Function TryStep( L,T: Codes): Elements;
  begin
    If L<>_Stay_ then ListOfTrees.Go(L);
    Case CurrentTree<>NIL of
      Ok : TryStep := CurrentTree.Go(T);
      No : TryStep := NIL;
    end;
    If (Result=NIL) and (CurrentTree<>NIL) then
      TryStep := TryStep( _Step_, _First_);
  end;
//
begin
  Case C of
    _First_  : Go := TryStep( _First_, _First_);
    _Step_   : Go := TryStep( _Stay_, _Step_);
    _BFS_    : Go := NIL; // not implemented
  Else
    Go := inherited Go(C);
  end;
  CurrentLink := Result;
end;

Procedure Forests.Arrange( k,l,Dk,Dl: Integer);
Var
  x1,y1,
  x2,y2: Integer;
begin
  ListOfTrees.Go(_First_);
  While CurrentTree<>NIL do begin
    With CurrentTree do begin
      Arrange(k,l,Dk,Dl);
      Frame(x1,y1,x2,y2);
      k := x2+Dk+20;
    end;
    ListOfTrees.Go(_Step_);
  end;
end;

(* ======================================================================= *)

begin
end.
