(* ----------------------------------------------------------------------- *)
UNIT CONSTANTS_;
(* ----------------------------------------------------------------------- *)

INTERFACE
(* ----------------------------------------------------------------------- *)

USES

  Graphics;
(* ----------------------------------------------------------------------- *)

TYPE

  Codes = Integer;        // A 'Description' of What Is Asked For

  Areas = Record
            x1,y1,
            x2,y2 : Integer;
            Hit   : String;
          end;
(* ----------------------------------------------------------------------- *)

CONST

(* --- Limit Values ------------------------------------------------------ *)
  _MaximumSize_ =  2000;
  _MaxHitAreas_ = 20000;

(* --- Logical ----------------------------------------------------------- *)
  Ok = TRUE;              // ShortHand for TRUE
  No = FALSE;             // ShortHand for FALSE

(* --- Order ------------------------------------------------------------- *)
  _Less_        =  1;     // First element is LESS than the second
  _Equal_       =  2;     // First element is EQUAL to the second
  _Greater_     =  3;     // First element is GREATER than the second

(* --- General ----------------------------------------------------------- *)
  _Null_        =  1001;  // Representing a NULL result
  _Any_         =  1002;  // Asks for anything of the relevant type
  _All_         =  1003;  // Perform something to all elements

(* --- Places ------------------------------------------------------------ *)
  _UpLf_        =  1101;  // Point (x2,y2) is UPPER-LEFT  w.r. to (x1,y1)
  _UpRt_        =  1102;  // Point (x2,y2) is UPPER-RIGHT w.r. to (x1,y1)
  _DnLf_        =  1103;  // Point (x2,y2) is DOWN-LEFT   w.r. to (x1,y1)
  _DnRt_        =  1104;  // Point (x2,y2) is DOWN-RIGHT  w.r. to (x1,y1)

(* --- LINKING ----------------------------------------------------------- *)

(* --- General  ---------------------------------------------------------- *)
  _Owner_       =  1301;  // Reserved for later use
  _Key_         =  1302;  // Asks for the Key Part of stored info
  _Info_        =  1303;  // Asks for the Info of a linking element
  _HeapInfo_    =  1304;  // Asks for the HeapInfo of special elements
  _Stay_        =  1305;  // Ask to remain inside the same Storage Scheme
  _NewInfo_     =  1306;  // Asks for Info to be replaced with new one

(* --- List Linking ------------------------------------------------------ *)
  _Head_        =  1401;  // Asks for the FIRST element of a List
  _Tail_        =  1402;  // Asks for the LAST element of a List
  _Next_        =  1403;  // Asks for the NEXT element (after CURRENT<>NIL)
  _Prev_        =  1404;  // Asks for the PREVIOUS element (CURRENT<>NIL)
  _Before_      =  1405;  // Link AFTER Current (or as LAST if NIL)
  _After_       =  1406;  // Link BEFORE Current (or as FIRST if NIL)

(* --- Tree Linking, Basic ----------------------------------------------- *)
  _Root_        =  1501;  // Refers to the Root of 'current' tree
  _Parent_      =  1502;  // Refers to the PARENT of given tree-element
  _GrandPa_     =  1503;  // Refers to the PARENT of the PARENT
  _FirstChild_  =  1504;  // Refers to the 1st CHILD of given tree-element
  _Child_       =  1505;  // Link as CHILD of given tree-element
  _LastChild_   =  1506;  // Refers to the LAST CHILD of given tree-element
  _NextSibling_ =  1507;  // Refers to the NEXT SIBLING of given element
  _PrevSibling_ =  1508;  // Refers to the NEXT SIBLING of given element

  _AllDn_       =  1531;  // Walk all the way Down following First Child
  _AllUp_       =  1532;  // Walk all the way Up following Parent links
  _AllLf_       =  1533;  // Walk all the way Down following Left links
  _AllRt_       =  1534;  // Walk all the way Down following Right links

(* --- Tree Linking, Binary Trees ---------------------------------------- *)
  _Lf_          =  1601;  // Asks for the LEFT child
  _Rt_          =  1602;  // Asks for the RIGHT child

(* --- Tree Linking, Forest ---------------------------------------------- *)
  _FirstRoot_   =  1701;  // Refering to ROOT of the FIRST tree in forest
  _NextRoot_    =  1702;  // Refering to ROOT of the NEXT tree in forest
  _NewTree_     =  1703;  // LINK as NEW TREE in the forest

(* --- SkipLists Linking ------------------------------------------------- *)
  _UpLink_      =  1801;  // Asks for the element one level UP
  _DnLink_      =  1802;  // Asks for the element one level DOWN
  _UpPrev_      =  1803;  // Asks for the first PREVIOUS with UpLINK<>NIL
  _UpNext_      =  1804;  // Asks for the first NEXT with DnLINK<>NIL

(* --- 'Standard' Enumeration -------------------------------------------- *)
  _First_       =  2001;  // Asks for 'FIRST' element of storage scheme
  _Step_        =  2002;  // Asks for 'NEXT' element of scheme's 'Current'
  _Last_        =  2003;  // Asks for 'LAST' element of storage scheme
  _BackStep_    =  2004;  // Asks for 'PREV' element
  _Reported_    =  2005;  // Denotes the 'reported' enumeration

(* --- Enumeration, Trees ------------------------------------------------ *)
  _PreFirst_    =  2101;  // Asks for FIRST of a tree w.r. to PRE-ORDER
  _PreStep_     =  2102;  // Asks for NEXT of 'Current' w.r. to PRE-ORDER

  _PostFirst_   =  2103;  // Asks for FIRST of a tree w.r. to POST-ORDER
  _PostStep_    =  2104;  // Asks for NEXT of 'Current' w.r. to POST-ORDER

  _InFirst_     =  2105;  // Asks for FIRST of a bin. tree w.r. to IN-ORDER
  _InStep_      =  2106;  // Asks for NEXT of 'Current' w.r. to IN-ORDER

  _BFS_         =  2107;  // Asks for 'Breadth First Search' enumeration

(* --- Searching --------------------------------------------------------- *)
  _Top_         =  2201;  // Asks for the LEAST in the 'Order' element
  _Min_         =  2202;  // Asks for the Minimum
  _Max_         =  2203;  // Asks for the Maximum
  _Median_      =  2204;  // Reserved for the Median element

(* --- Which is Which ---------------------------------------------------- *)
  _None_        =  1;     // None element is as desired
  _Both_        =  2;     // Both elements is as desired
  _1st_         =  3;     // (only) 1st element is as desired
  _2nd_         =  4;     // (only) 2nd element is as desired

(* --- Operations -------------------------------------------------------- *)
  _Insert_      =  1;     // For later use
  _Delete_      =  2;
  _FindAny_     =  3;
  _FindMin_     =  4;
  _FindNext_    =  5;
  _Promote_     =  6;
  _ReportLess_  =  7;
  _ReportRange_ =  8;
  _ReportArea_  =  9;

(* --- Binary Tree Children ---------------------------------------------- *)
   Child : Array[0..1] of Codes = ( _Lf_, _Rt_);

(* --- Colors  ----------------------------------------------------------- *)
  _FrameColor_  =  clBLUE;
  _FillColor_   =  clLTGRAY;
  _LinkColor_   =  clGREEN;
  _KeyColor_    =  clBLACK;
  _InsColor_    =  clLIME;
  _DelColor_    =  clRED;
  _ReportColor_ =  clAQUA;
  _FindColor_   =  clAQUA;
  _UpDateColor_ =  clFUCHSIA;
  _NumberColor_ =  clMAROON;
  _RankColor_   =  clNAVY;
  _UsedColor_   =  clYELLOW;
  _NoteColor_   =  clRED;
  _BkGrColor_   =  clWHITE;
(* ----------------------------------------------------------------------- *)

VAR

  MarkUsed,
  ShowRanks,
  ShowWeights,
  ShowNumbers : Boolean;

  FontSize    : Integer;
  GridWidth,
  GridHeight  : Integer;
  NmrOfCells,
  CellWidth,
  CellHeight  : Integer;

  HitAreas : Array[1.._MaxHitAreas_] of Areas;
  NmrOfHitAreas : Integer;

(* ======================================================================= *)

Function Mirror( C: Codes): Codes;
(* ======================================================================= *)

IMPLEMENTATION
(* ----------------------------------------------------------------------- *)

Function Mirror( C: Codes): Codes;
begin
  Mirror := _Null_;
  Case C of
    _Lf_ : Mirror := _Rt_;
    _Rt_ : Mirror := _Lf_;
  end;
end;

(* ======================================================================= *)

begin
  MarkUsed    := False;  // Watch out: 'False' is the Default value
  ShowRanks   := False;
  ShowWeights := False;
  ShowNumbers := False;
  FontSize    := 10;
end.

