(* ----------------------------------------------------------------------- *)
UNIT EXAMPLES_;
(* ----------------------------------------------------------------------- *)

INTERFACE

(* ----------------------------------------------------------------------- *)

USES

  SysUtils,
  Math,

  CONSTANTS_,
  ELEMENTS_,
  DATA_,
  SSCHEMES_,
  DSTRUCTURES_;
(* ----------------------------------------------------------------------- *)

Function KeyEG( Range: Integer): Integer;
Function DataEG( Key: Integer): Elements;
Function StructureEG( DS: Codes; SS: Codes; Size: Integer): Structures;
(* ======================================================================= *)

IMPLEMENTATION
(* ======================================================================= *)

Function KeyEG( Range: Integer): Integer;
begin
  KeyEG := Random(Min(9*Range+99,9999));
end;

Function DataEG( Key: Integer): Elements;
begin
  DataEG := KeyData.SetUp( Key, _InsColor_);
end;

Function StructureEG( DS: Codes; SS: Codes; Size: Integer): Structures;
var
  k : Integer;
begin
  StructureEG := NIL;
  (* --------------------------------------------------------------------- *)
  Case DS of
   (* --- Catalogs ------------------------------------------------------- *)
    1: Case SS of
         1: StructureEG := CAT_Tables.SetUp( Tables.SetUp(100));
         2: StructureEG := CAT_DynamicTables.SetUp( DynamicTables.SetUp(10));
         3: StructureEG := CAT_Lists.SetUp(NIL);
       end;

   (* --- Stacks --------------------------------------------------------- *)
    2: Case SS of
         1: StructureEG := STACK_Tables.SetUp(NIL);
         2: StructureEG := STACK_Lists.SetUp(NIL);
       end;

   (* --- Queues --------------------------------------------------------- *)
    3: Case SS of
         1: StructureEG := QUEUE_CircularTables.SetUp(NIL);
         2: StructureEG := QUEUE_Lists.SetUp(NIL);
       end;

   (* --- Priority Queues ------------------------------------------------ *)
    4: Case SS of
       // Binary Heaps in a Table
         1: StructureEG := PQ_HeapTables.SetUp(NIL);

       // Heaps in a Binary Tree
         2: StructureEG := PQ_BinaryHeaps.SetUp(NIL);
         3: StructureEG := PQ_LeftistHeaps.SetUp(NIL);

       // Variable Degree Forest Heaps
         4: StructureEG := PQ_BinomialHeaps.SetUp;          // Non Amortized
         5: StructureEG := PQ_AmortizedBinomialHeaps.SetUp; // Amortized Cost
         6: StructureEG := PQ_FibonacciHeaps.SetUp;         // Fast Promoting

       // Self Adjusted Heaps
         7: StructureEG := PQ_SkewHeaps.SetUp(NIL);
         8: StructureEG := PQ_PairingHeaps.SetUp(NIL);

       end;

   (* --- Directories ---------------------------------------------------- *)
    5: Case SS of
       // Static, Binary Search
        1: StructureEG := DIR_SortedTables.SetUp(NIL);

       // Binary Search Trees, Balanced
        2: StructureEG := DIR_BinaryTrees.SetUp( NIL );
        3: StructureEG := DIR_BinaryTrees.SetUp( AVLTrees.SetUp );
        4: StructureEG := DIR_BinaryTrees.SetUp( HalfBalancedTrees.SetUp );

       // Multi-Way Search Trees
        5: StructureEG := DIR_Btrees.SetUp(2,5);
        6: StructureEG := DIR_BTrees.SetUp(4,9);

       // Self Adjusted Schemes
        7: StructureEG := DIR_SelfAdjustedLists.SetUp(NIL);
        8: StructureEG := DIR_BinaryTrees.SetUp(SplayTrees.SetUp);
        9: StructureEG := DIR_BinaryTrees.SetUp(RotateToRootTrees.SetUp);

       // Randomized
       10: StructureEG := DIR_BinaryTrees.SetUp(RandomizedTrees.SetUp);
       11: StructureEG := DIR_BinaryTrees.SetUp(RandomizedSearchTrees.SetUp);
       12: StructureEG := DIR_SkipLists.SetUp( SkipLists.SetUp);

       // Hashing Techniques
       13: StructureEG := DIR_ClosedHashTables.SetUp(59);
       14: StructureEG := DIR_OpenHashTables.SetUp(Tables.SetUp(59));
       15: StructureEG := DIR_OpenDynamicHashTables.SetUp(DynamicTables.SetUp(11));
       16: StructureEG := DIR_DoubleEntryHashTables.SetUp(59);
      end;

   (* --- Priority Search Directories ------------------------------------ *)
    6: Case SS of
         1: StructureEG := PSDIR_BinaryTrees.SetUp(NIL);
         2: StructureEG := PSDIR_BinaryTrees.SetUp(AVLTrees.SetUp);
         3: StructureEG := PSDIR_BinaryTrees.SetUp(HalfBalancedTrees.SetUp);
       end;
   (* -------------------------------------------------------------------- *)
  end;
  (* --------------------------------------------------------------------- *)
  Case DS of
    1,
    2,
    3,
    4: NmrOfCells := 1;
    5: Case SS of
         1..4  : NmrOfCells := 1;
         5     : NmrOfCells := 5;
         6     : NmrOfCells := 9;
         7..16 : NmrOfCells := 1;
      end;
    6: NmrOfCells := 2;
  end;
  (* --------------------------------------------------------------------- *)
  If Result<>NIL then begin
    MarkUsed := True;
    For k := 1 to Size do
      Result.Insert( DataEG(KeyEG(Size)) );
    MarkUsed := False;
  end;
end;

(* ======================================================================= *)

begin
end.
