(* ----------------------------------------------------------------------- *)
UNIT MAIN_;
(* ----------------------------------------------------------------------- *)

INTERFACE
(* ----------------------------------------------------------------------- *)

USES

  Windows,    Messages,     SysUtils,     Classes,    Graphics,
  Controls,   Forms,        Dialogs,      Menus,      ToolWin,
  ComCtrls,   Buttons,      ExtCtrls,     StdCtrls,   Mask,

  HELP_,
  CONSTANTS_,
  DRAW_,
  ELEMENTS_,
  DATA_,
  LINKS_,
  SSCHEMES_,
  DSTRUCTURES_,
  EXAMPLES_,
  REPORTING_;
(* ----------------------------------------------------------------------- *)

TYPE

  TdsForm = Class(TForm)
   (* --- Main and Sub Menus --------------------------------------------- *)
    dsMainMenu: TMainMenu;
      mmView: TMenuItem;
        smRanks: TMenuItem;
        smWeights: TMenuItem;
        smNumber: TMenuItem;
      mmEnumerate: TMenuItem;
        nmrStandard: TMenuItem;
        nmrPreOrder: TMenuItem;
        nmrPostOrder: TMenuItem;
        nmrInOrder: TMenuItem;
        nmrBFS: TMenuItem;
      mmHelp: TMenuItem;
        smMemo: TMenuItem;
        smButtonsMemo: TMenuItem;
        smTreeMemo: TMenuItem;
        smNodesMemo: TMenuItem;
        smAbout: TMenuItem;
   (* --- Tool & Status Bar ---------------------------------------------- *)
    dsToolBar: TToolBar;
    dsStatusBar: TStatusBar;
   (* --- Tree & Structure Frames ---------------------------------------- *)
    TreeView: TTreeView;
    Splitter: TSplitter;
    Panel: TPanel;
    ScrollBox: TScrollBox;
    PaintBox: TPaintBox;
   (* --- Task Buttons Tree ---------------------------------------------- *)
    InsertBtn: TBitBtn;
    InsBox: TMaskEdit;
    FindKeyBtn: TBitBtn;
    FindBox: TEdit;
    FindBtn: TBitBtn;
    FindCombo: TComboBox;
    DeleteBtn: TBitBtn;
    PromoteBtn: TBitBtn;
    Report: TBitBtn;
    ReSizeBtn: TBitBtn;
    SizeBox: TEdit;
    PrintSB: TSpeedButton;
    ZoomInSB: TSpeedButton;
    ZoomOutSB: TSpeedButton;
   (* -------------------------------------------------------------------- *)
    procedure PaintBoxPaint(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure TreeViewDblClick(Sender: TObject);
   (* --- View DS Options ------------------------------------------------ *)
    procedure smRanksClick(Sender: TObject);
    procedure smWeightsClick(Sender: TObject);
    procedure smNumberClick(Sender: TObject);
   (* --- Numbering Options ---------------------------------------------- *)
    procedure nmrStandardClick(Sender: TObject);
    procedure nmrPreOrderClick(Sender: TObject);
    procedure nmrPostOrderClick(Sender: TObject);
    procedure nmrInOrderClick(Sender: TObject);
    procedure nmrBFSClick(Sender: TObject);
   (* --- Help Options --------------------------------------------------- *)
    procedure smMemoClick(Sender: TObject);
    procedure smNodesMemoClick(Sender: TObject);
    procedure smButtonsMemoClick(Sender: TObject);
    procedure smTreeMemoClick(Sender: TObject);
    procedure smAboutClick(Sender: TObject);
   (* --- Operations on DS ----------------------------------------------- *)
    procedure InsertBtnClick(Sender: TObject);
    procedure FindKeyBtnClick(Sender: TObject);
    procedure FindBtnClick(Sender: TObject);
    procedure DeleteBtnClick(Sender: TObject);
    procedure PromoteBtnClick(Sender: TObject);
    procedure ReportClick(Sender: TObject);
    procedure ReSizeBtnClick(Sender: TObject);
    procedure PrintSBClick(Sender: TObject);
    procedure ZoomInSBClick(Sender: TObject);
    procedure ZoomOutSBClick(Sender: TObject);
    procedure PaintBoxMouseDown( Sender: TObject; Button: TMouseButton;
                                 Shift: TShiftState; X, Y: Integer);
   (* --- Enter ---------------------------------------------------------- *)
    procedure InsBoxKeyPress(Sender: TObject; var Key: Char);
    procedure FindBoxKeyPress(Sender: TObject; var Key: Char);
    procedure FindComboKeyPress(Sender: TObject; var Key: Char);
    procedure SizeBoxKeyPress(Sender: TObject; var Key: Char);
   (* --- Scrolling ------------------------------------------------------ *)
    procedure FormMouseWheelUp( Sender: TObject; Shift: TShiftState;
                                MousePos: TPoint; var Handled: Boolean);
    procedure FormMouseWheelDown( Sender: TObject; Shift: TShiftState;
                                  MousePos: TPoint; var Handled: Boolean);
   (* -------------------------------------------------------------------- *)
  private
    { Private declarations }
  public
    { Public declarations }
   (* -------------------------------------------------------------------- *)
    CurrentKey : Elements;
    DStructure : Structures;
    _AsFirst_,
    _AsStep_   : Codes;
   (* -------------------------------------------------------------------- *)
    Procedure RefreshForm;
   (* -------------------------------------------------------------------- *)
  end;
(* ----------------------------------------------------------------------- *)

VAR

  dsForm: TdsForm;
(* ----------------------------------------------------------------------- *)

IMPLEMENTATION
(* ----------------------------------------------------------------------- *)

{$R *.DFM}

(* ----------------------------------------------------------------------- *)
procedure TdsForm.FormCreate(Sender: TObject);
begin
 (* ------ *)
  Randomize;
 (* ------ *)
  TreeView.FullExpand;
  FindCombo.ItemIndex := 0;
  CurrentKey := NIL;
  DStructure := NIL;
  _AsFirst_  := _First_;
  _AsStep_   := _Step_;
end;

Procedure TdsForm.RefreshForm;
var
  x1,y1,
  x2,y2 : Integer;
//
  Procedure Check( M: TMenuItem; C: Codes);
  begin
    M.Enabled := (DStructure.Scheme.Go(C)<>NIL);
    M.Checked := ShowNumbers and (C=_AsFirst_);
  end;
//
begin
  If DStructure<>NIL then begin
    With DStructure.Scheme do begin
      With Canvas do begin
        Font.Size  := FontSize;
        CellWidth  := Canvas.TextWidth('8888'); // One cell: 4 digit number
        CellHeight := Canvas.TextHeight('8');   // Each key below previous
       (* --- Do not show content if small -------- *)
        If (FontSize<6) then CellWidth := CellHeight;
      end;
      GridWidth  := CellWidth;
      GridHeight := NmrOfCells*CellHeight;
      If ShowNumbers then GridWidth := GridWidth + (CellWidth Div 2);
      If ShowRanks Or ShowWeights then GridHeight := GridHeight + CellHeight;
      //
      Arrange( 50, 150, GridWidth+CellWidth, GridHeight+CellHeight);
      Frame( x1,y1,x2,y2);
      Enumerate( _AsFirst_, _AsStep_, 0, 1);
    end;
    With DStructure do begin
      dsStatusBar.Panels[0].Text := AsText;
      InsertBtn.Enabled  := (Size<Scheme.MaxCapacity);
    end;
    //
    With ScrollBox do begin
      HorzScrollBar.Range := x2+250+GridWidth;
      VertScrollBar.Range := y2+200+GridHeight;
    end;
    PaintBox.ReFresh;
    //
    Check( nmrStandard,  _First_);
    Check( nmrInOrder,   _InFirst_);
    Check( nmrPreOrder,  _PreFirst_);
    Check( nmrPostOrder, _PostFirst_);
    Check( nmrBFS,       _BFS_);
    smNumber.Checked := ShowNumbers;
  end;
end;

procedure TdsForm.PaintBoxPaint(Sender: TObject);
begin
  If (DStructure<>NIL) then
    With DStructure do
      If (Scheme<>NIL) and (Size<=_MaximumSize_) then begin
        // Ready to gather clickable 'areas'
        NmrOfHitAreas := 0;
        //
        DrawLinks( PaintBox.Canvas, CellWidth Div 2, CellHeight Div 2);
        Scheme.DrawLinks( PaintBox.Canvas, 0,0);
        Scheme.Draw( PaintBox.Canvas, 0,0 );
      end;
end;

procedure TdsForm.TreeViewDblClick(Sender: TObject);
var
  i,
  Index : Integer;
  DS    : Structures;
begin
  DS := NIL;
  With TreeView, Items do
    Index := TreeView.Selected.AbsoluteIndex;
  Case Index of
                                        // 00 :
                                        // 01  :
     02: DS := StructureEG( 1, 1, 30);  // 02  
     03: DS := StructureEG( 1, 2, 30);  // 03  
     04: DS := StructureEG( 1, 3, 30);  // 04  

                                        // 05 :
     06: DS := StructureEG( 2, 1, 30);  // 06  
     07: DS := StructureEG( 2, 2, 30);  // 07  () 

                                        // 08  :
     09: DS := StructureEG( 3, 1, 30);  // 09   
     10: DS := StructureEG( 3, 2, 30);  // 10  () 

                                        // 11  :
     12: DS := StructureEG( 4, 1, 30);  // 12   -  
                                        // 13   :
     14: DS := StructureEG( 4, 2, 30);  // 14  
     15: DS := StructureEG( 4, 3, 30);  // 15  
     16: DS := StructureEG( 4, 4, 30);  // 16   
                                        // 17 :
     18: DS := StructureEG( 4, 5, 30);  // 18   
     19: DS := StructureEG( 4, 6, 30);  // 19   Fibonacci
                                        // 20 :
     21: DS := StructureEG( 4, 7, 30);  // 21   
     22: DS := StructureEG( 4, 8, 30);  // 22   

                                        // 23 :
     24: DS := StructureEG( 5, 1, 30);  // 24   
                                        // 25    :
     26: DS := StructureEG( 5, 2, 30);  // 26  
     27: DS := StructureEG( 5, 3, 30);  // 27 AVL-
     28: DS := StructureEG( 5, 4, 30);  // 38 - 
                                        // 39    :
     30: DS := StructureEG( 5, 5, 30);  // 30 - / 3..6  
     31: DS := StructureEG( 5, 6, 30);  // 31 - / 5..10 
                                        // 32 :
     33: DS := StructureEG( 5, 7, 30);  // 33   
     34: DS := StructureEG( 5, 8, 30);  // 34   .
     35: DS := StructureEG( 5, 9, 30);  // 35  "Rotate-to-Root" 
                                        // 36 :
     37: DS := StructureEG( 5,10, 30);  // 37   
     38: DS := StructureEG( 5,11, 30);  // 38   
     39: DS := StructureEG( 5,12, 30);  // 39   
                                        // 40   :
     41: DS := StructureEG( 5,13, 30);  // 41  
     42: DS := StructureEG( 5,14, 30);  // 42  
     43: DS := StructureEG( 5,15, 30);  // 43  . 
     44: DS := StructureEG( 5,16, 30);  // 44  

                                        // 45  :
     46: DS := StructureEG( 6, 1, 30);  // 46   
     47: DS := StructureEG( 6, 2, 30);  // 47  AVL-
     48: DS := StructureEG( 6, 3, 30);  // 48  - 
  end;
  If DS<>NIL then begin
    If DStructure<>NIL then DStructure.Clear;
    DStructure := DS;
   (* -------------------------------------------------------- *)
    With dsToolBar do
      For i := 0 to ButtonCount-1 do Buttons[i].Enabled := True;
    InsBox.Text        := IntToStr(KeyEG(DS.Size));
    PromoteBtn.Enabled := (11<=Index) And (Index<=22);
   (* -------------------------------------------------------- *)
    MarkUsed := False;
    RefreshForm;
   (* -------------------------------------------------------- *)
  end;
end;

procedure TdsForm.InsertBtnClick(Sender: TObject);
Var
  Sample: Elements;
begin
  If DStructure<>NIL then begin
    MarkUsed := True;
    With DStructure do
      Try
        Sample := KeyData.SetAs( InsBox.Text, _InsColor_);
        Scheme.ColorAll( _FillColor_);
       (* ------------- *)
        INSERT( Sample );
       (* ------------- *)
        Sample.Color := _InsColor_;
        FindBox.Text := InsBox.Text;
        InsBox.Text  := IntToStr(KeyEG(Size));
        CurrentKey   := Sample;
      except
        (* --- never mind --- *)
      end;
    MarkUsed := False;
    RefreshForm;
  end;
end;

procedure TdsForm.FindKeyBtnClick(Sender: TObject);
Var
  Sample : Elements;
begin
  If DStructure<>NIL then begin
    MarkUsed := True;
    With DStructure do
      Try
        Sample := KeyData.SetAs( FindBox.Text, _FindColor_);
        Scheme.ColorAll( _FillColor_);
       (* ---------------------------- *)
        CurrentKey := FINDKEY( Sample );
       (* ---------------------------- *)
        Case CurrentKey<>NIL of
          Ok: begin
                CurrentKey.Color := _FindColor_;
                Sample.Clear;
              end;
          No: begin
                InsBox.Text := FindBox.Text;
              end;
        end;
      except
       (* --- never mind --- *)
      end;
    MarkUsed := False;
    RefreshForm;
  end;
end;

procedure TdsForm.FindBtnClick(Sender: TObject);
Var
  Sample : Elements;
begin
  If DStructure<>NIL then
    With DStructure do begin
      MarkUsed := True;
      Sample := NIL;
      Scheme.ColorAll( _FillColor_);
      //
      Case FindCombo.ItemIndex of
       (* --------------------- *)
        0: begin
             MarkUsed := False;
             Sample := FIND(_Any_);
             MarkUsed := True;
           end;
        1: Sample := FIND(_Top_);
        2: Sample := FIND(_Min_);
        3: Sample := FIND(_Max_);
       (* --------------------- *)
      end;
      //
      If Sample<>NIL then begin
        CurrentKey := Sample;
        CurrentKey.Color := _FindColor_;
        FindBox.Text := IntToStr( KeyData(CurrentKey).Key);
      end;
      MarkUsed := False;
      RefreshForm;
    end;
end;

procedure TdsForm.DeleteBtnClick(Sender: TObject);
Var
  Sample : Elements;
begin
  If (DStructure<>NIL) and (CurrentKey<>NIL) then
    With DStructure do begin
      //
      MarkUsed := True;
      Scheme.ColorAll( _FillColor_);
      MarkUsed := False;
      Sample := FindKey(CurrentKey);
      MarkUsed := True;
      //
      If Sample<>NIL then begin
       (* ------------- *)
        Sample := DELETE;
       (* ------------- *)
        FindBox.Text := IntToStr(KeyData(Sample).Key);
        Sample.Clear;
        CurrentKey := NIL;
      end;
      MarkUsed := False;
      RefreshForm;
    end;
end;

procedure TdsForm.PromoteBtnClick(Sender: TObject);
Var
  Sample : Elements;
  NewKey : Integer;
begin
  If (DStructure<>NIL) and (CurrentKey<>NIL) then
    With DStructure do begin
      //
      MarkUsed := True;
      Scheme.ColorAll( _FillColor_);
      MarkUsed := False;
      With Scheme do begin                        // GoTo 'Current' element
        Go(_First_);
        While Current<>CurrentKey do Go(_Step_);
        Sample := Current;
      end;
      //
      MarkUsed := True;
      If Sample<>NIL then begin                   // Replace it with new one
       (* --- Watch out: equal 'Keys' may occur -------------------------- *)
        With CurrentKey As KeyData do NewKey := Key * 4 div 5;
        Sample := KeyData.SetUp( NewKey, _KeyColor_ );
       (* -------------- *)
        PROMOTE( Sample );
       (* -------------- *)
        Sample.Color := _UpDateColor_;
        FindBox.Text := IntToStr(NewKey);
        CurrentKey   := Sample;
      end;
      MarkUsed := False;
      RefreshForm;
    end;
end;

procedure TdsForm.ReportClick(Sender: TObject);
Var
  K1, K2, K3: KeyData;
  Report: Integer;
begin
  Report := 0;
  NumberIt.SetUp;
  _AsFirst_ := _Reported_;
  _AsStep_  := _Reported_;
  With ReportForm do
    If ShowModal = mrOK then
      If RadioButton1.Checked then Report := 1
      Else
      If RadioButton2.Checked then Report := 2
      Else
      If RadioButton3.Checked then Report := 3;
  //
  If DStructure<>NIL then
    With DStructure do begin
      MarkUsed := True;
      Scheme.Enumerate( _First_, _Step_, 0,0);        // No Numbers
      Scheme.ColorAll( _FillColor_);                  // and No Colors
      Try
        With ReportForm do
          Case Report of
            1: begin
                 K1 := KeyData.SetAs( Edit1.Text, _KeyColor_);
                (* --------------------- *)
                 REPORTLESS( K1, NumberIt);
                (* --------------------- *)
               end;
            2: begin
                 K1 := KeyData.SetAs( Edit2.Text, _KeyColor_);
                 K2 := KeyData.SetAs( Edit3.Text, _KeyColor_);
                (* ---------------------------- *)
                 REPORTinRANGE( K1, K2, NumberIt);
                (* ---------------------------- *)
               end;
            3: begin
                 K1 := KeyData.SetAs( Edit4.Text, _KeyColor_);
                 K2 := KeyData.SetAs( Edit5.Text, _KeyColor_);
                 K3 := KeyData.SetAs( Edit6.Text, _KeyColor_);
                (* ------------------------------- *)
                 REPORTinAREA( K3, K1, K2, NumberIt);
                (* ------------------------------- *)
               end;
          end;
      except
        (* --- never mind --- *)
      end;
      MarkUsed := False;
      RefreshForm;
    end;
end;

procedure TdsForm.ReSizeBtnClick(Sender: TObject);
Var
  NewSize : Integer;
  Sample  : Elements;
begin
  If DStructure<>NIL then
    With DStructure do begin
      MarkUsed := True;
     (* ---------------------------------- *)
      Try
        NewSize := StrToInt(SizeBox.Text);
      except
        On EConvertError do NewSize := Size;
      end;
     (* ---------------------------------- *)
      If (0<=NewSize) and (NewSize<=Scheme.MaxCapacity) then begin
        Scheme.ColorAll( _FillColor_);
        While Size<NewSize do Insert(DataEG(KeyEG(NewSize)));
        While Size>NewSize do begin
          Sample := Find(_Any_);
          If Sample<>NIL then begin
            FindKey(Sample);
            Delete;
          end;
        end;
      end;
      Case NewSize=0 of
        Ok: SizeBox.Text := IntToStr(30+Random(20));
        No: SizeBox.Text := '0';
      end;
      CurrentKey := NIL;
      MarkUsed := False;
      RefreshForm;
    end;
end;

procedure TdsForm.PrintSBClick(Sender: TObject);
begin
  Print;
end;

procedure TdsForm.ZoomInSBClick(Sender: TObject);
begin
  If FontSize<=(16) then begin
    Inc(FontSize);
    RefreshForm;
  end;
  ZoomInSB.Enabled := (FontSize<16);
  ZoomOutSB.Enabled := (FontSize>4);
end;

procedure TdsForm.ZoomOutSBClick(Sender: TObject);
begin
  If FontSize>=(4) then begin
    Dec(FontSize);
    RefreshForm;
  end;
  ZoomInSB.Enabled := (FontSize<16);
  ZoomOutSB.Enabled := (FontSize>4);
end;

procedure TdsForm.smRanksClick(Sender: TObject);
begin
  With smRanks do begin
    Checked := Not Checked;
    ShowRanks := Checked;
  end;
  RefreshForm;
end;

procedure TdsForm.smWeightsClick(Sender: TObject);
begin
  With smWeights do begin
    Checked := Not Checked;
    ShowWeights := Checked;
  end;
  RefreshForm;
end;

procedure TdsForm.smNumberClick(Sender: TObject);
begin
  With smNumber do begin
    Checked := Not Checked;
    ShowNumbers := Checked;
  end;
  RefreshForm;
end;

procedure TdsForm.nmrStandardClick(Sender: TObject);
begin
  _AsFirst_ := _First_;
  _AsStep_  := _Step_;
  ShowNumbers := True;
  RefreshForm;
end;

procedure TdsForm.nmrPostOrderClick(Sender: TObject);
begin
  _AsFirst_ := _PostFirst_;
  _AsStep_  := _PostStep_;
  ShowNumbers := True;
  RefreshForm;
end;

procedure TdsForm.nmrPreOrderClick(Sender: TObject);
begin
  _AsFirst_ := _PreFirst_;
  _AsStep_  := _PreStep_;
  ShowNumbers := True;
  RefreshForm;
end;

procedure TdsForm.nmrInOrderClick(Sender: TObject);
begin
  _AsFirst_ := _InFirst_;
  _AsStep_  := _InStep_;
  ShowNumbers := True;
  RefreshForm;
end;

procedure TdsForm.nmrBFSClick(Sender: TObject);
begin
  _AsFirst_ := _BFS_;
  _AsStep_  := _BFS_;
  ShowNumbers := True;
  RefreshForm;
end;

procedure TdsForm.smMemoClick(Sender: TObject);
begin
  With HelpForm do begin
    PageControl.ActivePage := TabSheet1;
    Show;
  end;
end;

procedure TdsForm.smButtonsMemoClick(Sender: TObject);
begin
  With HelpForm do begin
    PageControl.ActivePage := TabSheet2;
    Show;
  end;
end;

procedure TdsForm.smTreeMemoClick(Sender: TObject);
begin
  With HelpForm do begin
    PageControl.ActivePage := TabSheet3;
    Show;
  end;
end;

procedure TdsForm.smNodesMemoClick(Sender: TObject);
begin
  With HelpForm do begin
    PageControl.ActivePage := TabSheet4;
    Show;
  end;
end;

procedure TdsForm.smAboutClick(Sender: TObject);
begin
  With HelpForm do begin
    PageControl.ActivePage := TabSheet5;
    Show;
  end;
end;

procedure TdsForm.FormMouseWheelUp(Sender: TObject; Shift: TShiftState;
                                   MousePos: TPoint; var Handled: Boolean);
begin
  With ScrollBox do
    HorzScrollBar.Position := HorzScrollBar.Position - 4*ClientWidth div 5;
  Handled := True;
end;

procedure TdsForm.FormMouseWheelDown(Sender: TObject; Shift: TShiftState;
                                     MousePos: TPoint; var Handled: Boolean);
begin
  With ScrollBox do
    HorzScrollBar.Position := HorzScrollBar.Position + 4*ClientWidth div 5;
  Handled := True;
end;

procedure TdsForm.PaintBoxMouseDown( Sender: TObject; Button: TMouseButton;
                                     Shift: TShiftState; X, Y: Integer);
var
  i      : Integer;
  HitKey : String;
  MouseX,
  MouseY : Integer;
begin
 (* --- Absolute Coordinates because X,Y may suffer overflow problems ---- *)
  MouseX := Mouse.CursorPos.X;
  MouseY := Mouse.CursorPos.Y;
 (* --- Relative w.r.t. PaintBox so that all OK in large DStructures ----- *)
  With PaintBox.ClientOrigin do begin
    MouseX := MouseX - X;
    MouseY := MouseY - Y;
  end;
 (* ---------------------------------------------------------------------- *)
  Case Button of
    mbLeft : If (DStructure<>NIL)and(DStructure.Scheme<>NIL) then begin
             (* --- Use 'HitAreas' to pick the pointed element ----------- *)
             (* --- Watch out what happens if double keys exist! --------- *)
               HitKey := '';
               i      := 1;
               While (i<=NmrOfHitAreas) and (HitKey='') do begin
                 With HitAreas[i] do
                   If (x1<=MouseX) and (MouseX<=x2) and
                      (y1<=MouseY) and (MouseY<=y2) then HitKey := Hit;
                 Inc(i);
               end;
               Case (HitKey<>'') of
                 Ok: begin
                       FindBox.Text := HitKey;
                       FindKeyBtnClick(Sender);
                     end;
                 No: begin
                       MarkUsed := True;
                       DStructure.Scheme.ColorAll( _FillColor_);
                       MarkUsed := False;
                       RefreshForm;
                     end;
               end;
             end;
             (* ---------------------------------------------------- *)
  end;
end;

procedure TdsForm.InsBoxKeyPress(Sender: TObject; var Key: Char);
begin
  If Key = #13 then begin InsertBtnClick(Sender); Key := #0; end;
end;

procedure TdsForm.FindBoxKeyPress(Sender: TObject; var Key: Char);
begin
  If Key = #13 then begin FindKeyBtnClick(Sender); Key := #0; end;
end;

procedure TdsForm.FindComboKeyPress(Sender: TObject; var Key: Char);
begin
  If Key = #13 then begin FindBtnClick(Sender); Key := #0; end;
end;

procedure TdsForm.SizeBoxKeyPress(Sender: TObject; var Key: Char);
begin
  If Key = #13 then begin ReSizeBtnClick(Sender); Key := #0; end;
end;

(* ======================================================================= *)

begin
end.
